// *****************************************************************************
// * Copyright (c) 1999      Southeastern Universities Research Association,
// *                         Thomas Jefferson National Accelerator Facility
// * 
// * This software was developed under a United States Government license
// * described in the NOTICE file included as part of this distribution.
// *****************************************************************************
// *
// * File:        PBS_Job.h
// * Author:      Walt Akers
// * Description: This file contains the class definition for the PBS_Job
// *              class.  This class is generally contained by a PBS_Server 
// *              class which provides populates it and provides access to 
// *              its components.
// *****************************************************************************

#ifndef _PBS_JOB_H_
#define _PBS_JOB_H_ 1

#include <PBS_Base.h>
#include <time.h>

// *****************************************************************************
// * class PBS_Job :
// *    This class provides the mechanisms required to read all available
// *    status/configuration information from a PBS job.  It is assumed
// *    that this information will be used later for assessing conditions
// *    when scheduling jobs.
// *
// *    The following data items are defined in this class:
// *
// *         name            : A character string containing the name of the 
// *                           job as it is known to the PBS_Server.
// *         account_name    : A string containing a user defined account name.
// *         checkpoint      : A string containing an attribute identifying
// *                           when checkpointing will be performed by PBS
// *                           on behalf of the job.
// *         depend          : A list of other jobs that this job depends on
// *                           the completion of in order to start.
// *         error_path      : The host:filename that will be used to store
// *                           messages written to stderr.
// *         execution_time  : The time afterwhich the job may be executed -
// *         group_list      : A list of group_names@host that determine 
// *                           the group as which the job is executed on a given
// *         hold_type      : A string containing characters that identify
// *                           the types of holds that have been placed on a
// *                           job:
// *                           u - user hold
// *                           s - system hold 
// *                           o - other hold
// *         job_name        : A string containing the job name that was
// *                           assigned by the user.
// *         join_path       : A boolean value, if TRUE the standard error and 
// *                           standard output wil be merged into the file 
// *                           specified by output_path.
// *         keep_files      : A string value, if it contains the letter 'o' 
// *                           then standard output files will be kept on the
// *                           execution host. If it contains 'e' then standard
// *                           error files will be kept.  By default all files
// *                           are returned to the submission host.
// *         mail_points     : A string identifying at which points mail will
// *                           be sent to the user. It may contain one or more
// *                           of the following flags.
// *                           a - Send mail if the job aborts
// *                           b - Send mail when the job begins
// *         mail_users      : A list of users who will receive mail at the 
// *                           specified mail_points.
// *         output_path     : The host:filename that wil be used to store
// *                           messages written to stdout.
// *                           host.
// *         priority        : The user assigned priority of the job.
// *         rerunable       : A boolean flag, if TRUE the job can be rerun
// *                           if it aborts.
// *         resource_list   : A PBS_Resource object with a list of resources 
// *                           required by the job.
// *         shell_path_list : A list of absolute paths to the shell program on 
// *                           each host that will be used to execute the 
// *                           batch script.
// *         stage_in        : The list of files to be staged in prior to 
// *                           job execution.
// *         stage_out       : The list of jobs to be staged out following 
// *                           job execution.
// *         user_list       : A list of user_names@host that determine the 
// *                           username as which a job will be executed on a 
// *                           given host. 
// *         variable_list   : The list of environment variables passed to the
// *                           queue job request.
// *         comment         : A string that contains comments about the job 
// *                           that were generated by the system.
// *         ctime           : The time that the job was created - seconds since 
// *                           epoch.
// *         etime           : The time that the job became eligible to run - 
// *                           in seconds since epoch.
// *         duration        : This is the job's duration as expressed in the
// *                           walltime resource.
// *         exec_host       : A '+' separated list of the hosts on which the
// *                           job is executing.
// *         egroup          : If the job is queued, this string contains the 
// *                           group as which the job will be run.
// *         euser           : If the job is queued, this string contains the
// *                           user as which the job will be run.
// *         hashname        : Hashname used as the prefix for job output files.
// *         interactive     : A boolean value, that is TRUE if the job is
// *                           an interactive PBS job.
// *         job_owner       : The name and host of the user who submitted the
// *                           job.  The data is provided in name@host format.
// *         job_owner_id    : A locally generated and maintained integer
// *                           that uniquely identifies a username.
// *         job_state       : A character representing the state of the job.
// *                           E - Exiting the job has completed execution, 
// *                               with or without errors, and the batch system
// *                               is doing post-execution clean-up.
// *                           H - Held.  One or more holds has been placed on
// *                               the job.
// *                           Q - Queued.  The job is in a routing or execution
// *                               queue andis awaiting routing or execution.
// *                           R - Running. 
// *                           S - Suspended.  The job was running and has been
// *                               suspended.
// *                           T - Transiting.  The job is in the process of 
// *                               being routed or moved.
// *                           W - Waiting. The job is not being held, but the
// *                               specified execution time has not been 
// *                               reached.
// *         mtime           : The time at which the job was last modified or
// *                           changed states - in seconds since epoch.
// *         proc_cnt        : The computed total number of processors that
// *                           was requested for this job.
// *         qtime           : The time that the job entered the current queue -
// *                           in seconds since epoch.
// *         queue_name      : The name of the queue that holds the job.
// *	     queue_id        : An integer identifier that uniquely identifies
// *	                       the queue.  This value is locally generated and
// *	                       has no significance outside the generating 
// *	                       process.
// *         queue_rank      : An ordered, non-sequential number indicating
// *                           the jobs rank in the queue.
// *         queue_type      : A character identifying the type of queue -
// *                           E - execution
// *                           R - routing
// *         resources_used  : This is a PBS_Resource object that contains a
// *                           list of resources that have been used by the
// *                           job.
// *         session_id      : If the job is running, this integer is the 
// *                           session id of the first executing task.'
// *         startTime       : This is the time that the job was actually 
// *                           started.  It is computed during the loadData
// *                           method, by subtracting the jobs walltime from
// *                           the current time. Presumably, this value cannot
// *                           be weighted in the same manner as the cpu_time.
// *
// *         weight          : This is the job weight as computed by the 
// *                           scheduler.  This value will be calculated during 
// *                           the scheduling cycle and then used to sort the
// *                           list of jobs.
// *****************************************************************************
class PBS_Job : public PBS_Base
{
friend class PBS_Server;
	
public:
	// *********************************************************************
	// * This is the default job duration if no walltime resource is 
	// * specified.
	// *********************************************************************
	enum { DEFAULT_DURATION = (60*60*60) };
	
	// *********************************************************************
	// * This static array of strings represents the names associated
	// * with each of the job states.
	// *********************************************************************
	static char * JobStateList[7];

	// *********************************************************************
	// * These values can be combined bitwise to identify the type of hold
	// * that has been placed on a job.
	// *********************************************************************
	typedef enum {
		HOLD_NONE   = 0,
		HOLD_USER   = 1,
		HOLD_SYSTEM = 2,
		HOLD_OTHER  = 4
		} HoldType;
	
	// *********************************************************************
	// * These values can be combined bitwise to identify the files that 
	// * should be kept on the execution host.
	// *********************************************************************
	typedef enum {
		KEEP_NOTHING = 0,
		KEEP_STDERR  = 1,
		KEEP_STDOUT  = 2
		} KeepFiles;
		
	// *********************************************************************
	// * These values can be combined bitwise to identify the events that
	// * should trigger mail being sent to the user.
	// *********************************************************************
	typedef enum {
		MAIL_DEFAULT  = 0,
		MAIL_ON_ABORT = 1,
		MAIL_ON_BEGIN = 2,
		MAIL_ON_EXIT  = 4
		} MailPoints;
		
	// *********************************************************************
	// * Each of the integers in this enumeration are used to identify a 
	// * unique job state.
	// *********************************************************************
	typedef enum {
		EXITING = 0,
		HELD,
		QUEUED,
		RUNNING,
		SUSPENDED,
		TRANSIT,
		WAITING
		} JobState;
		
protected:
	
	// *********************************************************************
	// * This is a class to hold arrays of strings that represent
	// * dependencies associated with this job.
	// *********************************************************************
	class Dependency
	{
	public:
		char *  dataBuf;
		char ** dataArray;
		
		char ** after;
		char ** afterok;
		char ** afternotok;
		char ** afterany;
		char ** before;
		char ** beforeok;
		char ** beforenotok;
		char ** beforeany;

		Dependency      (void);
		~Dependency     (void);
		int  loadData   (char * str);
		void clear      (void);
	};

	// *********************************************************************
	// * PBS_Server that owns this job.
	// *********************************************************************
	class PBS_Server & server;
	
	// *********************************************************************
	// * Job attributes.
	// *********************************************************************
	char *       name;
	char *       account_name;
	char *       checkpoint;
	Dependency   depend;
	char *       error_path;
	time_t       execution_time;
	char **      group_list;
	int          hold_type;
	char *       job_name;
	Boolean      join_path;
	int          keep_files;
	int          mail_points;
	char **      mail_users;
	char *       output_path;
	int          priority;
	Boolean      rerunable;
	PBS_Resource resource_list;
	char **      shell_path_list;
	char **      stage_in;
	char **      stage_out;
	char **      user_list;
	char **      variable_list;
	char *       comment;
	time_t       ctime;
	time_t       etime;
	time_t       duration;
	char **      exec_host;
	char *       egroup;
	char *       euser;
	char *       hashname;
	Boolean      interactive;
	char *       job_owner;
	int          job_owner_id;
	JobState     job_state;
	time_t       mtime;
	time_t       qtime;
	size_t       proc_cnt;
	char *       queue_name;
	int          queue_id;
	int          queue_rank;
	QueueType    queue_type;
	PBS_Resource resources_used;
	int          session_id;
	time_t       startTime;
	double       weight;
	
	// *********************************************************************
	// * These are loader methods which are called to load data items in
	// * formats that are unique to the PBS_Job object.  Loader methods
	// * for more common formats are defined in the PBS_Base class.
	// *********************************************************************
	void readHoldTypeValue  (int      & val, struct attrl * attrib);
	void readMailPointValue (int      & val, struct attrl * attrib);
	void readJobStateValue  (JobState & val, struct attrl * attrib);
	
public:
	// *********************************************************************
	// * PBS_Job::PBS_Job :
	// *	This constructor is called with a reference to the PBS_Server
	// *	object and a pointer to a batch_status structure that contains
	// *	configuration information for the job.
	// *	
	// *	This method is protected to prevent it's execution by anyone
	// *	except the PBS_Server object.
	// *********************************************************************	
	PBS_Job ( PBS_Server & Server, struct batch_status *status )
		: server         (Server),
		  name           (NULL),
		  account_name   (NULL),
		  checkpoint     (NULL),
		  depend         (),
		  error_path     (NULL),
		  execution_time (0),
		  group_list     (NULL),
		  hold_type      (HOLD_NONE),
		  job_name       (NULL),
		  join_path      (FALSE),
		  keep_files     (KEEP_NOTHING),
		  mail_points    (MAIL_DEFAULT),
		  mail_users     (NULL),
		  output_path    (NULL),
		  priority       (0),
		  rerunable      (TRUE),
		  resource_list  (),
		  shell_path_list(NULL),
		  stage_in       (NULL),
		  stage_out      (NULL),
		  user_list      (NULL),
		  variable_list  (NULL),
		  comment        (NULL),
		  ctime          (0),
		  etime          (0),
		  duration       (DEFAULT_DURATION),
		  exec_host      (NULL),
		  egroup         (NULL),
		  euser          (NULL),
		  hashname       (NULL),
		  interactive    (FALSE),
		  job_owner      (NULL),
		  job_owner_id   (-1),
		  job_state      (QUEUED),
		  mtime          (0),
		  qtime          (0),
		  proc_cnt       (0),
		  queue_name     (NULL),
		  queue_id       (-1),
		  queue_rank     (0),
		  queue_type     (EXECUTION),
		  resources_used (),
		  session_id     (0),
		  startTime      (0),
		  weight         (1.0)
		{
		loadData(status);
		}

	// *********************************************************************
	// * PBS_Job::~PBS_Job :
	// *	The destructor for the object.  It calls the clear method 
	// *	to release all memory that was allocated during the loadData
	// *    process.
	// *	
	// *	This method is protected to prevent it's execution by anyone
	// *	except the PBS_Server object.
	// *********************************************************************
	~PBS_Job ( void )
		{
		clear();
		}
		
	// *********************************************************************
	// * PBS_Job::loadData :
	// *	Calls pbs_statserver to read the pbs server configuration
	// *	strings and then calls various helper functions to parse and 
	// *	store this information.
	// *	
	// *	This method is protected to prevent it's execution by anyone
	// *	except the PBS_Server object.
	// *********************************************************************
	int  loadData   ( struct batch_status * job );
	
	// *********************************************************************
	// * PBS_Job::clear :
	// *	Frees allocated memory and reinitializes variables to their
	// *	default values.
	// *	
	// *	This method is protected to prevent it's execution by anyone
	// *	except the PBS_Server object.
	// *********************************************************************
	void clear ( void );

public:
	// *********************************************************************
	// * PBS_Job::asciiDump :
	// *	Generates a text dump of the contents of the object.  This 
	// *	output is particularly useful for debugging and reporting the 
	// *	state of the server.
	// *********************************************************************
	void asciiDump  ( FILE * fp = stdout );

	// *********************************************************************
	// * Accessor methods :
	// *	These methods provide access to the data that is stored inside
	// *	the PBS_Job object.
	// *********************************************************************
	inline PBS_Server &   getServer                ( void ) { return server; }
	inline const char *   getName                  ( void ) { return (const char *) name; }
	inline const char *   getAccountName           ( void ) { return (const char *) account_name; }
	inline const char *   getCheckpoint            ( void ) { return (const char *) checkpoint; }
	inline const char **  getAfterDependency       ( void ) { return (const char **) depend.after; }
	inline const char **  getAfterOkDependency     ( void ) { return (const char **) depend.afterok; }
	inline const char **  getAfterNotOkDependency  ( void ) { return (const char **) depend.afternotok; }
	inline const char **  getAfterAnyDependency    ( void ) { return (const char **) depend.afterany; }
	inline const char **  getBeforeDependency      ( void ) { return (const char **) depend.before; }
	inline const char **  getBeforeOkDependency    ( void ) { return (const char **) depend.beforeok; }
	inline const char **  getBeforeNotOkDependency ( void ) { return (const char **) depend.beforenotok; }
	inline const char **  getBeforeAnyDependency   ( void ) { return (const char **) depend.beforeany; }
	inline const char *   getErrorPath             ( void ) { return (const char *) error_path; }
	inline time_t         getExecutionTime         ( void ) { return execution_time; }
	inline const char **  getGroupList             ( void ) { return (const char **) group_list; }
	inline int            getHoldType              ( void ) { return hold_type; }
	inline const char *   getUserJobName           ( void ) { return (const char *)job_name; } 
	inline Boolean        getJoinPathFlag          ( void ) { return join_path; }
	inline int            getKeepFilesFlag         ( void ) { return keep_files; }
	inline int            getMailPointFlag         ( void ) { return mail_points; }
	inline const char **  getMailUsers             ( void ) { return (const char **)mail_users; }
	inline const char *   getOutputPath            ( void ) { return (const char *)output_path; }
	inline int            getPriority              ( void ) { return priority; }
	inline Boolean        getRerunableFlag         ( void ) { return rerunable; }
	inline const char **  getShellPathList         ( void ) { return (const char **)shell_path_list; }
	inline const char **  getStageInFiles          ( void ) { return (const char **)stage_in; }
	inline const char **  getStageoutFile          ( void ) { return (const char **)stage_out; }
	inline const char **  getUserList              ( void ) { return (const char **)user_list; }
	inline const char **  getVariableList          ( void ) { return (const char **)variable_list; }
	inline const char *   getComment               ( void ) { return (const char *)comment; }
	inline time_t         getTimeCreated           ( void ) { return ctime; }
	inline time_t         getTimeEligible          ( void ) { return etime; }
	inline time_t         getTimeStarted           ( void ) { return startTime; }
	inline time_t         getDuration              ( void ) { return duration; }
	inline const char **  getExecutionHosts        ( void ) { return (const char **)exec_host; }
	inline const char *   getEffectiveGroup        ( void ) { return (const char *)egroup; }
	inline const char *   getEffectiveUser         ( void ) { return (const char *)euser; }
	inline const char *   getHashname              ( void ) { return (const char *)hashname; }
	inline Boolean        getInteractiveFlag       ( void ) { return interactive; }
	inline const char *   getOwner                 ( void ) { return (const char *)job_owner; }
	inline int            getOwnerID               ( void ) { return job_owner_id; }
	inline JobState       getState                 ( void ) { return job_state; }
	inline const char *   getStateName             ( void ) { return JobStateList[job_state]; }
	inline time_t         getTimeModified          ( void ) { return mtime; }
	inline time_t         getTimeQueued            ( void ) { return qtime; }
	inline const char *   getQueueName             ( void ) { return (const char *)queue_name; }
	inline int            getQueueID               ( void ) { return queue_id; }
	inline int            getQueueRank             ( void ) { return queue_rank; }
	inline QueueType      getQueueType             ( void ) { return queue_type; }
	inline int            getSessionID             ( void ) { return session_id; }
	
	inline PBS_Resource * getRequestedResources    ( void ) { return &resource_list; }
	inline PBS_Resource * getUsedResources         ( void ) { return &resources_used; }

	// *********************************************************************
	// * The weight is an advisor value that is computed by the scheduler.
	// * These methods allow the scheduled to get and set the weight for
	// * a particular job based on local scheduling criteria.
	// *********************************************************************
	inline double         getWeight                ( void )     { return weight; }
	inline void           setWeight                ( double w ) { weight = w; }        

	// *********************************************************************
	// * This is an advisory value that is computed using the nodes and 
	// * nodect resources for the job.
	// *********************************************************************
	inline size_t         getProcCnt               ( void )     { return proc_cnt; }
};



// *****************************************************************************
// * class PBS_JobList:
// *	This is an array of PBS_Job objects and a collection of mechanisms
// *	used to select from among them.
// *****************************************************************************
class PBS_JobList 
{
protected:
	PBS_Job ** jobs;
	size_t     jobCnt;
	
	// *********************************************************************
	// * PBS_JobList::PBS_JobList :
	// *	Default constructor for the class.  It initializes all internal
	// *	variables to their default values.
	// *********************************************************************
	PBS_JobList ( void );

	// *********************************************************************
	// * PBS_JobList::addJobs :
	// *	This method populates the object using an external array of 
	// *	jobs that is termimated by a NULL pointer.  The array becomes 
	// *	the property of the class and should not be accessed directly 
	// *	by the caller.
	// *********************************************************************
	void addJobs ( PBS_Job ** joblist );
	
public:
	// *********************************************************************
	// * PBS_JobList::PBS_JobList :
	// *	Parameterized constructor for the class.  It initializes the
	// *	internal variables using an array of jobs that is terminated
	// *	by a NULL pointer.
	// *
	// *	The array becomes the property of the class and should not be
	// *	accessed directly by the caller.
	// *********************************************************************
	PBS_JobList ( PBS_Job ** joblist );
	
	// *********************************************************************
	// * PBS_JobList::~PBS_JobList :
	// *	Destructor for the object.  It calls clear to free the array 
	// *	structure associated with the object. The PBS_Server is 
	// *	responsible for deleteing the individual PBS_Jobs.
	// *********************************************************************
	~PBS_JobList ( void );

	// *********************************************************************
	// * PBS_JobList::getJob :
	// *	This method returns a specific job by name.  If the job doesn't
	// *	exist in the list, then a NULL will be returned.
	// *********************************************************************
	PBS_Job * getJob ( char * name );
		
	// *********************************************************************
	// * PBS_JobList::getJobCount :
	// *	Returns a count of all jobs that are currently in the list.
	// *********************************************************************
	size_t getJobCount ( void );

	// *********************************************************************
	// * PBS_JobList::getJobs :
	// *	This method returns an array of all jobs that are available in 
	// *	the list.  The user is responsible for deleteing the array 
	// *	that is allocated by this call.
	// *********************************************************************
	PBS_Job ** getJobs ( void );

	// *********************************************************************
	// * PBS_JobList::getJobsByQueue :
	// *	This method returns an array of all jobs in the PBS_Server that 
	// *	are currently in a specified queue.  The caller is responsible 
	// *	for deleting the array that is returned by this call.
	// *********************************************************************
	PBS_Job ** getJobsByQueue ( char * queueName );
	
	// *********************************************************************
	// * PBS_JobList::getJobsByState :
	// *	This method returns an array of all jobs in the PBS_Server that 
	// *	are currently in the specified state.  The caller is 
	// *	responsible for deleting the array that is returned by this call.
	// *********************************************************************
	PBS_Job ** getJobsByState ( PBS_Job::JobState state );
	
	// *********************************************************************
	// * PBS_JobList::get<State>Jobs :
	// *	These methods allow the caller to retrieve a list of jobs that
	// *	are in a particular state.
	// *********************************************************************
	inline PBS_Job **     getTransitingJobs     ( void ) { return getJobsByState(PBS_Job::TRANSIT); }
	inline PBS_Job **     getQueuedJobs         ( void ) { return getJobsByState(PBS_Job::QUEUED); }
	inline PBS_Job **     getHeldJobs           ( void ) { return getJobsByState(PBS_Job::HELD); }
	inline PBS_Job **     getWaitingJobs        ( void ) { return getJobsByState(PBS_Job::WAITING); }
	inline PBS_Job **     getRunningJobs        ( void ) { return getJobsByState(PBS_Job::RUNNING); }
	inline PBS_Job **     getExitingJobs        ( void ) { return getJobsByState(PBS_Job::EXITING); }
	inline PBS_Job **     getSuspendedJobs      ( void ) { return getJobsByState(PBS_Job::SUSPENDED); }

	// *********************************************************************
	// * PBS_JobList::asciiDump :
	// *	Sequentially dumps all of the jobs in the list.
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout );
	
	// *********************************************************************
	// * PBS_JobList::operator [] :
	// *	This method allows the caller to walk through the internal array
	// *	of PBS_Job objects.
	// *********************************************************************
	PBS_Job * operator [] (size_t idx)
		{
		return (idx<jobCnt)?jobs[idx]:NULL;
		}
};
#endif /* _PBS_JOB_H_ */
