// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	This header file defines the PBS_JobCollection object which is used
//	to maintain and return information from the PBS_ServerDatabase.
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#ifndef _JOB_COLLECTION_H_
#define _JOB_COLLECTION_H_ 1

#include "stdio.h"
#include "PBS_Hash.h"

class PBS_JobRecord;

// *****************************************************************************
// * class PBS_JobCollection :
// *	This class collects group of PBS_JobRecord objects into a string hash table.
// *	These jobs are stored and retrieved using their jobID strings.  Each
// *	PBS_JobCollection object has a name value that is used in subclasses to 
// *	uniquely identify it and include it in other hash tables.
// *****************************************************************************
class PBS_JobCollection
{
protected:
	char *        name;
	PBSStringHash jobHash;

public:
	PBS_JobCollection  ( const char * Name, size_t hashSize = 256 );
	~PBS_JobCollection ( void );

	virtual void                   clear   ( void );
	virtual PBS_JobRecord        * addJob  ( PBS_JobRecord * job ); 
	inline virtual PBS_JobRecord * getJob  ( const char * job );
	inline virtual PBS_JobRecord * delJob  ( const char * job );
	inline virtual const char    * getName ( void );

	virtual void asciiDump ( FILE * fp = stdout, char * prefix = "\0");
};

// *****************************************************************************
// * PBS_JobCollection::delJob :
// *	This method removes a PBS_JobRecord object from the jobHash.  The record
// *	is then returned to the caller for deletion or other disposition.
// *****************************************************************************
inline PBS_JobRecord * PBS_JobCollection::delJob ( const char * job )
	{
	return (PBS_JobRecord *)jobHash.remove(job);
	}
	
// *****************************************************************************
// * PBS_JobCollection::getJob :
// *	This method locates a PBS_JobRecord in the jobHash by using the caller 
// *	provided 'job' string.  The PBS_JobRecord that is returned will have a 
// *	JobID that matches the caller provided 'job' string.
// *****************************************************************************
inline PBS_JobRecord * PBS_JobCollection::getJob ( const char * job )
	{
	return (PBS_JobRecord *)jobHash.find(job);
	}
	
// *****************************************************************************
// * PBS_JobCollection::getName :
// *	This method returns the name string that is stored in the class.  This
// *	name string is used to uniquely identify the object and can be used to
// *	store/retrieve it from a hash table.
// *****************************************************************************
inline const char * PBS_JobCollection::getName (  void )
	{
	return (const char * ) name;
	}

#endif /* _JOB_COLLECTION_H_ */
