// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	This file realizes the methods that are defined in the PBS_JobRecord.h
//	file.
//
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#include "PBS_JobRecord.h"
#include "PBS_ServerDatabase.h"
#include "PBS_JobArray.h"
#include "sys/time.h"

// *****************************************************************************
// * PBS_JobRecord::PBS_JobRecord :
// *	This is the constructor, it initializes all of the internals of the 
// *	PBS_JobRecord object to their default values.  Note that the caller MUST
// *	provide a PBS_ServerDatabase object and a job identification string to 
// *	create the object.
// *****************************************************************************
PBS_JobRecord::PBS_JobRecord ( PBS_ServerDatabase & Server, const char * JobID )
	: server        (Server), 
	  jobID         ((JobID && *JobID)?strdup(JobID):NULL),
	  queue         (NULL),
	  owner         (NULL),
	  jobName       (NULL),
	  timeEnqueued  (0),
	  timeStarted   (0),
	  timeCompleted (0),
	  exitStatus    (0),
	  cpuTimeUsed   (0L),
	  wallTimeUsed  (0L),
	  physMemUsed   (0L),
	  virtMemUsed   (0L)
	{
	}
	
// *****************************************************************************
// * PBS_JobRecord::~PBS_JobRecord :
// *	This is the destructor for the PBS_JobRecord object.  It deletes any memory
// *	that was allocated by the class and it removes the object from its 
// *	associated PBS_ServerDatabase, PBS_JobArray, and PBS_JobArray prior to
// *	deletion. 
// *****************************************************************************
PBS_JobRecord::~PBS_JobRecord ( void )
	{
	if(jobID)
		{
		server.delJob(jobID);
		if(queue) queue->delJob(jobID);
		if(owner) owner->delJob(jobID);
		free (jobID);
		}
	if(jobName) free (jobName);
	}
	
// *****************************************************************************
// * PBS_JobRecord::setQueue :
// *	This method will allow the caller to set/change the queue that the job
// *	is a member of. The method will use the name of the queue to locate
// *	a corresponding PBS_JobArray in the PBS_ServerDatabase. 
// *
// *	If a queue has already been specified for this PBS_JobRecord, then the 
// *	delJob method will be used to detach the PBS_JobRecord from the old
// *	queue before calling addJob to attach it to the new queue.
// *****************************************************************************
void PBS_JobRecord::setQueue ( const char * Queue )
	{	
	if(Queue && *Queue)
		{
		PBS_JobArray * newQueue = server.getQueue(Queue);
		if(queue && queue!=newQueue) queue->delJob(jobID);
		queue = newQueue;
		queue->addJob(this);
		}
	}

// *****************************************************************************
// * PBS_JobRecord::setOwner :
// *	This method allows the caller to set/change the PBS_JobArray that the
// *	job is associated with,  The method will use the owner name to locate
// *	a corresonding PBS_JobArray in the PBS_ServerDatabase.
// *	
// *	If an owner has already been specified for the jobRecord, then the 
// *	delJob method will be called to detach the PBS_JobRecord from the old 
// *	owner before calling addjob to attach it to the new owner.
// *****************************************************************************
void PBS_JobRecord::setOwner ( const char * Owner )
	{
	if(Owner && *Owner)
		{
		PBS_JobArray * newOwner = server.getUser(Owner);
		if(owner && owner!=newOwner) owner->delJob(jobID);
		owner = newOwner;
		owner->addJob(this);
		}
	}


// *****************************************************************************
// * PBS_JobRecord::asciiDump :
// *	This method performs a diagnostic dump of the PBS_JobRecord object to 
// *	a specified FILE pointer.  This method is useful for debugging and
// *	general purpose output.
// *****************************************************************************
void PBS_JobRecord::asciiDump ( FILE * fp, char * prefix, int prtHeader )
	{
	if(prtHeader)
		{
		fprintf(fp, "%s%*s     Queue     Owner  Exit   CPU Time  Wall Time    Phys Mem    Virt Mem    Time Enqueued      Time Started      Time Completed    Job Name\n", prefix, strlen(jobID), "Job ID");
		fprintf(fp, "%s-----------------------------------------------------------------------------------------------------------------------------------------------------------------\n", prefix);
		}
		
	char str[128];
	int  a, b, c;
	
	if(owner)
		{
		c = strcspn(owner->getName(), "@");
		strncpy(str, owner->getName(), c);
		str[c]=0;
		}
	else strcpy(str, "UNKNOWN");
	
	fprintf(fp, "%s%s  %8s  %8s  %4d  ", prefix, jobID, queue?queue->getName():"UNKNOWN", str, exitStatus);
	a = (int)(cpuTimeUsed/3600);
	b  = (int)((cpuTimeUsed%3600)/60);
	c  = (int)(cpuTimeUsed%60);
	fprintf(fp, "%3d:%02d:%02d  ", a, b, c);
	a = (int)(wallTimeUsed/3600);
	b  = (int)((wallTimeUsed%3600)/60);
	c  = (int)(wallTimeUsed%60);
	fprintf(fp, "%3d:%02d:%02d  ", a, b, c);
	fprintf(fp, "%7d kb  ", physMemUsed/1024);
	fprintf(fp, "%7d kb  ", virtMemUsed/1024);
	
	struct tm t = *localtime(&timeEnqueued);
	if(timeEnqueued) fprintf(fp, "%02d/%02d/%02d %02d:%02d:%02d  ", t.tm_mon+1, t.tm_mday, t.tm_year%100, t.tm_hour, t.tm_min, t.tm_sec);
	else             fprintf(fp, "                      ");
	
	t = *localtime(&timeStarted);
	if(timeStarted) fprintf(fp, "%02d/%02d/%02d %02d:%02d:%02d  ", t.tm_mon+1, t.tm_mday, t.tm_year%100, t.tm_hour, t.tm_min, t.tm_sec);
	else             fprintf(fp, "                      ");
	
	t = *localtime(&timeCompleted);
	if(timeCompleted) fprintf(fp, "%02d/%02d/%02d %02d:%02d:%02d  ", t.tm_mon+1, t.tm_mday, t.tm_year%100, t.tm_hour, t.tm_min, t.tm_sec);
	else             fprintf(fp, "                      ");
	fprintf(fp, "%s\n", jobName?jobName:"");
	}
