// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	This file realizes the methods that are defined in the PBS_Parameters
//	header file.  
//
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#include "PBS_Parameters.h"
#include "PBS_Profile.h"

PBS_Parameters & PBS_Parameters::attach ( void )
	{
	static PBS_Parameters p;

	return p; 
	}

PBS_Parameters::PBS_Parameters ( void )
	: profile (NULL), configPath(NULL)
	{
	reload();
	}

void PBS_Parameters::reload ( void )
	{
	const char     * ptr;
	char           * base;
	
	if(profile)    delete profile;
	if(configPath) delete[] configPath;
	
	// *********************************************************************
	// * Allocate a new PBS_ProfileManager for accessing data in profile
	// * databases.
	// *********************************************************************
	profile = new ProfileManager;

	// *********************************************************************
	// * Obtain the value of the configuration directory from the local
	// * environment, or use the current directory if it is not defined.
	// *********************************************************************
	if((ptr    = getenv("PBS_CONFIG_DIR"))!=NULL) base = strdup(ptr);
	else base  = strdup("/usr/spool/PBS");
	configPath = new char[strlen(base)+strlen("/scheduler.cfg")+1];
	sprintf(configPath,    "%s%s", base, "/scheduler.cfg");
	free(base);
	
	// *********************************************************************
	// * Call reload here to populate the entries from the disk.
	// *********************************************************************
	profile->reload(configPath);
	}
	
// *****************************************************************************
// * PBS_Parameters::~PBS_Parameters :
// *	This is the destructor for the PBS_Parameters object. It deletes any
// *	memory that was previously allocated to the class.
// *****************************************************************************
PBS_Parameters::~PBS_Parameters ( void )
	{
	delete profile;

	if(configPath)    delete[] configPath;
	}

long PBS_Parameters::getConfigInt ( char * header, char * entry, long Default )
	{
	const char * ptr = profile->getProfileString(configPath, header, entry);
	return (ptr && *ptr)?atoi(ptr):Default;
	}
	
double PBS_Parameters::getConfigFloat ( char * header, char * entry, double Default )
	{
	const char * ptr = profile->getProfileString(configPath, header, entry);
	return (ptr && *ptr)?atof(ptr):Default;
	}

const char * PBS_Parameters::getConfigString ( char * header, char * entry, const char * Default )
	{
	const char * ptr = profile->getProfileString(configPath, header, entry);
	return (ptr && *ptr)?ptr:Default;
	}


const char ** PBS_Parameters::getConfigHeaders ( void )
	{
	return profile->getProfileHeaders(configPath);
	}

const char ** PBS_Parameters::getConfigEntries ( char * header )
	{
	return profile->getProfileEntries(configPath, header);
	}
		 
void PBS_Parameters::asciiDump ( FILE * fp )
	{
	Profile * p;

	fprintf(fp, "CONTENTS OF PBS PARAMETERS OBJECT:\n");
	fprintf(fp, "   Scheduler Config File = %s\n", configPath);
	profile->asciiDump(fp);
	}
