// *****************************************************************************
// * Copyright (c) 1999      Southeastern Universities Research Association,
// *                         Thomas Jefferson National Accelerator Facility
// * 
// * This software was developed under a United States Government license
// * described in the NOTICE file included as part of this distribution.
// *****************************************************************************
// *
// * File:        PBS_Server.h
// * Author:      Walt Akers
// * Description: This file contains the class definition for the PBS_Server
// *              class.  This is the master class for this interface to a
// *              running PBS server.  When instanciated, this class will 
// *              contact the server and extract all known data about the
// *              server, queues, nodes and jobs.  Convenience methods are 
// *              provided to allow the developer to use this data for
// *              scheduling jobs.
// *****************************************************************************

#ifndef _PBS_SERVER_H_
#define _PBS_SERVER_H_

#include <PBS_Base.h>
#include <PBS_Log.h>
#include <PBS_Queue.h>
#include <PBS_Job.h>
#include <PBS_Node.h>

// *****************************************************************************
// * class PBS_Server :
// *    This class provides the mechanisms required to read all available
// *    status/configuration information from the PBS server.  It is assumed
// *    that this information will be used later for assessing conditions
// *    when scheduling jobs.
// *
// *    The following data items are defined in this class:
// *
// *        connectionID     : This is the connection number that is returned
// *                           when pbs_connect is called to attach to a server.
// *        name             : This is the na;me of the PBS Server.
// *        pbs_version      : This is a text string identifying the version
// *                           number of PBS that is running on the server.
// *        server_state     : This is the operating state of the PBS Server, it
// *                           may be one of the following values...
// *                           Active               - processing jobs
// *                           Idle                 - not scheduling jobs
// *                           Scheduling           - waiting for the scheduler 
// *                           Terminating          - performing shutdown
// *                           Terminating. Delayed - stalled during shutdown 
// *        scheduling       : Boolean value indicates if the server is actively
// *                           scheduling jobs.
// *        total_jobs       : The total number of jobs in queues and execution.
// *        state_count      : An array of integers identifing the number of
// *                           jobs in each of six states. The integers are 
// *                           associated with the following job states.
// *                           state_count[0] - Jobs in transit
// *                           state_count[1] - Jobs in a queue
// *                           state_count[2] - Jobs being held
// *                           state_count[3] - Jobs waiting 
// *                           state_count[4] - Jobs that are running
// *                           state_count[5] - Jobs that are exiting
// *         managers        : An array of names of users who have manager 
// *                           rights on the server.  The last string in the 
// *                           array is NULL to indicate the end of the list.
// *         operators       : An array of names of users who have operator 
// *                           rights on the server.  The last string in the 
// *                           array is NULL to indicate the end of the list.
// *         default_queue_name 
// *                         : The name of the default queue where jobs will
// *                           be submitted.
// *         default_node_name 
// *                         : The name of the default node where jobs will be
// *                           executed.
// *         log_events      : A bitmask that identifies which server events 
// *                           will be logged.  PBS 2.1 currently defines the
// *                           following bits.
// *                           1    Internal PBS Errors
// *                           2    Operating System Errors
// *                           4    Administrative Events
// *                           8    Job Events
// *                           16   Job Resource Usage Accounting
// *                           32   Security Events/Violations
// *                           64   Scheduler Events
// *                           128  Level One Debugging Output (Common)
// *                           256  Level Two Debugging Output (Less Common)
// *         mail_from       : The user name that will be associated with the
// *                           mail that is sent from the PBS Server to 
// *                           users to inform them of job completion or
// *                           resource usage.
// *         scheduler_iteration
// *                         : This is the number of seconds at which the 
// *                           scheduler will be periodically called. 
// *         max_running     : The maximum number of jobs that can be running 
// *                           at one time.
// *         max_user_run    : The maximum number of jobs that a user
// *                           can have running at one time.
// *         max_group_run   : The maximum number of jobs that a group can
// *                           have running at one time.
// *         acl_host_enable : This boolean flag indicates that the 
// *                           host access control list must be used to
// *                           identify which hosts can attach to the server.
// *         acl_hosts       : An array of hosts that have permission to 
// *                           attach to the server. The last string in the 
// *                           array is NULL to indicate the end of the list.
// *         acl_user_enable : This boolean flag indicates that the 
// *                           user access control list must be used to
// *                           identify which users can attach to the server.
// *         acl_users       : An array of users that have permission to 
// *                           attach to the server. The last string in the 
// *                           array is NULL to indicate the end of the list.
// *         resources_default
// *         resources_max
// *         resources_available
// *         resources_assigned
// *                         : These are tables of resources that have
// *                           been defined on the server.  See the 
// *                           PBS_Resource class for information on how
// *                           to extract information from these objects.
// *         queueCnt        : The number of queues defined in this server.
// *         queues          : An array of PBS_Queues that are defined in the
// *                           PBS_Server.
// *         loadTime        : This is a time_t value that represents the time
// *                           when the loadData method was called to populate
// *                           the object.
// *****************************************************************************
class PBS_Server : public PBS_Base, 
		   public PBS_JobList,
		   public PBS_QueueList,
		   public PBS_NodeList
{
public:
	// *********************************************************************
	// * An enumerated list of possible server states.
	// *********************************************************************
	typedef enum 
		{
		ACTIVE=0, 
		IDLE,
		SCHEDULING,
		TERMINATING,
		DELAYED_TERMINATING
		} ServerState;

protected:
	// *********************************************************************
	// * An array of strings defined in PBS_Server.cc that provide a text
	// * string name for each of the integer values in the ServerState
	// * enumeration.
	// *********************************************************************
	static char * ServerStateList[5];
	
	// *********************************************************************
	// * These are the data variables that wil be populated from the server.
	// *********************************************************************
	int          connectionID;
	char *       name;
	ServerState  server_state;
	Boolean      scheduling;
	size_t       total_jobs;
	StateCount   state_count;
	char **      managers;
	char **      operators;
	char *       default_queue_name;
	int          log_events;
	char *       mail_from;
	int          scheduler_iteration;
	char *       default_node_name;
	size_t       max_running;
	size_t       max_user_run;
	size_t       max_group_run;
	Boolean      acl_host_enable;
	char **      acl_hosts;
	Boolean      acl_user_enable;
	char **      acl_users;
	char *       pbs_version;
	time_t       loadTime;
	
	PBS_Resource resources_default;
	PBS_Resource resources_max;
	PBS_Resource resources_assigned;
	PBS_Resource resources_available;
		
	// *********************************************************************
	// * These are loader methods which are called to load data items in
	// * formats that are unique to the PBS_Server object.  Loader methods
	// * for more common formats are defined in the PBS_Base class.
	// *********************************************************************
	void readServerStateValue   (ServerState & server_state, struct attrl * attrib);
	void loadQueues             (void);
	void loadJobs               (void);
	void loadNodes              (void);
	
public:
	// *********************************************************************
	// * PBS_Server::PBS_Server :
	// *	This is the default constructor... it initializes all 
	// *	variables to their default values.
	// *********************************************************************	
	inline PBS_Server ( void );
		
	// *********************************************************************
	// * PBS_Server::PBS_Server :
	// *	This constructor is called if the connectionID to the PBS Server
	// *	has already been obtained.  The constructor will initialize
	// *	all internal variables to their default values, and then will 
	// *	call loadData to read configuration information from the 
	// *    server.
	// *********************************************************************	
	inline PBS_Server ( int connID );
		
	// *********************************************************************
	// * PBS_Server::~PBS_Server :
	// *	The destructor for the object.  It calls the clear method 
	// *	to release all memory that was allocated during the loadData
	// *    process.
	// *********************************************************************
	inline ~PBS_Server ( void );
		
	// *********************************************************************
	// * PBS_Server::loadData :
	// *	Calls pbs_statserver to read the pbs server configuration
	// *	strings and then calls various helper functions to parse and 
	// *	store this information.
	// *********************************************************************
	int loadData ( int connID );
	
	// *********************************************************************
	// * PBS_Server::clear :
	// *	Frees allocated memory and reinitializes variables to their
	// *	default values.
	// *********************************************************************
	void clear ( void );

	// *********************************************************************
	// * PBS_Server::asciiDump :
	// *	Generates a text dump of the contents of the object.  This 
	// *	output is particularly useful for debugging and reporting the 
	// *	state of the server.
	// *********************************************************************
	void asciiDump  ( FILE * fp = stdout );

	// *********************************************************************
	// * Accessor methods :
	// *	These methods provide access to the data that is stored inside
	// *	the PBS_Server object.
	// *********************************************************************
	inline int            getConnectionID       ( void ) { return connectionID; }	
	inline const char *   getName               ( void ) { return (const char *)name; }
	inline ServerState    getState              ( void ) { return server_state; }
	inline const char *   getStateName          ( void ) { return ServerStateList[server_state]; }
	inline Boolean        getSchedulingFlag     ( void ) { return scheduling; }
	inline size_t         getTotalJobCount      ( void ) { return total_jobs; }
	inline size_t         getTransitJobCount    ( void ) { return state_count[TRANSIT]; }
	inline size_t         getQueuedJobCount     ( void ) { return state_count[QUEUED]; }
	inline size_t         getHeldJobCount       ( void ) { return state_count[HELD]; }
	inline size_t         getWaitingJobCount    ( void ) { return state_count[WAITING]; }
	inline size_t         getRunningJobCount    ( void ) { return state_count[RUNNING]; }
	inline size_t         getExitingJobCount    ( void ) { return state_count[EXITING]; }
	inline const char **  getManagers           ( void ) { return (const char **)managers; }
	inline const char **  getOperators          ( void ) { return (const char **)operators; }
	inline const char *   getDefaultQueueName   ( void ) { return (const char *)default_queue_name; }
	inline int            getLogEventsFlag      ( void ) { return log_events; }
	inline const char *   getMailName           ( void ) { return (const char *)mail_from; }
	inline int            getSchedulerIter      ( void ) { return scheduler_iteration; }
	inline const char *   getDefaultNodeName    ( void ) { return (const char *)default_node_name; }
	inline size_t         getMaxRunning         ( void ) { return max_running; }
	inline size_t         getMaxUserRunning     ( void ) { return max_user_run; }
	inline size_t         getMaxGroupRunning    ( void ) { return max_group_run; }
	inline Boolean        getACLHostEnableFlag  ( void ) { return acl_host_enable; }
	inline const char **  getACLHostList        ( void ) { return (const char **)acl_hosts; }
	inline Boolean        getACLUserEnableFlag  ( void ) { return acl_user_enable; }
	inline const char **  getACLUserList        ( void ) { return (const char **)acl_users; }
	inline const char *   getPBSVersion         ( void ) { return (const char *)pbs_version; }
	inline PBS_Resource * getDefaultResources   ( void ) { return &resources_default; }
	inline PBS_Resource * getMaxResources       ( void ) { return &resources_max; }
	inline PBS_Resource * getAssignedResources  ( void ) { return &resources_assigned; }
	inline PBS_Resource * getAvailableResources ( void ) { return &resources_available; }
	inline time_t         getLoadDataTime       ( void ) { return loadTime; }
};

// *****************************************************************************
// * PBS_Server::PBS_Server :
// *	This is the default constructor... it initializes all 
// *	variables to their default values.
// *****************************************************************************
inline PBS_Server::PBS_Server ( void )
	: connectionID       (-1),
	  name               (NULL),
	  server_state       (ACTIVE),
	  scheduling         (FALSE),
	  total_jobs         (0),
	  managers           (NULL),
	  operators          (NULL),
	  default_queue_name (NULL),
	  log_events         (0),
	  mail_from          (NULL),
	  scheduler_iteration(0),
	  default_node_name  (NULL),
	  max_running        (0),
	  max_user_run       (0),
	  max_group_run      (0),
	  acl_host_enable    (FALSE),
	  acl_hosts          (NULL),
	  acl_user_enable    (FALSE),
	  acl_users          (NULL),
	  pbs_version        (NULL),
	  PBS_JobList        (),
	  PBS_QueueList      (),
	  PBS_NodeList       (),
	  loadTime           (0)
	{
	memset(state_count, 0, sizeof(state_count));
	}
	
// *****************************************************************************
// * PBS_Server::PBS_Server :
// *	This constructor is called if the connectionID to the PBS Server
// *	has already been obtained.  The constructor will initialize
// *	all internal variables to their default values, and then will 
// *	call loadData to read configuration information from the 
// *    server.
// *****************************************************************************
inline PBS_Server::PBS_Server ( int connID )
	: connectionID       (-1),
	  name               (NULL),
	  server_state       (ACTIVE),
	  scheduling         (FALSE),
	  total_jobs         (0),
	  managers           (NULL),
	  operators          (NULL),
	  default_queue_name (NULL),
	  log_events         (0),
	  mail_from          (NULL),
	  scheduler_iteration(0),
	  default_node_name  (NULL),
	  max_running        (0),
	  max_user_run       (0),
	  max_group_run      (0),
	  acl_host_enable    (FALSE),
	  acl_hosts          (NULL),
	  acl_user_enable    (FALSE),
	  acl_users          (NULL),
	  pbs_version        (NULL),
	  PBS_JobList        (),
	  PBS_QueueList      (),
	  PBS_NodeList       (),
	  loadTime           (0)
	{
	loadData(connID);
	}
			

// *****************************************************************************
// * PBS_Server::~PBS_Server :
// *	The destructor for the object.  It calls the clear method 
// *	to release all memory that was allocated during the loadData
// *    process.
// *****************************************************************************
inline PBS_Server::~PBS_Server ( void )
	{
	clear();
	}
	
#endif /* _PBS_SERVER_H_ */
