// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
//	This header file contains classes to create a stopwatch that can be
// 	used to measure the performance of various applications and tests.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#ifndef _STOP_WATCH_H_
#define _STOP_WATCH_H_

class StopWatch
{
private:
	struct timeval   start_;
	struct timeval   end_;
	struct timeval   pause_;
	int              paused_;
	int              running_;
	
	struct timeval * lap_;
	size_t           lapMax_;
	size_t           lapCnt_;
			
public:
	StopWatch ( void )
		: lapMax_(16)
		{
		lap_ = (struct timeval *)malloc(lapMax_*sizeof(struct timeval));
		clear();
		}

	~StopWatch ( void )
		{
		free(lap_);
		}	

	inline void   clear      ( void );
	inline void   start      ( void );
	inline void   stop       ( void );
	inline void   pause      ( void );
	inline void   resume     ( void );
	inline void   lap        ( void );
	inline double getReading ( void );
	inline double getLap     ( size_t lapIdx );
	inline size_t getLapCnt  ( void );
};

inline void StopWatch::clear ( void )
	{
	start_.tv_sec=(start_.tv_usec)=0;
	end_.tv_sec=(end_.tv_usec)=0;
	pause_.tv_sec=(pause_.tv_usec)=0;
	lapCnt_  = 0;
	running_ = 0;
	paused_  = 0;
	memset(lap_, 0, lapMax_*sizeof(struct timeval));
	}
	
inline void StopWatch::start ( void )
	{
	clear();
	running_ = 1;		
	gettimeofday(&start_, NULL);
	}

inline void StopWatch::stop ( void )
	{
	if(paused_) resume();
	running_ = 0;
	gettimeofday(&end_, NULL);
	}

inline void StopWatch::pause ( void )
	{
	if(!paused_)
		{
		gettimeofday(&pause_, NULL);
		paused_ = 1;
		}
	}

inline void StopWatch::resume ( void )
	{
	if(paused_)
		{
		struct timeval now;
		gettimeofday(&now, NULL);
		
		if(now.tv_usec < pause_.tv_usec)
			{
			now.tv_usec += 1000000;
			now.tv_sec  -= 1;
			}		
		now.tv_sec  -= pause_.tv_sec;
		now.tv_usec -= pause_.tv_usec;
		
		start_.tv_sec  += now.tv_sec;
		if((start_.tv_usec += now.tv_usec)>=1000000)
			{
			start_.tv_sec  += 1;
			start_.tv_usec -= 1000000; 
			}
		
		for(int idx=0; idx<lapCnt_; idx++)
			{
			lap_[idx].tv_sec += now.tv_sec;
			if((lap_[idx].tv_usec += now.tv_usec)>=1000000)
				{
				lap_[idx].tv_sec  += 1;
				lap_[idx].tv_usec -= 1000000;
				}
			}
		paused_ = 0;
		}
	else start();
	}

inline void StopWatch::lap ( void )
	{
	if(running_)
		{
		if(lapCnt_>=lapMax_)
			{	
			lapMax_*=2;
			lap_ = (struct timeval *)realloc(lap_, lapMax_*sizeof(struct timeval));
			}
		if(!paused_) gettimeofday(&lap_[lapCnt_], NULL);
		else 	{
			lap_[lapCnt_].tv_sec = pause_.tv_sec;
			lap_[lapCnt_].tv_usec = pause_.tv_usec;
			}
		lapCnt_++;
		}
	}

inline double StopWatch::getReading ( void )
	{
	struct timeval t;
	
	if(running_)
		{
		if(paused_)
			{
			t.tv_sec = pause_.tv_sec;
			t.tv_usec = pause_.tv_usec;
			}
		else gettimeofday(&t, NULL);
		}	
	else	{
		t.tv_sec  = end_.tv_sec;
		t.tv_usec = end_.tv_usec;
		}
	t.tv_sec  -= start_.tv_sec;
	if((t.tv_usec -= start_.tv_usec)<0)
		{
		t.tv_usec += 1000000;
		t.tv_sec  -= 1;
		}
	return ((double)t.tv_sec)+((double)t.tv_usec)/1000000.0;
	}

inline double StopWatch::getLap ( size_t lapIdx )
	{
	double result = getReading();
	
	if(lapIdx<lapCnt_)
		{
		struct timeval t;
		
		t.tv_sec  = lap_[lapIdx].tv_sec - (lapIdx?lap_[lapIdx-1].tv_sec:start_.tv_sec);
		if((t.tv_usec = lap_[lapIdx].tv_usec - (lapIdx?lap_[lapIdx-1].tv_usec:start_.tv_usec))<0)
			{
			t.tv_usec += 1000000;
			t.tv_sec  -= 1;
			}
		result = ((double)t.tv_sec)+((double)t.tv_usec)/1000000.0;
		}
	return result;
	}

inline size_t StopWatch::getLapCnt ( void )
	{
	return lapCnt_;
	}
	
#endif /* _STOP_WATCH_H_ */
