// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
//	BitField.h contains the class definition for a variable width bitmask
// 	object.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#ifndef _BIT_FIELD_H_
#define _BIT_FIELD_H_ 1

#include <stdio.h>	// Needed for stdout and FILE structure
#include <stdlib.h>	// Needed for memory allocation functions
#include <string.h>	// Needed for memset

// *****************************************************************************
// * class BitField :
// *	This class defines a variable width bit field that is used by the
// *	MutualRange and NodeEntry classes to identify when certain criteria 
// *	have been met.  The bit count starts from 0 and proceeds upward.
// *****************************************************************************
class BitField
{
private:
	// *********************************************************************
	// * size_t valArray :
	// *	This is an array that represents the value of each bit in a
	// *	character.
	// *********************************************************************
	size_t valArray[8];
	
	// *********************************************************************
	// * size_t maxBits :
	// *	This variable defines the maximum number of bits that can be
	// *	be set in the bit field.  By default this value is set to
	// *	32.
	// *********************************************************************
	size_t maxBits;

	// *********************************************************************
	// * size_t charCnt :
	// *	This is the number of characters that were allocated to the
	// *	bits array.  This number is always broken on a 4 byte boundary
	// *	to ensure that the data is alligned properly for comparison.
	// *********************************************************************
	size_t charCnt;
	
	// *********************************************************************
	// * char * bits:
	// *	This is an array of characters that represents the bit field.
	// *	It is allocated at construction to accomodate the number of bits
	// *	specified in the maxBits variable.
	// *********************************************************************
	char * bits;
	
public:
	// *********************************************************************
	// * BitField::BitField :
	// *	This is the constructor for the BitField object - it will 
	// *	allocate a sufficient number of characters to the bits array
	// *	to accomodate the specified number of bits.
	// *********************************************************************
	BitField ( size_t MaxBits = 32 )
		: maxBits (0),
		  charCnt (0), 
		  bits    (NULL)
		{
		// *************************************************************
		// * Never allow a bitfield to be created with less than 32 
		// * bits.
		// *************************************************************
		if(MaxBits<32) MaxBits = 32;
		resize(MaxBits);
				
		// *************************************************************
		// * Populate the valArray object with bit specific values.
		// *************************************************************
		valArray[0] = 1;
		valArray[1] = 2;
		valArray[2] = 4;
		valArray[3] = 8;
		valArray[4] = 16;
		valArray[5] = 32;
		valArray[6] = 64;
		valArray[7] = 128;
		}
		
	// *********************************************************************
	// * BitField::~BitField :
	// *	This is the destructor for the BitField object. It will delete
	// *	the bits array that was allocated during the constructor.
	// *********************************************************************
	~BitField ( void )
		{
		free(bits);
		}
		

	// *********************************************************************
	// * BitField::resize :
	// *	This method will increase the maximum number of bits that
	// *	the object can handle...  This is provided to allow the 
	// *	BitField to be dynamically resized. If the clear flag is
	// *	set (default), then the current contents of the object will
	// *	be elliminated.
	// *	Note: If the caller requests fewer bits than are currently
	// *	allocated, then the object will NOT be downsized.
	// *********************************************************************
	void resize ( size_t MaxBits, int clear = 1 )
		{
		if(MaxBits>maxBits)
			{
			maxBits = MaxBits+(MaxBits%8);
			if(maxBits<32) maxBits = 32;
			charCnt = maxBits/8;
			charCnt = (charCnt+(charCnt%4));
			if(bits) bits=(char *)realloc(bits, charCnt*sizeof(char));
			else     bits=(char *)malloc (charCnt*sizeof(char));
			}
		if(clear) memset(bits, 0, (charCnt*sizeof(char)));
		}
		
	// *********************************************************************
	// * BitField::isSet :
	// *	This returns a boolean result indicating if the specified bit
	// *	number has been set. Remember that the bit count starts from
	// *	0 and goes to 7 for each byte in the array.
	// *********************************************************************
	int isSet ( size_t idx )
		{
		size_t byte = idx/8;
		size_t bit  = idx%8;
		return ((byte<charCnt)?(bits[byte]&valArray[bit]):0);
		}
		
	// *********************************************************************
	// * BitField::set :
	// *	This method turns on one of the bits in the bitfield.
	// *********************************************************************
	void set ( size_t idx )
		{
		size_t byte = idx/8;
		size_t bit  = idx%8;
		if (byte<charCnt) bits[byte]|=valArray[bit];
		}
		
	// *********************************************************************
	// * BitField::unset :
	// *	This method turns off one of the bits in the bitfield.
	// *********************************************************************
	void unset ( size_t idx )
		{
		size_t byte = idx/8;
		size_t bit  = idx%8;
		if (byte<charCnt) bits[byte]^=valArray[bit];
		}

	// *********************************************************************
	// * BitField::clear :
	// *	This method turns off all of the bits in the bitfield.
	// *********************************************************************
	void clear ( void )
		{
		memset(bits, 0, charCnt);
		}
		
	// *********************************************************************
	// * BitField::operator == :
	// *	This method compares two BitField objects for equality.
	// *********************************************************************
	int operator == ( BitField & field )
		{
		return (field.charCnt==charCnt)?(!memcmp(bits, field.bits, charCnt)):0;
		}

	// *********************************************************************
	// * BitField::operator != :
	// *	This method compares two BitField objects for inequality.
	// *********************************************************************
	int operator != ( BitField & field )
		{
		return !(operator ==)(field);
		}

	// *********************************************************************
	// * BitField::asciiDump :
	// *	This is a diagnostic method that will dump the contents of the
	// *	BitField object.
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout )
		{
		for(int idx=0; idx<charCnt; idx++)
			{
			for(int cIdx=0; cIdx<8; cIdx++)
				{
				fprintf(fp, "Bit %5i is %s\n", idx*8+cIdx, bits[idx]&valArray[cIdx]?"ON":"OFF"); 
				}
			}
		}
};

#endif /* _BIT_FIELD_H_ */
