// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distributioon.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	NodeCommodity.cc realizes the methods defined within the NodeCommodity
// 	class.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#include <Brokerage.h>		// Included for NodeCommodity definition
#include <ClusterNode.h>	// Included for ClusterNode class Definition
#include <stdio.h>		// Include for FILE structure and stdio

// *****************************************************************************
// * NodeCommodity::freeList :
// *	This is the static definition for the freeList that is contained within
// *	the NodeCommodity class.
// *****************************************************************************
NodeCommodity * NodeCommodity::freeList;

// *****************************************************************************
// * operator NodeCommodity::new :
// *	This operator is called whenever a new NodeCommodity object is 
// *	requested. The operator will first attempt to extract a NodeCommodity
// *	object from those preallocated on the freeList.  If no NodeCommodity
// *	objects are available from the freeList, then the freeList will be 
// *	repopulated.
// *****************************************************************************
void * NodeCommodity::operator new ( size_t )
	{
	if(freeList==NULL)
		{
		int i;
		
		freeList = ::new NodeCommodity[32];
		for(i=0; i<31; i++) freeList[i].setNext(&freeList[i+1]);
		freeList[i].setNext(NULL);
		}
	NodeCommodity * retval = freeList;
	freeList = retval->getNext();
	retval->setNext(NULL);
	
	return (void *)retval;
	}
	
// *****************************************************************************
// * operator NodeCommodity::delete :
// *	This is the delete operator for the NodeCommodity class.  It will
// *	return the object to the freeList so that it can be provided (without
// *	reallocation) the next time the user requires one.
// *****************************************************************************
void NodeCommodity::operator delete ( void * ptr )
	{
	((NodeCommodity *)ptr)->setNext(freeList);
	freeList = (NodeCommodity *)ptr;
	}
	
	
// *****************************************************************************
// * NodeCommodity::NodeCommodity:
// *	This is a private constructor for the NodeCommodity object.  It will
// *	be called each time the NodeCommodity::new operator has to repopulate
// *	the freeList.
// *****************************************************************************
NodeCommodity::NodeCommodity ( void )
	: next    (NULL),
	  nodeType(0),
	  node    (NULL),
	  holder  (NULL)
	{
	}
	
// *****************************************************************************
// * NodeCommodity::NodeCommodity :
// *	This is the public constructor for the NodeCommodity object. It will
// *	be called when the user needs to allocate a new NodeCommodity object.
// *****************************************************************************
NodeCommodity::NodeCommodity ( ClusterNode & Node, size_t Type )
	: next     (NULL),
	  node     (&Node),
	  nodeType (Type),
	  holder   (NULL)
	{
	}
		
// *****************************************************************************
// * NodeCommodity::~NodeCommodity :
// *	This is the destructor for the NodeCommodity object. Since the object
// *	does not allocate any memory, this method is used as a placeholder only.
// *****************************************************************************
NodeCommodity::~NodeCommodity ( void )
	{
	}
	
// *****************************************************************************
// * NodeCommodity::asciiDump :
// *	This method is called to generate an ascii text report of the contents
// *	of the NodeCommodity object.  This data will be reported to the file
// *	descriptor specified in the fp parameter.  Each line of the report will
// *	be prefixed by the string specified in the prefix parameter. The 
// *	compressed flag indicates that all data should be reported to the same
// *	output line.
// *****************************************************************************
void NodeCommodity::asciiDump ( FILE * fp, char * prefix )
	{
	char c = 0;
	char * prefixString = prefix?prefix:&c;
	
	fprintf(fp, "%sNodeCommodity for node %s\n", 
		prefixString, node->getName());
	fprintf(fp, "%s   => Index : %i\n", 
		prefixString, node->getIndex());
	fprintf(fp, "%s   => Node Types Supported :\n",
		prefixString);
	for(int idx=0; idx<SchedStatic::MAXSUBSETS; idx++)
		{
		if(hasType(idx))
			{
			fprintf(fp, "%s      => %i\n", prefixString, idx);
			}
		}
	}
	
// *****************************************************************************
// * NodeBroker::NodeBroker :
// *	This is the public constructor for the object. It will 
// *	initialize all internal variables.
// *****************************************************************************
NodeBroker::NodeBroker ( void )
	{
	clear();
	}

// *****************************************************************************
// * NodeBroker::~NodeBroker :
// *	This is the destructor for the NodeBroker object.  It currently
// *	performs no real operations.
// *****************************************************************************
NodeBroker::~NodeBroker ( void )
	{
	clear ();
	}

// *****************************************************************************
// * NodeBroker::issueNode :
// *	This method is used to give a node to the NodeBroker object. The
// *	method will install the NodeCommodity object to the end of the
// *	list, will update the typeMask to represent all capabilities
// *	that are provided by this object and will increment the 
// *	nodeCnt variable.
// *****************************************************************************
void NodeBroker::issueNode ( NodeCommodity & node )
	{
	// *********************************************************************
	// * First logically OR the local typeMask with the capabilities
	// * provided by the node.  This will be used to express all of the 
	// * capabilties that are currently available by NodeCommodities
	// * within this NodeBroker.
	// *
	// * NOTE : The typeMask provided by the NodeCommodity has already
	// *        been converted from subset indices to a bitmask... Direct
	// *	    ORing is allowed.
	// *********************************************************************
	typeMask |= node.getType();
	
	// *********************************************************************
	// * Set the holder for the NodeCommodity to point to this 
	// * NodeBroker object.
	// *********************************************************************
	node.setHolder(this);

	// *********************************************************************
	// * Set the next pointer for the NodeCommodity to NULL.
	// *********************************************************************
	node.setNext(NULL);

	// *********************************************************************
	// * Next, increment the nodeCnt to reflect the presence of a new node.
	// *********************************************************************
	nodeCnt++;
	
	// *********************************************************************
	// * Finally, add the new node to the end of the list.
	// *********************************************************************
	if(nodes==NULL) nodes = &node;
	else            nodeTail->setNext(&node);
	nodeTail = &node;
	}


// *****************************************************************************
// * NodeBroker::requestNode :
// *	This method will search through the list of nodes to find one
// *	that represents a specific type.  If the client has this type
// *	of node, then it will provide it to the requestor - otherwise,
// *	a NULL is returned.
// *****************************************************************************
NodeCommodity * NodeBroker::requestNode ( size_t type )
	{
	NodeCommodity * result = NULL;
	
	if(type<SchedStatic::MAXSUBSETS && hasType(type))
		{
		NodeCommodity * prevPtr = NULL;
		NodeCommodity * currPtr = nodes;
		
		// *************************************************************
		// * Set the typeMask to 0 and recalculate it as you walk 
		// * through the list until you find a matching node.
		// *************************************************************
		typeMask = 0;
		while(currPtr && !currPtr->hasType(type))
			{
			// *****************************************************
			// * NOTE : The typeMask provided by the NodeCommodity 
			// *        has already been converted from subset 
			// *        indices to a bitmask... Direct ORing is 
			// *        allowed.
			// *****************************************************
			typeMask |= currPtr->getType();
			prevPtr   = currPtr;
			currPtr   = prevPtr->getNext();
			}
		// *************************************************************
		// * If a matching node was found, remove it from the list
		// * and place its address in the result variable.
		// *************************************************************
		if(currPtr)
			{
			result = currPtr; 
			nodeCnt--;
			if(prevPtr) prevPtr->setNext(currPtr->getNext());
			else nodes = currPtr->getNext();
			currPtr = currPtr->getNext();

			// *****************************************************
			// * Set the holder for the NodeCommodity to NULL - 
			// * indicating that the Commodity is an orphan.
			// *****************************************************
			result->setHolder(NULL);
			
			// *****************************************************
			// * Set the next pointer for the NodeCommodity to NULL.
			// *****************************************************
			result->setNext(NULL);
			}
		// *************************************************************
		// * Continue walking through the list until you have 
		// * obtained the typeMasks associated with all of the 
		// * nodes that remain inside of the NodeBroker object.
		// *************************************************************
		while(currPtr!=NULL)
			{
			// *****************************************************
			// * NOTE : The typeMask provided by the NodeCommodity 
			// *        has already been converted from subset 
			// *        indices to a bitmask... Direct ORing is 
			// *        allowed.
			// *****************************************************
			typeMask |= currPtr->getType();
			currPtr = currPtr->getNext();
			}
		}

	// *********************************************************************
	// * Return the NodeCommodity that was obtained from the list.
	// *********************************************************************
	return result;
	}

// *****************************************************************************
// * NodeBroker::revokeNode :
// *	This method should only be called when the type associated with
// *	a NodeCommodity has been drastically changed.  This will walk
// *	through the list of NodeCommodities to find the one that has
// *	the specified index and will remove it from the list.
// *****************************************************************************
NodeCommodity * NodeBroker::revokeNode ( size_t nodeIdx )
	{
	NodeCommodity * result = NULL;
	
	NodeCommodity * prevPtr = NULL;
	NodeCommodity * currPtr = nodes;
		
	// *********************************************************************
	// * Set the typeMask to 0 and recalculate it as you walk 
	// * through the list until you find a matching node.
	// *********************************************************************
	typeMask = 0;
	while(currPtr && currPtr->getNode().getIndex()!=nodeIdx)
		{
		// *************************************************************
		// * NOTE : The typeMask provided by the NodeCommodity 
		// *        has already been converted from subset 
		// *        indices to a bitmask... Direct ORing is 
		// *        allowed.
		// *************************************************************
		typeMask |= currPtr->getType();
		prevPtr   = currPtr;
		currPtr   = prevPtr->getNext();
		}
	// *********************************************************************
	// * If a matching node was found, remove it from the list
	// * and place its address in the result variable.
	// *********************************************************************
	if(currPtr)
		{
		result = currPtr; 
		nodeCnt--;
		if(prevPtr) prevPtr->setNext(currPtr->getNext());
		else nodes = currPtr->getNext();
		currPtr = currPtr->getNext();

		// *************************************************************
		// * Set the holder for the NodeCommodity to NULL - 
		// * indicating that the Commodity is an orphan.
		// *************************************************************
		result->setHolder(NULL);

		// *************************************************************
		// * Set the next pointer for the NodeCommodity to NULL.
		// *************************************************************
		result->setNext(NULL);
		}
	// *********************************************************************
	// * Continue walking through the list until you have 
	// * obtained the typeMasks associated with all of the 
	// * nodes that remain inside of the NodeBroker object.
	// *********************************************************************
	while(currPtr!=NULL)
		{
		// *************************************************************
		// * NOTE : The typeMask provided by the NodeCommodity 
		// *        has already been converted from subset 
		// *        indices to a bitmask... Direct ORing is 
		// *        allowed.
		// *************************************************************
		typeMask |= currPtr->getType();
		currPtr = currPtr->getNext();
		}

	// *********************************************************************
	// * Return the NodeCommodity that was obtained from the list.
	// *********************************************************************
	return result;
	}
	
// *****************************************************************************
// * NodeBroker::refresh :
// *	Because a NodeCommodity's characteristics may be modified 
// *	externally, each of the brokers must be refreshed prior to
// *	use to ensure that their typeMask is up to date.
// *****************************************************************************
void NodeBroker::refresh ( void )
	{
	nodeCnt  = 0;
	typeMask = 0;
	
	for(NodeCommodity * node = nodes; node!=NULL; node=node->getNext())
		{
		nodeCnt++;
		// *************************************************************
		// * NOTE : The typeMask provided by the NodeCommodity 
		// *        has already been converted from subset 
		// *        indices to a bitmask... Direct ORing is 
		// *        allowed.
		// *************************************************************
		typeMask |= node->getType();
		}
	}
	
// ****************************************************************************
// * NodeBroker::asciiDump :
// *	This method allows the caller to generate an ascii text
// *	diagnostic dump of the contents of the NodeBroker object. 
// ****************************************************************************
void NodeBroker::asciiDump ( FILE * fp, char * prefix )
	{
	char c = 0;
	char * prefixString = prefix?prefix:&c;

	fprintf(fp, "%sNodeBroker for Node Type %i\n", 
		prefixString, typeIndex);
	fprintf(fp, "%s   => Node Count : %i\n", 
		prefixString, nodeCnt);
	fprintf(fp, "%s   => Nodes Required : %i\n",
		prefixString, nodeReq);
	fprintf(fp, "%s   => Nodes :\n",
		prefixString );
	
	for(NodeCommodity * node = nodes; node!=NULL; node=node->getNext())
	
	for(int idx=0; idx<SchedStatic::MAXSUBSETS; idx++)
		{
		fprintf(fp, "%s         %s\n", prefixString, node->getNode().getName());
		}
	}
