// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
//	The file ClusterNode.cc contains the methods and attributes for the 
// 	class definitions in ClusterNode.h
// 	class.  
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#include <CombinedTimeSlots.h>	// Needed for class CombinedTimeSlots
#include <ClusterNode.h>	// Needed for local class definitions
#include <SimpleProcessor.h>	// Needed for class SimpleProcessor

// *****************************************************************************
// * ClaimSortStruct :
// *	This is an implementation object that is not known outside of this
// *	source file.
// *****************************************************************************
typedef struct 
	{
	FreeTimeSlot    * slot;
	SimpleProcessor * proc;
	} ProcClaimSortStruct;
	
// *****************************************************************************
// * ProcClaimSort :
// *	This is the function that qsort will use to sort the ProcClaimSortStruct
// *	objects that are generated during the ClusterNode::claim method.
// *
// *	Return value:
// *		-1  = a is less than b
// *		 0  = a is equal to b
// *		 1  = a is greater than b
// *****************************************************************************
static int ProcClaimSort ( const void * s1, const void * s2 )
	{
	long result = 0;
	
	if(s1 && s2)
		{
		FreeTimeSlot & a = *((ProcClaimSortStruct *)s1)->slot;
		FreeTimeSlot & b = *((ProcClaimSortStruct *)s2)->slot;

		if((result = a.getDuration()-b.getDuration())==0)
			{
			result = b.getStartTime()-a.getStartTime();
			}
		}
	else if (s1) result = -1;
	else if (s2) result =  1;
	
	return (result<0)?(-1):((result>0)?1:0);
	}


// *****************************************************************************
// * ClusterNode::ClusterNode :
// *	This is the public constructor for the object. It will 
// *	initialize all internal variables, allocate each of the 
// *	SimpleProcessor objects and prepare the object for immediate 
// *	utilization.
// *****************************************************************************
ClusterNode::ClusterNode ( PBS_Node & Node, 
			   PBS_KeywordTable & keyTable, 
			   size_t Index)
	: node    (Node),
	  keys    (keyTable),
	  index   (Index),
	  nProcs  (0),
	  slots   (NULL),
	  procs   (NULL),
	  pMatrix (),
	  dirty   (1)
	{
	nProcs = (node.getNumProcessors()<=0)?1:node.getNumProcessors();
	slots  = new CombinedTimeSlots * [nProcs];
	procs  = new SimpleProcessor   * [nProcs];	
	for(int idx=0; idx<nProcs; idx++)
		{
		slots[idx] = NULL;
		procs[idx] = new SimpleProcessor ((char *)node.getName(), idx);
		}

	char ** keywords = (char **)node.getKeywords();
	while(keywords && *keywords) keys.addKey(*(keywords++));

	refresh();
	}

// *****************************************************************************
// * ClusterNode::~ClusterNode :
// *	This is the destructor for the ClusterNode object.  It will 
// *	delete any memory or sub-objects that were created during its
// *	lifecycle.
// *****************************************************************************
ClusterNode::~ClusterNode ( void )
	{
	for(int idx=0; idx<nProcs; idx++)
		{
		if(slots[idx]!=NULL) delete slots[idx];
		if(procs[idx]!=NULL) delete procs[idx];
		}
	delete[] slots;
	delete[] procs;
	}

// *****************************************************************************
// * ClusterNode::refresh :
// *	This method is called to cause the ClusterNode to regenerate
// *	it's FreePeriodMatrix.  This matrix is then used to refresh the
// *	list of CombinedTimeSlots objects.
// *****************************************************************************
void ClusterNode::refresh ( void )
	{
	pMatrix.clear();
	for(int idx=0; idx<nProcs; idx++) pMatrix.append(procs[idx]);
	dirty = 0;
	}
	
// *****************************************************************************
// * ClusterNode::mark :
// *	This method is called internally to set the dirty flag to TRUE
// *	and then to mark all of the CombinedTimeSlots object as dirty.
// *****************************************************************************
void ClusterNode::mark ( void )
	{
	dirty = 1;
	for(int idx=0; idx<nProcs; idx++) if(slots[idx]) slots[idx]->mark();
	}
	
// *****************************************************************************
// * ClusterNode::claim :
// *	This method is called to claim a period of time on a specified
// *	number of processors.  The period of time MUST have been 
// *	calculated from the data returned by the getRange method.  This
// *	method will return an array of character strings that identify
// *	the processors that have been allocated to the job.  It is the
// *	responsibility of the caller to delete the array - but, not
// *	the strings.
// *
// *	Note: In the original design this method called the claim method
// *	for the size specific CombinedTimeSlots object... Unfortunately,
// *	this resulted in a less than optimal usage of the nodes.  The 
// *	CombinedTimeSlots object is still used to determine the periods
// *	of availability on the ClusterNode, however, the SimpleProcessors
// *	are individually polled to find the best fit for the job.
// *****************************************************************************
char ** ClusterNode::claim ( char * job, size_t procCnt, time_t start, time_t duration)
	{
	static ProcClaimSortStruct css[SchedStatic::MAXPROC];
	char ** result = NULL;

	// *********************************************************************
	// * If the number of processors requested is within range AND the 
	// * FreeTimeSlots object for that number of processors has been 
	// * created AND the ClusterNode is NOT dirty - then attempt to 
	// * allocate the range.
	// *********************************************************************
	if(procCnt>0 && procCnt<=nProcs && procCnt<SchedStatic::MAXPROC && !dirty)
		{
		size_t cssIdx = 0;
		size_t prcIdx = 0;

		// *************************************************************
		// * Walk through the list of simple processors and collect
		// * any FreeTimeSlot object that is compatible with the time
		// * requested.  Place these objects (along with the associated
		// * SimpleProcessor pointer) into the ProcClaimSortStruct array.
		// *************************************************************
		for(prcIdx=0; prcIdx<nProcs; prcIdx++)
			{
			if((css[cssIdx].slot=procs[prcIdx]->getSlot(start, duration))!=NULL)
				{
				css[cssIdx++].proc = procs[prcIdx];
				}
			}

		// *************************************************************
		// * If I have collected a sufficient number of FreeTimeSlots
		// * to satisfy the job, then process them.
		// *************************************************************
		if(cssIdx>=procCnt)
			{
			// *****************************************************
			// * If I have more than enough, then use the smallest
			// * and or latest window that will fit the requirement.
			// * This will leave the earlier/larger windows for 
			// * later utilization.
			// *****************************************************
			if(cssIdx>procCnt) 
				qsort(css, cssIdx, sizeof(ProcClaimSortStruct), ProcClaimSort);
			
			// *****************************************************
			// * Allocate an array of pointers to hold the processor
			// * names that will be collected.
			// *****************************************************
			result = new char * [procCnt+1];
			
			// *****************************************************
			// * Walk into the array and claim the first procCnt
			// * processors - placing the name of the processor in
			// * the result array for return to the caller.
			// *****************************************************
			for(prcIdx=0; prcIdx<procCnt; prcIdx++)
				{
				result[prcIdx] = css[prcIdx].proc->claim(job, start, duration);
				}
			result[prcIdx] = NULL;
			}
		}

	// *********************************************************************
	// * Mark the ClusterNode and all underlying CombinedTimeSlots as dirty.
	// *********************************************************************
	mark();
	
	// *********************************************************************
	// * Return the array of strings to the caller.
	// *********************************************************************
	return result;
	}

// *****************************************************************************
// * ClusterNode::incorporate :
// *	The incorporate method is called to add the statistics for
// *	a running job to the underlying SimpleProcessor.
// *****************************************************************************
void ClusterNode::incorporate ( char * proc, char * job, time_t start, time_t duration )
	{
	// *********************************************************************
	// * Because the format of a processor is node/procNum, we will locate 
	// * the last '/' character in the string and then will read the index 
	// * number that follows it into the 'procIdx' variable.
	// *********************************************************************
	char * indexStr = proc?strrchr(proc, '/'):NULL;
	size_t procIdx = indexStr?atoi(indexStr+1):0;
	incorporate(procIdx, job, start, duration);	
	}

// *****************************************************************************
// * ClusterNode::incorporate :
// *	The incorporate method is called to add the statistics for
// *	a running job to the underlying SimpleProcessor.
// *****************************************************************************
void ClusterNode::incorporate ( size_t procIdx, char * job, time_t start, time_t duration )
	{
	if(procIdx<nProcs && procs[procIdx]!=NULL)
		{
		procs[procIdx]->claim(job, start, duration);
		mark();
		}
	}
	
// *****************************************************************************
// * ClusterNode::available :
// *	This method will check each of the underlying SimpleProcessor
// *	objects and will return TRUE if any of the processors is not
// *	either running a job or scheduled to immediately start a job.
// *****************************************************************************
int ClusterNode::available ( void )
	{
	int result = 0;
	
	for(int idx=0; !result && idx<nProcs; idx++)
		{
		result=procs[idx]->available()?1:0;
		}

	return result;
	}
	
// *****************************************************************************
// * ClusterNode::availableTime :
// *	This method will return the longest time slot that is 
// *	immediately available on any ONE processor in the ClusterNode.
// *****************************************************************************
time_t ClusterNode::availableTime ( void )
	{
	time_t result = 0;
	
	for(int idx=0; idx<nProcs; idx++)
		{
		time_t procTime = procs[idx]->availableTime();
		if(result < procTime) result = procTime;
		}
	return result;
	}

// *****************************************************************************
// * ClusterNode::getRange :
// *	This method is called to get a list of free times when a 
// *	specified number of processors are concurrently available. The
// *	caller will then select the best time from this list and will
// *	call the claim method to reserve that period of time.
// *****************************************************************************
RangeList * ClusterNode::getRange ( RangeDefinition & rangeDef )
	{
	RangeList * result  = NULL;
	size_t      procCnt = rangeDef.getProcsPerNode();
	
	// *********************************************************************
	// * First, ensure that the RangeDefinition does not request more
	// * processors than this ClusterNode can provide...
	// *********************************************************************
	if(procCnt>0 && procCnt<=nProcs)
		{
		// *************************************************************
		// * If the ClusterNode is dirty, call refresh to repopulate
		// * the FreeProcMatrix object with valid data.
		// *************************************************************
		if(dirty) refresh();

		// *************************************************************
		// * If the specified CombinedTimeSlots object has never been 
		// * used before, then create a new one.
		// *************************************************************
		if(slots[procCnt-1]==NULL)
			{
			slots[procCnt-1] = new CombinedTimeSlots(procCnt);
			}
		
		// *************************************************************
		// * If the CombinedTimeSlots object for this processor count
		// * has become dirty ( or has never been initialized) call 
		// * the refresh method to populate it with valid data.
		// *************************************************************
		if(slots[procCnt-1]->isDirty())
			slots[procCnt-1]->refresh(pMatrix.merge(procCnt));

		// *************************************************************
		// * Allocate a RangeList obejct that will be used to hold
		// * the range data retrieved from the CombinedTimeSlots object.
		// *************************************************************
		result = new RangeList(*this, rangeDef);
		
		// *************************************************************
		// * Pass the new RangeList objects to the slots object for
		// * processing and population.
		// *************************************************************
		slots[procCnt-1]->getRange(*result);
		}
	// *********************************************************************
	// * Return the result to the caller.
	// *********************************************************************
	return result;
	}

// *****************************************************************************
// * ClusterNode::getUtilization :
// *	This method will return a double value that represents the 
// *	percentage of processors USED on this ClusterNode at a specific
// *	time.  This value will be used in concert with Range objects
// *	to provide better load balancing in a modestly loaded system.
// *
// *	NOTE: The behavior of this function is dependent on the 
// *	SchedStatic::balanceFlag.  If the balanceFlag is TRUE, then the
// *	method will return the percentage of the system that is utilized.
// *	If the balanceFlag is FALSE, then the method will return the 
// *	percentage of the system that is UN-utilized.
// *****************************************************************************
double ClusterNode::getUtilization ( time_t startTime )
	{
	double inUse  = 0.0;
	double total  = nProcs;
	double result = 0.0;
	
	for(int idx=0; idx<nProcs; idx++)
		{
		if(procs[idx] && procs[idx]->scheduled(startTime)) inUse++;
		}

	// *********************************************************************
	// * If loadBalancing is occurring, return the percent of the system
	// * that is inuse.
	// *********************************************************************
	if(SchedStatic::getBalanceFlag()) result = (total>0)?(inUse/total):1.0;

	// *********************************************************************
	// * Otherwise, return the percent of the system that is not inuse.
	// *********************************************************************
	else result = (total>0)?((total-inUse)/total):0.0;

	return result;
	}

// *****************************************************************************
// * ClusterNode::reserveTime :
// *	This method is called when the system needs to reserve a block
// *	of time to prevent it from being utilized by a job.  This method
// *	will walk through the list of processor objects and reserve
// *	the specified period of time on each of them.
// *****************************************************************************
void ClusterNode::reserveTime ( time_t start, time_t duration )
	{
	for(size_t idx=0; idx<nProcs; idx++)
		{
		procs[idx]->reserveTime(start, duration);
		}
	}


// *****************************************************************************
// * ClusterNode::asciiDump :
// *	This method allows the caller to generate an ascii text summary of the
// *	contents of the ClusterNode object.  It will write the contents of this
// *	object and the contents of all of the contained objects to a specified
// *	file pointer.
// *****************************************************************************
void ClusterNode::asciiDump ( FILE * fp, char * prefix )
	{
	int idx;
	char c = 0;
	char * prefixString = prefix?prefix:&c;
	char * subPrefix = new char [strlen(prefixString)+10];
	sprintf(subPrefix, "%s        ", prefixString);
	
	// *********************************************************************
	// * Because the ClusterNode is a repository for dynamic data that is
	// * only updated as needed, call the refresh method if this item is
	// * dirty to clean up the underlying objects.
	// *********************************************************************
	if(dirty) refresh();
	
	fprintf(fp, "%sNode Name %s\n", prefixString, getName());
	fprintf(fp, "%s   - Index      : %i\n", prefixString, index);
	fprintf(fp, "%s   - Num Procs  : %i\n", prefixString, nProcs);
	fprintf(fp, "%s   - Processors :\n", prefixString);
	for(idx=0; idx<nProcs; idx++) procs[idx]->asciiDump(fp, subPrefix);
	fprintf(fp, "%s   - Free Slots :\n", prefixString);
	for(idx=0; idx<nProcs; idx++)
		{
		if(slots[idx]!=NULL) slots[idx]->asciiDump(fp, subPrefix);
		}
	fprintf(fp, "%s   - Keys       :\n", prefixString);
	keys.asciiDump(fp, subPrefix);
	delete[] subPrefix;
	}
