// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	CombinedTimeSlots.cc contains the functions definitions for the 
// 	methods of the CombinedTimeSlots class.  
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#include <CombinedTimeSlots.h>		// Needed for local class definitions
#include <FreeTimeSet.h>		// Needed for class FreeTimeSet
#include <SimpleProcessor.h>		// Needed for class SimpleProcessor

// *****************************************************************************
// * CombinedTimeSlots::CombinedTimeSlots :
// *	This is the constructor for the object.  It is provided with 
// *	a reference to its owning ClusterNode and the number of 
// *	processors that are required to represent a quorum.
// *****************************************************************************
CombinedTimeSlots::CombinedTimeSlots ( size_t Size )
	: size  (Size),
	  dirty (1),
	  slots (NULL)
	{
	}
	
// *****************************************************************************
// * CombinedTimeSlots::~CombinedTimeSlots :
// *	This is the destructor for the CombinedTimeSlots object.  It
// *	is responsible for deleteing all of the FreeTimeSet objects 
// *	that it has allocated to track the periods of availability.
// *****************************************************************************
CombinedTimeSlots::~CombinedTimeSlots ( void )
	{
	while(slots!=NULL)
		{
		FreeTimeSet * slot = slots;
		slots = slot->getNext();
		delete slot;
		}
	}

// *****************************************************************************
// * CombinedTimeSlots::refresh :
// *	This method is called to force the CombinedTimeSlots object
// *	to reload its contents from the data stored in the owning
// *	ClusterNode object.  If the dirty flag is set, this method is 
// *	executed prior to allocating any new periods of availability 
// *	because the underlying FreeTimeSets are corrupted each time a 
// *	job is allocated to the node.
// *****************************************************************************
void CombinedTimeSlots::refresh ( FreeTimeSet * Slots )
	{
	while(slots!=NULL)
		{
		FreeTimeSet * slot = slots;
		slots = slot->getNext();
		delete slot;
		}
	slots = Slots;
	dirty = 0;
	}

// *****************************************************************************
// * CombinedTimeSlots::asciiDump :
// *	This method performs an ascii diagnostic dump of the contents of
// *	the CombinedTimeSlots object.  This includes a list of all 
// *	underlying periods of availability and the associated 
// *	processors.
// *****************************************************************************
void CombinedTimeSlots::asciiDump ( FILE * fp, char * prefix )
	{
	char c = 0;
	char * prefixString = prefix?prefix:&c;
	char * subPrefixString = new char[strlen(prefixString)+5];

	sprintf(subPrefixString, "%s    ", prefixString);
	
	fprintf(fp, "%sCombined Time Slots of %i Processors\n", prefixString, size);
	
	FreeTimeSet * slot = slots;
	while(slot!=NULL)
		{
		slot->asciiDump(fp, subPrefixString);
		slot = slot->getNext();	
		}

	delete subPrefixString;
	}
