// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description: SortableLink.h : This is the header file for the sortable
//               linked list collection of objects.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#ifndef _SORTABLE_LINK_H_
#define _SORTABLE_LINK_H_

// *****************************************************************************
// * class SortableLink :
// *	This is the base class for a single linked list that can be sorted with
// *	the SortableLink::sort method.  It is a pure virtual base class that 
// *	defines the interfaces that the SortableLink::sort method will use to 
// *	sort linked lists of these objects.
// *****************************************************************************
class SortableLink 
{
private:
	// *********************************************************************
	// * size_t 
	// *********************************************************************
	// * size_t sortIndex :
	// *	This value is used by the sort algorithm to determine the
	// *	respective order of different objects in the list.
	// *********************************************************************
	size_t sortIndex;

	
public:
	// *********************************************************************
	// * SortableLink::getNext :
	// *	This method will return a pointer to the next SortableLink
	// *	object in the list.
	// *********************************************************************
	virtual SortableLink * getNext ( void ) = 0;

	// *********************************************************************
	// * SortableLink:setNext :
	// *	This method is used to set the next object in the SortableLink
	// *	object.
	// *********************************************************************
	virtual SortableLink * setNext ( void ) = 0;
	
	// *********************************************************************
	// * SortableLink::isA :
	// *	This method will return a character string (which should be
	// *	static in the implementation class) that uniquely identifies
	// *	the name of the implementation class.
	// *********************************************************************
	virtual const char * isA   ( void ) = 0;	
	
	// *********************************************************************
	// * SortableLink::comp :
	// *	This method will compare two SortableLink objects.  The method will
	// *	return the following value:
	// *
	// *	-1 : object a is less than object b
	// *	 0 : object a is quivalent to object b
	// *	 1 : object a is greater than object b
	// *********************************************************************
	virtual int comp (const SortableLink *a, const SortableLink *b) = 0;

	// *********************************************************************
	// * SortableLink::sort :
	// *	This method is responsible for sorting the SortableLink objects
	// *	based on the user provided comp function.  The caller must 
	// *	provide a reference to the head pointer to the list of objects
	// *	to be sorted.  You'll note that the head pointer can be changed
	// *	during this sort procedure.
	// *********************************************************************
	static void sort ( SortableLink * &list );
};

#endif /* _SORTABLE_LINK_H_ */
