// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description: This file contains the class definitions for a collection of
// 	objects that allow the scheduler to manage available nodes as a 
//	named set.  These sets will be used to allow jobs to be scheduled and
//	run on a restricted list of nodes.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#include <NodeSet.h>
#include <ClusterNode.h>
#include <PBS_Parameters.h>

// *****************************************************************************
// * class NodeSet :
// *	This class is used to store a collection of ClusterNode objects 
// *	associated with a specific list of hosts.  At the start of the 
// *	scheduling cycle, this NodeSet object will be populated with the list 
// *	of operational nodes from the host list.  When a job from that queue 
// *	becomes eligible to run, it will be scheduled against the ClusterNode 
// *	objects from that list.
// *
// *	Note: This object is refreshed using a PBSStringHash object that 
// *	contains all of the operational ClusterNode objects hashed against their
// *	node name.  This will improve performance when filling the matching the
// *	list of candidate nodes against the list of operational nodes.
// *****************************************************************************
class NodeSet 
{
protected:
	char         * name;
	char        ** hosts;
	size_t         hostCnt;
	size_t         hostMax;
	PBSStringHash  hostHash;
	ClusterNode ** nodes;
	size_t         nodeCnt;
	NodeSet      * next;
	
public:	
	NodeSet         ( const char * Name );
	~NodeSet        ( void );
	void associate  ( const char * host );
	void startCycle ( ClusterNode ** available, size_t availableCnt );
	void endCycle   ( void );
	void asciiDump  ( FILE * fp=stdout, char *prefix=NULL);

	NodeSet      * getNext    ( void )           { return next; }
	void           setNext    ( NodeSet * Next ) { next = Next; }
	const char   * getName    ( void )           { return name; }  
	const char  ** getHosts   ( void )           { return (const char **)hosts; }
	size_t         getHostCnt ( void )           { return hostCnt; } 
	ClusterNode ** getNodes   ( void )           { return nodes; }
	size_t         getNodeCnt ( void )           { return nodeCnt; }
};


// *****************************************************************************
// * class NodeSet :
// *	This is the constructor for the NodeSet object. It will take a NodeSet
// *	name, a list of hosts and the nmber of hosts in the list. From this data
// *	it will initialize all of the internal variables so that it can collect
// *	the associated ClusterNodes at the beginning of the scheduling cycle.
// *****************************************************************************
NodeSet::NodeSet ( const char * Name )
	: name    (NULL),
	  hosts   (NULL),
	  hostCnt (0),
	  hostMax (16),
	  hostHash(),
	  nodes   (NULL),
	  nodeCnt (0),
	  next    (NULL)
	{
	name = (Name && *Name)?strdup(Name):NULL;
	
	hosts    = (char **)malloc((hostMax+1)*sizeof(char *));
	hosts[0] = NULL;
	nodes    = (ClusterNode **)malloc((hostMax+1)*sizeof(ClusterNode *));
	nodes[0] = NULL;
	} 

// *****************************************************************************
// * NodeSet::NodeSet :
// *	This is the destructor for the NodeSet object. It will delete any memory
// *	that was allocated by the object during its use.
// *****************************************************************************
NodeSet::~NodeSet ( void )
	{
	if(hosts)
		{
		while(hostCnt>0)
			{
			if(hosts[--hostCnt])
				{
				hostHash.remove(hosts[hostCnt]);
				free(hosts[hostCnt]);
				}
			}
		free(hosts);
		}	
	if(name)  free(name);
	if(nodes) free(nodes);
	}

// *****************************************************************************
// * NodeSet::associate :
// *	This method is called to associate a host name with a NodeSet object. It
// *	will increase the size of the array if necessary to make room for the 
// *	new name.  If the name already exists in the array, then no action will
// *	be taken.
// *****************************************************************************
void NodeSet::associate ( const char * host )
	{
	if(host && *host && hostHash.find(host)==NULL)
		{
		if((hostCnt+1)>=hostMax)
			{
			hostMax*=2;
			hosts = (char **)realloc(hosts, (hostMax+1)*sizeof(char *));
			nodes = (ClusterNode **)realloc(nodes, (hostMax+1)*sizeof(char *));
			}
		hosts[hostCnt]=strdup(host);
		hostHash.insert(hosts[hostCnt], hosts[hostCnt]);
		hostCnt++;
		}
	}


// *****************************************************************************
// * NodeSet::startCycle :
// *	This method is called by the NodeSetManager at the beginning of the 
// *	Scheduling cycle to initialize the list of nodes from the available 
// *	nodes that are referenced in the nodeHash.
// *****************************************************************************
void NodeSet::startCycle ( ClusterNode ** available, size_t availableCnt )
	{
	if(nodeCnt>0) endCycle();
	nodeCnt = 0;
	for(size_t idx=0; idx<availableCnt; idx++)
		{
		if(hostHash.find(available[idx]->getName())!=NULL)
			{
			nodes[nodeCnt++] = available[idx];
			}
		}
	nodes[nodeCnt] = NULL;
	}


// *****************************************************************************
// * NodeSet::endCycle :
// *	This method is called at the end of each scheduling cycle to clear the
// *	list of ClusterNodes that were available to this object during to
// *	schedule or run jobs.
// *****************************************************************************
void NodeSet::endCycle ( void )
	{
	while(nodeCnt>0) nodes[--nodeCnt] = NULL;
	}
	
	
// *****************************************************************************
// * NoteSet::asciiDump :
// *	This method is called to write a diagnostic dump of the contents of the
// *	NodeSet object to the specified file descriptor.
// *****************************************************************************
void NodeSet::asciiDump ( FILE * fp, char * prefix )
	{
	char c = 0;
	if(prefix==0 || *prefix==0) prefix = &c;
	size_t len = strlen((name && *name)?name:"Unknown")+3;
	fprintf(fp, "%s[%s] Node Set : %i Hosts\n", prefix, (name && *name)?name:"Unknown", hostCnt);
	if(hosts) for(size_t idx=0; idx<hostCnt && hosts[idx]!=NULL; idx++)
		{
		fprintf(fp, "%s%*c%s\n", prefix, len, ' ', hosts[idx]);
		}
	fprintf(fp, "\n");
	}

// *****************************************************************************
// * NodeSetManager::NodeSetManager:
// *	This is the constructor for the NodeSetManager object. It will read from
// *	the configuration files a list of all queues that are registered in the
// *	'Restricted Queues' section.  It will then read the list of nodes
// *	associated with each of these queues and will create a NodeSet object
// *	that contains the node list retrieved from the configuration file.
// *
// *	If a queue is listed as restructed, only those nodes that are listed
// *	in its scheduler.cfg file will be available for processing.
// *****************************************************************************
NodeSetManager::NodeSetManager ( void )
	: available   (NULL), 
	  availableCnt(0),
	  nodeSets    (NULL),
	  nodeSetHash ()
	{
	PBS_Parameters & params = PBS_Parameters::attach();
	
	const char ** entries = params.getConfigEntries("Restricted Nodes");
	for(size_t idx=0; entries && entries[idx]; idx++)
		{
		const char * ptr;
		char         queueName[256];
		const char * queueList = params.getConfigString("Restricted Nodes", (char *)entries[idx], NULL);
		
		if((ptr=queueList)!=NULL) while(*ptr)
			{
			size_t qIdx=0;
			
			while(*ptr && (isspace(*ptr) || *ptr==',')) ptr++;
			while(*ptr && !isspace(*ptr) && *ptr!=',' && qIdx<255)
				{
				queueName[qIdx++] = *(ptr++);
				}
			queueName[qIdx] = 0;
			if(*queueName)
				{
				NodeSet * set = (NodeSet *)nodeSetHash.find(queueName);
				if(!set)
					{
					set = new NodeSet(queueName);
					nodeSetHash.insert(set->getName(), set);
					set->setNext(nodeSets);
					nodeSets = set;
					}
				set->associate(entries[idx]);
				}
			if(*ptr) ptr++;
			}
		}
	if(entries) free(entries);
	}


// *****************************************************************************
// * NodeSetManager::~NodeSetManager :
// *	This is the destructor for the NodeSetManager object.  It will free all
// *	memory that was allocated during the objects lifetime.
// *****************************************************************************
NodeSetManager::~NodeSetManager ( void )
	{
	endCycle();
	while(nodeSets!=NULL)
		{
		NodeSet *curr = nodeSets;
		nodeSets = curr->getNext();
		delete curr; 
		}
	}

// *****************************************************************************
// * NodeSetManager::startCycle :
// *	This method is called at the beginning of the scheduling cycle with a
// *	list of all of the available ClusterNode objects.  A ClusterNode is
// *	considered available if its State is NOT "Down", "Unknown" or "OffLine".
// *	The constructor will then use this list of nodes to populate the 
// *	NodeSet objects that correspond to each of the "Restricted Nodes" 
// *	definitions.
// *****************************************************************************
void NodeSetManager::startCycle ( ClusterNode ** Available, size_t NodeCnt )
	{
	PBSStringHash nodeHash;
	const char *  namePtr;
	
	if(available) endCycle();
	available    = Available;
	availableCnt = NodeCnt;
	
	NodeSet * curr = nodeSets;
	while(curr!=NULL)
		{
		curr->startCycle(available, availableCnt);
		curr = curr->getNext();
		}
	}


// *****************************************************************************
// * NodeSetManager::endCycle :
// *	This method is called at the end of the scheduling cycle to clear the 
// *	non-local variables that were passed to the object during the 
// *	startCycle method.
// *****************************************************************************
void NodeSetManager::endCycle ( void )
	{
	available = NULL;
	availableCnt = 0;
	NodeSet * curr = nodeSets;
	while(curr!=NULL)
		{
		curr->endCycle();
		curr=curr->getNext();
		}
	}



// *****************************************************************************
// * NodeSetManager::getNodes :
// *	This method will obtain a list of all of the available nodes that are
// *	associated with a queue.  In the event that the specified name was not
// *	registered, then a list of all of the available ClusterNode objects 
// * 	will be returned.
// *****************************************************************************
ClusterNode ** NodeSetManager::getNodes ( const char * name, size_t & cnt )
	{
	ClusterNode ** result   = available;
	NodeSet      * selected = NULL;
	
	cnt    = availableCnt;
	
	if(nodeSets!=NULL &&
	  (selected = (NodeSet *)nodeSetHash.find(name))!=NULL)
		{
		result = selected->getNodes();
		cnt    = selected->getNodeCnt();
		}
	return result;
	}

// *****************************************************************************
// * NodeSetManager::asciiDump :
// *	This method generates an ascii summary of the contents of the 
// *	NodeSetManager object.
// *****************************************************************************
void NodeSetManager::asciiDump ( FILE * fp, char * prefix )
	{
	char c = 0;
	if(prefix==0 || *prefix==0) prefix=&c;

	NodeSet * curr = nodeSets;
	
	fprintf(fp, "%s-------------------------------------\n", prefix);
	fprintf(fp, "%sContents of the NodeSetManager Object\n", prefix);
	fprintf(fp, "%s-------------------------------------\n", prefix);
	while(curr!=NULL)
		{
		curr->asciiDump(fp, prefix);
		curr = curr->getNext();
		}
	}
