// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	Range.cc realizes the methods defined within the Range
// 	class.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#include <Range.h>		// Needed for Range class definition
#include <RangeList.h>		// Needed for RangeList class definition
#include <FreeTimeSet.h>	// Needed for FreeTimeSet class definition

// *****************************************************************************
// * Range::freeList :
// *	This is the static definition for the freeList that is contained within
// *	the Range class.
// *****************************************************************************
Range * Range::freeList;

// *****************************************************************************
// * operator Range::new :
// *	This operator is called whenever a new Range object is 
// *	requested. The operator will first attempt to extract a Range
// *	object from those preallocated on the freeList.  If no Range
// *	objects are available from the freeList, then the freeList will be 
// *	repopulated.
// *****************************************************************************
void * Range::operator new ( size_t )
	{
	if(freeList==NULL)
		{
		int i;
		
		freeList = ::new Range[32];
		for(i=0; i<31; i++) freeList[i].next = &freeList[i+1];
		freeList[i].next = NULL;
		}

	Range * retval = freeList;
	freeList       = retval->next;
	retval->next   = NULL;
	
	return (void *)retval;
	}
	
// *****************************************************************************
// * operator Range::delete :
// *	This is the delete operator for the Range class.  It will
// *	return the object to the freeList so that it can be provided (without
// *	reallocation) the next time the user requires one.
// *****************************************************************************
void Range::operator delete ( void * ptr )
	{
	((Range *)ptr)->next = freeList;
	freeList = (Range *)ptr;
	}
	
	
// *****************************************************************************
// * Range::Range : 
// *	This is a private constructor for the Range object.  It will
// *	be called each time the Range::new operator has to repopulate
// *	the freeList.
// *****************************************************************************
Range::Range ( void )
	: owner(NULL), rangeStart(0), rangeDuration(0), utilization(0.0)
	{
	}
	
// *****************************************************************************
// * Range::Range :
// *	This is the public constructor for the Range object. It will
// *	be called when the user needs to allocate a new Range object.
// *****************************************************************************
Range::Range ( RangeList & Owner )
	: owner(&Owner), rangeStart (0), rangeDuration(0), utilization(0.0)
	{
	}
	
// *****************************************************************************
// * Range::Range :
// *	This is the public constructor for the Range object. It will
// *	be called when the user needs to allocate a new Range object. Note
// *	that before this method is called, the owning RangeList object must
// *	confirm that the timeSet has sufficient duration to hold the job...
// *	otherwise, the rangeDuration will have an invalid value.
// *****************************************************************************
Range::Range ( RangeList & Owner, FreeTimeSet & timeSet )
	: owner(&Owner), rangeStart (0), rangeDuration(0), utilization(0.0)
	{
	setStartTime(timeSet.getStartTime());
	setDuration (timeSet.getDuration()-Owner.getDuration());
	}
	
// *****************************************************************************
// * Range::~Range :
// *	This is the destructor for the Range object. Since no memory is 
// * 	allocated locally, it is a placeholder.
// *****************************************************************************
Range::~Range ( void )
	{
	}
	
// *****************************************************************************
// * Range::asciiDump :
// *	This method will generate an asciiDump of the contents of the
// *	Range object.
// *****************************************************************************
void Range::asciiDump ( FILE * fp, char * prefix )
	{
	char c = 0;
	char * prefixString = prefix?prefix:&c;
	
	fprintf(fp, "%sTime Range For Node %s\n", 
		prefixString, owner->getNode().getName());
	fprintf(fp, "%s   - Start Time  : %s\n", 
		prefixString, SchedStatic::getTimeString(rangeStart));
	fprintf(fp, "%s   - End Time    : %s\n", 
		prefixString, SchedStatic::getTimeString(getEndTime()));
	fprintf(fp, "%s   - Duration    : %i:%i:%i\n", 
		prefixString, rangeDuration/3600, (rangeDuration%3600)/60, 
		rangeDuration%60);
	fprintf(fp, "%s   - Utilization : %3.0f %%\n", 
		prefixString, utilization*100.0);
	}
