// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
//	Range.h contains the class definition for the methods and attributes
//	of the Range class.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#ifndef _RANGE_H_
#define _RANGE_H_ 1

#include <stdlib.h>		// Needed for size_t
#include <stdio.h>		// Needed for FILE structure and stdio
#include <RangeList.h>		// Needed for RangeList::getNode
#include <ClusterNode.h>		// Needed for ClusterNode::getUtilization

// *****************************************************************************
// * Class declarations for classes that will be used or referenced locally. By
// * minimizing the inclusion of header files, many external dependencies can
// * be elliminated.
// *****************************************************************************
class FreeTimeSet;

// *****************************************************************************
// * class Range:
// *	The Range class maintains a list of periods of availability for a
// *	specific job on a node (as specified in the owned RangeList object).
// *	The durations of these period of availability differ in this object
// *	from those specified in the FreeTimeSet objects.  The duration
// *	value in the range object indicates the period of time after the 
// *	startTime during which the job can be started and still complete
// *	without interfering with other scheduled jobs.
// *
// *	Because Range objects are dynamically allocated and deallocated 
// *	regularly, the class will maintain a freeList that will manage
// *	these objects when they are not in use.  Reducing the amount of 
// *	memory that is allocated and deallocated. 
// *****************************************************************************
class Range 
{
private:
	// *********************************************************************
	// * static Range * freeList:
	// *	This is the free list for the class. Whenever a user requests
	// *	a new Range object - it will be provided from those that are 
	// *	preallocated and stored in this list.  When the user deletes the
	// *	Range object it willbe returned to this list for later
	// *	distribution.
	// *********************************************************************
	static Range * freeList;
	
	// *********************************************************************
	// * Range * next;
	// *	This is the pointer to the next Range object in the linked list.
	// *	This pointer is also used to hold the deleted Range objects
	// *	on the freeList.
	// *********************************************************************
	Range * next;
	
	// *********************************************************************
	// * time_t rangeStart:
	// *	This is the time that period of availability for this Range 
	// *	object begins.
	// *********************************************************************
	time_t rangeStart;
	
	// *********************************************************************
	// * time_t rangeDuration:
	// *	This is the period of time following the rangeStart during 
	// *	which the job indicated in the RangeList object can be started
	// *	and still have time to finish without interfering with other
	// *	scheduled jobs.
	// *********************************************************************
	time_t rangeDuration;
	
	// *********************************************************************
	// * double utilization :
	// *	This is the utilization value from the node at the startTime
	// *	for this range.  This value can be used to provide an additional
	// *	weighting factors for Ranges that are otherwise identical.
	// *********************************************************************
	double utilization;

	// *********************************************************************
	// * RangeList * owner:
	// *	This is the owner of this Range object.  It maintains a list
	// *	of the pertinent job parameters, a reference to the node that
	// *	this Range is from and a list of all other Ranges that exist
	// *	on that node for that job.
	// *********************************************************************
	RangeList * owner;

	// *********************************************************************
	// * Range::Range :
	// *	This is the private constructor for the object.  It is used to
	// *	preallocate Range objects that will then be distributed from 
	// *	the FreeList.
	// *********************************************************************
	Range ( void );

public:
	// *********************************************************************
	// * Range::operator new :
	// *	This is the new operator for the class. Whenever the caller
	// *	requests a new Range object - this operator will provide
	// *	it from the freeList.  When the freeList becomes empty, it
	// *	will be repopulated automatically by this method.
	// *********************************************************************
	void * operator new ( size_t );
	
	// *********************************************************************
	// * Range::operator delete :
	// *	This is the delete operator for the class.  Whenever a Range
	// *	object is deleted, it will be returned to the freeList for later
	// *	reallocation.
	// *********************************************************************
	void operator delete ( void * );
	
	// *********************************************************************
	// * Range::Range :
	// *	This is the public constructor for this object.  When the user
	// *	requires a new Range object, this constructor will be called
	// *	with a reference to the owning RangeList.
	// *********************************************************************
	Range ( RangeList & Owner );
	
	// *********************************************************************
	// * Range:Range :
	// *	This is a constructor for the object.  Whenever the caller 
	// *	wishes to allocate a new Range object, tis constructor will be
	// *	called with a reference to the owning RangeList and a reference
	// *	to the FreeTimeSet object that contains the startTime and
	// *	duration.
	// *********************************************************************
	Range ( RangeList & Owner, FreeTimeSet & timeSet );
	
	// *********************************************************************
	// * Range::~Range :
	// *	This is the destructor for the object.
	// *********************************************************************
	~Range ( void );
	
	// *********************************************************************
	// * Range::asciiDump :
	// *	This method will generate an asciiDump of the contents of the
	// *	Range object.
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout, char * prefix = NULL );
	
	// *********************************************************************
	// * These are the inline accessor methods used to access the contents
	// * of the Range object.
	// *********************************************************************
	inline time_t        getStartTime   ( void )      { return rangeStart; }
	inline time_t        getEndTime     ( void )      { return rangeStart+rangeDuration; }
	inline time_t        getDuration    ( void )      { return rangeDuration; }
	inline double        getUtilization ( void )      { return utilization; }
	inline Range       * getNext        ( void )      { return next; }
	inline ClusterNode & getNode        ( void )      { return owner->getNode(); }	
	inline void          setDuration    ( time_t x )  { rangeDuration = x; }
	inline void          setNext        ( Range * x ) { next = x; }

	// *********************************************************************
	// * Range::setStartTime :
	// *	Whenever the startTime for a Range is modified, the Range should
	// *	contact its owning node to obtain the utilization fraction for
	// *	that time.
	// *********************************************************************
	inline void setStartTime ( time_t x )  
		{ 
		rangeStart = x; 
		utilization = owner?owner->getNode().getUtilization(x):1.0;
		}
};

#endif /* _RANGE_H_ */
