// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	This file contains the class definition of methods and attributes of
//	the RangeDefinition class.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#ifndef _RANGE_DEFINITION_H_
#define _RANGE_DEFINITION_H_ 1

#include <PBS_KeywordTable.h>	// Needed for class PBS_KeySet member object

// *****************************************************************************
// * Class declarations for classes that will be used or referenced locally. By
// * minimizing the inclusion of header files, many external dependencies can
// * be elliminated.
// *****************************************************************************
class ClusterNode;

// *****************************************************************************
// * class RangeDefinition:
// *	The RangeDefinition class defines the interface used to specify a
// *	jobs requirements.  This is the base class for the RangeList object
// *	and the JobCandidateSubSet object.
// *
// *	At construction the caller specifies the keywords associated with the
// *	job, the number of nodes required, the number of processors per node,
// *	and the job duration.  The scheduling system will then locate a 
// *	node or group of nodes that meet these requirements.
// *****************************************************************************
class RangeDefinition
{
protected:
	// *********************************************************************
	// * PBS_KeySet keys: 
	// *	These are the keywords that are associated with the 
	// *	RangeDefinition.  The caller must add keys to this list by 
	// *	using the addKey method.
	// *********************************************************************
	PBS_KeySet keys;

	// *********************************************************************
	// * time_t duration :
	// *	This is the user specified duration of the job in walltime.
	// *********************************************************************
	time_t duration;
	
	// *********************************************************************
	// * size_t nNodes;
	// *	This is the number of nodes that are required by the job.
	// *********************************************************************
	size_t nNodes;
	
	// *********************************************************************
	// * size_t nProcs;
	// *	This is the number of processors per node that are required
	// *	for the job.
	// *********************************************************************
	size_t nProcs;
	
public:
	// *********************************************************************
	// * RangeDefinition::RangeDefinition :
	// *	This is the constructor for the object.  It must be provided
	// *	with the following parameters. 
	// *
	// *	Duration : The walltime required of the nodes requested.
	// *	NNodes   : The number of nodes required.
	// *	NProcs   : The number of processors required.
	// *	keyTable : The PBS_KeywordTable used to manage keywords.
	// *********************************************************************
	RangeDefinition ( time_t, size_t, size_t, PBS_KeywordTable & );
	
	// *********************************************************************
	// * RangeDefinition::RangeDefinition :
	// *	This is the constructor for the object.  It will copy the
	// *	pertinent contents from an existing RangeDefinition object.
	// *********************************************************************
	RangeDefinition ( RangeDefinition & );

	// *********************************************************************
	// * RangeDefinition::~RangeDefinition :
	// *	This is teh destructor for the object.  Because this class will
	// *	almost always be used as a base for other objects,  this is 
	// *	a virtual place holder to allow the child class's constructor
	// *	to be called.
	// *********************************************************************
	virtual ~RangeDefinition ( void );
	
	// *********************************************************************
	// * RangeDefinition::operator ==:
	// *	This operator is used to determine if a RangeDefinition object's
	// *	requirements match the capabilities of a ClusterNode object.
	// *********************************************************************
	int operator == ( ClusterNode & );
	
	// *********************************************************************
	// * These are the inline accessor methods used to access the contents 
	// * of the RangeDefinition object.
	// *********************************************************************
	inline time_t       getDuration     ( void ) { return duration; }
	inline size_t       getNodesReq     ( void ) { return nNodes; }
	inline size_t       getProcsPerNode ( void ) { return nProcs; }
	inline PBS_KeySet & getKeys         ( void ) { return keys; }
	
	// *********************************************************************
	// * RangeDefinition::addKey :
	// *	This method is used to add a new key string to the PBS_KeySet
	// *	object (keys).
	// *********************************************************************
	inline void addKey ( char * key ) { keys.addKey(key); }

	// *********************************************************************
	// * RangeDefinition::asciiDump :
	// *	This method allows the caller to generate an ascii text report
	// *	of the contents of the RangeDefinition object.  This data may
	// *	be used either for debugging or for reporting purposes.
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout, char * prefix = NULL );
};

#endif /* _RANGE_DEFINITION_H_ */
