// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
//	RangeList.cc contains the method definitions for the member functions
// 	of the RangeList class.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#include <RangeList.h>		// Needed for local class definitions
#include <Range.h>		// Needed for class Range
#include <FreeTimeSet.h>	// Needed for class FreeTimeSet
#include <ClusterNode.h>	// Needed for class ClusterNode

// *****************************************************************************
// * RangeList::RangeList :
// *	This is the constructor for the RangeList object.  It is 
// *	responsible for initializing the internal variables.  The user
// *	must provide a reference to the owning ClusterNode object, the
// *	duration, number of nodes requested, number of processors-per-	
// *	node requested and a PBS_Keywordtable that can be used to 
// *	initialize the underlying RangeDefinition object.
// *****************************************************************************
RangeList::RangeList ( ClusterNode & Node, time_t duration, size_t nNodes, size_t nProcs, PBS_KeywordTable & keyTable)
	: RangeDefinition ( duration, nNodes, nProcs, keyTable ),
	  node            ( Node ),
	  range           ( NULL ),
	  next            ( NULL )
	{
	}

// *****************************************************************************
// * RangeList::RangeList :
// *	This is an alternate form of constructor.  It can be called 
// *	using only a reference to the ClusterNode object and a 
// *	reference to a RangeDefinition object.
// *****************************************************************************
RangeList::RangeList ( ClusterNode & Node, RangeDefinition & Def )
	: RangeDefinition ( Def ),
	  node            ( Node ),
	  range           ( NULL ),
	  next            ( NULL )
	{
	}

// *****************************************************************************
// * RangeList::~RangeList :
// *	This is the destructor for the RangeList class. It will delete
// *	any Range objects that have allocated during its lifetime.
// *****************************************************************************
RangeList::~RangeList ( void )
	{
	while ( range != NULL )
		{
		Range * curr = range;
		range = curr->getNext();
		delete curr;
		}
	}

// *****************************************************************************
// * RangeList::addRanges :
// *	This method will walk through a linked list of FreeTimeSet objects 
// *	and will add all FreeTimeSet objects that are compatible with the 
// *	jobs requirements.
// *****************************************************************************
void RangeList::addRanges ( FreeTimeSet * timeSets )
	{
	while(timeSets != NULL)
		{
		if(timeSets->getDuration()>=duration)
			addRange(*timeSets);
		timeSets = timeSets->getNext();
		}
	}

// *****************************************************************************
// * RangeList::addRange :
// *	This method will update the list of Ranges to reflect the free
// *	time provided by the new FreeTimeSet.  This is a highly complex
// *	method and should be read carefully prior to making modifications.
// *****************************************************************************
void RangeList::addRange ( FreeTimeSet & timeSet)
	{
	// *********************************************************************
	// * If the duration of the timeSet is less than the duration of the 
	// * job, then there is no reason to continue.
	// *********************************************************************
	if(timeSet.getDuration()<duration) return;
	
	// *********************************************************************
	// * For simplicity and to minimize calls, extract the pertinent data
	// * from the timeSet.  Note that the duration is the FreeTimeSet's
	// * duration minus the duration of the job that is being considered.
	// * Consequently, the tDuration variable contains the length of time
	// * after the startTime during which the job can be STARTED.
	// *********************************************************************
	time_t  tStart    = timeSet.getStartTime();
	time_t  tDuration = timeSet.getDuration()-duration;
	time_t  tEnd      = tStart+tDuration;
	
	// *********************************************************************
	// * Begin walking through the list of Range objects and stop when
	// * the Range object's startTime begins AT or AFTER the tStart.
	// *********************************************************************
	Range * prev=NULL, *curr = range;
	while(curr && (curr->getStartTime()<tStart))
		{
		prev = curr;
		curr = curr->getNext();
		}
	
	// *********************************************************************
	// * Inspect the prev object (if valid) and determine if the previous
	// * Range overlaps with this time set.  If it does - then extend it
	// * to reflect the overall window of availability.
	// *********************************************************************
	if(prev && (prev->getEndTime())>=tStart)
		{
		// *************************************************************
		// * If the window of availability for the new FreeTimeSet 
		// * lasts longer than the one in the prev Range... then,
		// * recalculate the duration of the Range object to reflect 
		// * the later time.
		// *************************************************************
		if(tEnd>prev->getEndTime()) 
			prev->setDuration(tEnd-prev->getStartTime());
		
		// *************************************************************
		// * Reset the curr variable to point to the prev Range...
		// * this variable will be used later when assessing if 
		// * any of the other ranges are overlapping.
		// *************************************************************
		curr = prev;
		}
	// *********************************************************************
	// * If the prev Range did not overlap with the FreeTimeSet, determine
	// * if the current range overlaps with the new FreeTimeSet. If it
	// * does, extend it to reflect the overall window of availability.
	// *********************************************************************
	else if(curr && (tEnd>=curr->getStartTime()))
		{
		// *************************************************************
		// * It will be necessary now to determine which of the objects
		// * (the FreeTimeSet or the curr Range) has the longest 
		// * duration - by comparing endTimes.  The duration of the
		// * range should be set from the latest endTime after the 
		// * Range's startTime has been updated.
		// *************************************************************
		if(tEnd>curr->getEndTime()) curr->setDuration(tEnd-tStart);
		else curr->setDuration(curr->getEndTime()-tStart);
		
		// *************************************************************
		// * Update the start time for the curr Range to reflect the 
		// * startTime of the new FreeTimeSet.
		// *************************************************************
		curr->setStartTime(tStart);
		}
	// *********************************************************************
	// * If the new FreeTimeSet overlaps with neither the previous nor the 
	// * current Ranges.  Then create a new Range and insert it between
	// * the two. 
	// *********************************************************************
	else	{
		Range * nRange = new Range(*this, timeSet);
		nRange->setNext(curr);

		if(prev==NULL) range = nRange;
		else           prev->setNext(nRange);
		
		// *************************************************************
		// * Set curr to NULL to prevent any more checks for overlapping
		// * times...
		// *************************************************************
		curr = nRange;
		}

	// *********************************************************************
	// * The final step is to walk forward from the curr Range and determine
	// * if any other adjacent Ranges were overlapped by the new 
	// * FreeTimeSet.
	// *********************************************************************
	while(curr && curr->getNext() && 
	     (curr->getEndTime()>=curr->getNext()->getStartTime())) 
		{
		// *************************************************************
		// * Extract the overlapped range.
		// *************************************************************
		Range * deadRange = curr->getNext();
		
		// *************************************************************
		// * Set the next range to the one after the overlapped range.
		// *************************************************************
		curr->setNext(deadRange->getNext());

		// *************************************************************
		// * Update the current Range to reflect the latest duration.
		// *************************************************************
		if(deadRange->getEndTime()>curr->getEndTime())
			curr->setDuration(deadRange->getEndTime()-curr->getStartTime());
		
		// *************************************************************
		// * Delete the deadRange and continue.
		// *************************************************************
		delete deadRange;
		}
	}

// *****************************************************************************
// * RangeList::asciiDump : 
// *	This method generates an ascii report of the contents of the RangeList
// *	object.  This data can then be used to evaluate that the object is
// *	functioning properly.
// *****************************************************************************
void RangeList::asciiDump ( FILE * fp, char * prefix )
	{
	char c = 0;
	char * prefixString = prefix?prefix:&c;
	
	fprintf(fp, "%sAvailable ranges on node %s - index %i\n", 
		prefixString, node.getName(), node.getIndex());
	fprintf(fp, "%s   - Procs required : %i\n", 
		prefixString, nProcs);
	fprintf(fp, "%s   - Job duration   : %i:%02i:%02i\n", 
		prefixString, duration/3600, (duration%3600)/60, duration%60);
	
	Range * curr = range;
	while(curr)
		{
		fprintf(fp, "%s    Job can start from %s to ", 
			prefixString, SchedStatic::getTimeString(curr->getStartTime()));
		fprintf(fp, "%s\n", SchedStatic::getTimeString(curr->getEndTime()));
		
		curr = curr->getNext();
		}
	}
