// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
//	RangeTable.cc contains the method definitions for the member functions
// 	of the RangeTable class.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#include <RangeTable.h>		// Needed for local class definitions
#include <ClusterNode.h>	// Needed for class ClusterNode
#include <JobCandidate.h>	// Needed for class JobCandidate
#include <JobCandidateSubSet.h>	// Needed for class RangeDefinition/JobCandidateSubSet
#include <JobDeployment.h>	// Needed for class JobDeployment
#include <MutualRange.h>	// Needed for class MutualRange
#include <ClusterRange.h>	// Needed for class ClusterRange definition

// *****************************************************************************
// * RangeTable::RangeTable :
// *	This is the constructor for the RangeTable object.  It sets all 
// *	variables to their default value.
// *****************************************************************************
RangeTable::RangeTable ( void )
	: nTypes   (0),
	  histogram(NULL),
	  job      (NULL)
	{
	clear();
	}

// *****************************************************************************
// * RangeTable::clear :
// *	This method is called to clear the RangeTable object in 
// *	preparation for it to be used to schedule the next job.
// *****************************************************************************
inline void RangeTable::clear ( void )
	{
	job    = NULL;
	nTypes = 0;
	memset(types, 0, sizeof(RangeDefinition *)*SchedStatic::MAXSUBSETS);
	
	while(histogram!=NULL)
		{
		MutualRange * range = histogram;
		histogram = range->getNext();
		delete range;
		}
	}	

// *****************************************************************************
// * RangeTable::init :
// *	This method initializes the RangeTable object with data that is 
// *	specific to the JobCandidate that is being scheduled.
// *****************************************************************************
void RangeTable::init ( JobCandidate * Job )
	{
	// *********************************************************************
	// * Save the address of the JobCandidate object so that it can be
	// * used later to extract necessary information such as the job name.
	// *********************************************************************
	job = Job;
	
	// *********************************************************************
	// * If the JobCandidate is NON-NULL and the number of subset types
	// * is less than SchedStatic::MAXSUBSETS...
	// *********************************************************************
	if(job && (nTypes = job->getSubsetCnt())<SchedStatic::MAXSUBSETS)
		{
		// *************************************************************
		// * Install a pointer to each defined subset into the types
		// * array.  In reality, these objects are JobCandidateSubSets,
		// * however, for the purposes of this class they are being
		// * stored as their base class - RangeDefinition.
		// *************************************************************
		for(size_t idx=0; idx<nTypes; idx++)
			{
			types[idx] = job->getSubset(idx);
			}
		}

	// *********************************************************************
	// * If there is no job or there is no type specification, then clear
	// * the object and return.
	// *********************************************************************
	if(job==NULL || nTypes==0) clear();
	return;
	}

// *****************************************************************************
// * RangeTable::process :
// *	This method is called with an array of ClusterNode objects that 
// *	represent the nodes where the job can be scheduled.  This method will
// *	examine the job requirements and the available nodes and will determine
// *	the best time and place for scheduling this job.  The method will return
// *	a JobDeployment object that reflects the optimal scheduling scenario.
// *****************************************************************************
JobDeployment * RangeTable::process ( ClusterNode ** nodes, size_t nodeCnt)
	{
	JobDeployment * result = NULL;
	size_t          typeIdx;
	
	// *********************************************************************
	// * The failed flag is used to indicate that there is no possibility of 
	// * executing the job with the nodes that are currently available.
	// *********************************************************************
	int failed = 0;
	
	// *********************************************************************
	// * If no job has been specified or there are no types, or no nodes
	// * have been provided... then it will be impossible to schedule this
	// * job for deployment.  Go ahead and fail.
	// *********************************************************************
	if(!(job && nTypes && nodes && nodeCnt)) failed = 1;

	// *********************************************************************
	// * Count up all of the nodes that have been requested for all types
	// * defined by the job.  If the requested number of nodes exceeds the
	// * number of nodes available, then the job cannot be scheduled.
	// *********************************************************************
	size_t allReqCnt = 0;
	for(typeIdx=0; typeIdx<nTypes && !failed; typeIdx++)
		{
		if((allReqCnt+=types[typeIdx]->getNodesReq())>nodeCnt) failed=1;
		}

	// *********************************************************************
	// * The typeIdx variable will point to the RangeDefinition that is 
	// * currently being processed. Walk the typeIdx variable from the 
	// * first types entry to the last, processing each entry as you go.
	// *********************************************************************
	for(typeIdx=0; !failed && typeIdx<nTypes; typeIdx++)
		{
		// *************************************************************
		// * Create a ClusterRange object that will be used to collect
		// * the RangeLists returned by each node.
		// *************************************************************
		ClusterRange ranges;
		
		// *************************************************************
		// * The nodeIdx variable will point to the entries in the nodes 
		// * array - starting with the first.
		// *************************************************************
		for(size_t nodeIdx = 0; !failed && nodeIdx<nodeCnt; nodeIdx++)
			{
			// *****************************************************
			// * Ensure that the ClusterNode is of the correct type
			// * prior to requesting a RangeList from it.
			// *****************************************************
			if((*types[typeIdx]) == (*nodes[nodeIdx]))
				{
				RangeList * list = nodes[nodeIdx]->getRange(*types[typeIdx]);
				if(list) ranges.incorporate(*list);
				}
			}

		// *************************************************************
		// * Test to see if sufficient nodes exist of the specified
		// * type.  If they do not - then fail this job because it
		// * cannot be satisfied by the existing resources.
		// *************************************************************
		if(ranges.getNodeCnt()<types[typeIdx]->getNodesReq()) failed = 1;
		else
			{
			// *****************************************************
			// * Extract the ordered list of Ranges from the 
			// * ClusterRange object.
			// *****************************************************
			Range * allRanges = ranges.getRange();

			// *****************************************************
			// * The currRange object represents the last used 
			// * MutualRange object.  Since all Range objects should
			// * progress forward in time, this variable elliminates
			// * the need to walk through MutualRange objects that
			// * we already know are too early.
			// *****************************************************
			MutualRange * currRange = histogram;
			
			// *****************************************************
			// * While Range objects remain in the list, pass each
			// * one to the MutualRange object for processing and
			// * disposition.
			// *****************************************************
			while (allRanges!=NULL)
				{
				// *********************************************
				// * In order to reduce complexity in this code,
				// * I am using a separate method (installRange)
				// * to incorporate the individual Range objects
				// * into the histogram. It will return a 
				// * pointer to the MutualRange object where the
				// * Range object was installed.
				// *********************************************
				currRange = installRange(currRange, *allRanges, typeIdx);
				
				// *********************************************
				// * Increment the allRanges pointer to the next
				// * Range object in the list.
				// *********************************************
				allRanges = allRanges->getNext();
				}
			// *****************************************************
			// * Call compact to elliminate any MutualRange objects
			// * that have not met the node requirements up to this
			// * point...  If they don't have the elements they need
			// * now, then they never will.
			// * 
			// * Note
			// *****************************************************
			compact(typeIdx);
			}
		
		// *************************************************************
		// * If the histogram is NULL at any time after the first group
		// * of ranges has been installed, then we know that this job
		// * will never be successful... Call fail to elliminate further
		// * processing.
		// *************************************************************
		if(histogram==NULL) failed = 1;
		}
	
	// *********************************************************************
	// * If a valid MutualRang has been generated, its time to turn that 
	// * into a JobDeployment object and populate it with processors.
	// *********************************************************************
	if(histogram && !failed) result =  histogram->claim();
		
	// *********************************************************************
	// * Return the resultant JobDeployment object - May be NULL if a 
	// * fault occurred.
	// *********************************************************************
	return result;
	}

// *****************************************************************************
// * MutualRange * RangeTable::installRange :
// *	This method is provided to reduce the complexity of the 
// *	RangeTable::process method.  It is private to the class and is called 
// *	exclusively by the process method to install an individual range
// *	to the correct location in the list of MutualRanges.
// *
// *	This method will return the MutualRange object where the allRanges
// *	object was installed.
// *****************************************************************************
MutualRange * RangeTable::installRange (MutualRange * currPtr, Range & range, size_t type )
	{
	// *********************************************************************
	// * If the histogram has not been craeted, then generate it using the
	// * specifications in the range object.
	// *********************************************************************
	if(histogram==NULL)
		{
		histogram = new MutualRange(*this, 
			range.getStartTime(), 
			range.getDuration());
		}

	// *********************************************************************
	// * If, for some reason, the range occurs earlier in time than the
	// * current histogram object - them roll the currPtr back to the 
	// * beginning of the list.
	// *********************************************************************
	if(currPtr==NULL || range.getEndTime()<currPtr->getStartTime()) 
		{
		currPtr=histogram;
		}

	// *********************************************************************
	// * Walk forward in the list of MutualRange objects until we find one
	// * that is either overlapping - or later than the current range.
	// *
	// * While the end time of the mutual range is less than the start time
	// * of the inbound range.
	// *********************************************************************
	MutualRange * prevPtr = currPtr;
	while(currPtr && currPtr->getEndTime()<range.getStartTime())
		{
		prevPtr = currPtr;
		currPtr = prevPtr->getNext();
		}
	
	// *********************************************************************
	// * If we have run off the end of the list, set the currPtr to the
	// * prevPtr and then add the range to that object.
	// *********************************************************************
	if(currPtr==NULL) currPtr = prevPtr;
	
	// *********************************************************************
	// * Since range objects are presented 'ordered by time and length', 
	// * logically, if a range object arrives that is earlier than the 
	// * first range in the list - then it must be of a different 'type'...
	// * Since it is earlier than the first MutualRange, it's time period 
	// * will inevitably lack one or more of the characteristics required 
	// * to  create a valid job environment... Do not use it.
	// *********************************************************************
	if(range.getEndTime()>currPtr->getStartTime()) 
		{
		// *************************************************************
		// * At this time, call the add method of the Range of the 
		// * current MutualRange object. It will walk forward within its
		// * own list, adding and modifying all of the additional 
		// * MutualRange objects necessary to represent the time 
		// * interval.  The MutualRange pointer returned from this 
		// * method, represents the first MutualRange object in the list 
		// * where the new range was successfully implemented.
		// *************************************************************
		currPtr  = currPtr->add(range, type);
		}

	// *********************************************************************
	// * Return the most recently utilized MutualRange object.
	// *********************************************************************
	return currPtr;
	}

// *****************************************************************************
// * RangeTable::compact :
// *	This method will walk through the MutualRange objects in the
// *	histogram and will ensure that they are satisfied for all types
// *	upto and including the type number specified in the typeIdx.
// *
// *	If a MutualRange object is not satisfied by the nodes that it 
// *	has now, then there is no way that it will be satisfied later.
// *****************************************************************************
void RangeTable::compact ( size_t typeIdx )
	{
	MutualRange * prevPtr = NULL;
	MutualRange * currPtr = histogram;
	
	// *********************************************************************
	// * Ensure that the typeIdx being passed to this method is NOT greater
	// * than the maximum value.
	// *********************************************************************
	if(typeIdx >= SchedStatic::MAXSUBSETS) typeIdx = SchedStatic::MAXSUBSETS-1;
	if(typeIdx >= nTypes)                  typeIdx = nTypes-1;
	
	// *********************************************************************
	// * Set the lastPass variable to TRUE if we are processing the final
	// * index.  This variable can trigger an early return if a satisfied
	// * MutualRange is located.
	// *********************************************************************
	int lastPass = (typeIdx==(nTypes-1));
	
	// *********************************************************************
	// * Start walking through the histogram, elliminate any MutualRange
	// * objects that are not satisfied up to the current index.
	// *********************************************************************
	while(currPtr!=NULL)
		{
		// *************************************************************
		// * If the currPtr is not satisfied...
		// *************************************************************
		if(!currPtr->satisfied(typeIdx))
			{
			// *****************************************************
			// * Remove the currPtr from the linked list.
			// *****************************************************
			if(prevPtr) prevPtr->setNext(currPtr->getNext());
			else        histogram = currPtr->getNext();
			
			// *****************************************************
			// * Delete the currPtr
			// *****************************************************
			delete currPtr;
			}
		// *************************************************************
		// * Otherwise, set the prevPtr to the currPtr.
		// *************************************************************
		else	{
			prevPtr = currPtr;
			
			// *****************************************************
			// * If we are testing the LAST subset type, the code
			// * will return as soon as it finds ONE fully satisfied
			// * MutualRange object.  By design, this MutualRange
			// * object will be at the top of the histogram AND it
			// * will be the earliest possible time.
			// *****************************************************
			if(lastPass) return;
			}
		// *************************************************************
		// * Advance the currPtr to the next entry in the list.  If 
		// * prevPtr is still NULL, then we are at the top of the list.
		// * If prevPtr is Non-NULL then currPtr should point to the 
		// * entry after it.
		// *************************************************************
		if(prevPtr) currPtr = prevPtr->getNext();
		else        currPtr = histogram;
		}
	}

// *****************************************************************************
// * void RangeTable::asciiDump :
// *	This method generates a diagnostic dump of the RangeTable object.  Since
// *	this class is never seen outside of the MatchMatrix object, this method
// *	is provided exclusively for internal debugging and testing.
// *****************************************************************************
void RangeTable::asciiDump ( FILE * fp, char * prefix )
	{
	char c = 0;
	char * prefixString = prefix?prefix:&c;
	char * subPrefixString = new char[strlen(prefixString)+8];
	sprintf(subPrefixString, "%s      ", prefixString);
	
	fprintf(fp, "%sRangeTable for job %s :\n", prefixString, job?job->getName():"None");
	fprintf(fp, "%s   => Number of Types  : %i\n", prefixString, nTypes);
	fprintf(fp, "%s   => Types            :\n", prefixString );
	
	for(int idx=0; idx<nTypes; idx++)
		types[idx]->asciiDump(fp, subPrefixString);

	delete subPrefixString;
	}
	
