// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
//	SchedStatic.cc contains the realization of the static variables and the
// 	methods defined for the SchedStatic class.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <sys/time.h>
#include <SchedStatic.h>

// *****************************************************************************
// * SchedStatic::now :
// *	This is the current time at the beginning of the scheduling cycle.  
// *	This variable is used to avoid making costly system calls to obtain
// *	the current time throughout the scheduling cycle.
// *****************************************************************************
time_t SchedStatic::now = 0;

// *****************************************************************************
// * SchedStatic::timeMax :
// *	This is the maximum time value that will be considered by the scheduling
// *	system.
// *****************************************************************************
time_t SchedStatic::timeMax = LONG_MAX;

// *****************************************************************************
// * SchedStatic::durationMin :
// *	This is the smallest amount of free time (beginning imediately) that
// *	any processor can have and still be considered immediately available.
// *	By default this value is 300 seconds (5 minutes).
// *****************************************************************************
time_t SchedStatic::durationMin = 300;

// *****************************************************************************
// * SchedStatic::freeSlotMin :
// *	This is the smallest period of free time that will be allowed to 
// *	represent available time on a processor.  If the period of free
// *	time is less than this - then it will be ignored. By default this
// *	value is 60 seconds.
// *****************************************************************************
time_t SchedStatic::freeSlotMin = 60;

// *****************************************************************************
// * time_t freeSetMin :
// *	This is the shortest duration of time for which a FreeTimeSet
// *	object wil be created.  Note the FreeTimeSet object represents
// *	a period of availability on multiple processors of a compute
// *	node.
// *****************************************************************************
time_t SchedStatic::freeSetMin = 60;

// *********************************************************************
// * time_t reservationMax :
// *	This is the longest contiguous period of time that can be
// *	requested for a reservation. By default it is one year.
// *********************************************************************
time_t SchedStatic::reservationMax = 3600*365*24;
	
// *****************************************************************************
// * time_t defaultWallTime :
// *	This value is used as the walltime for a job in the event that neither
// *	the user nor the system provided a wall time resource value.  This 
// *	value is set to 60 hours.
// *****************************************************************************
time_t SchedStatic::defaultWallTime = 60*60*60;

// *****************************************************************************
// * int balanceFlag :
// *	This is a boolean value that tells the system to use 
// *	utilization information to attempt to balance the load across 
// *	all systems.
// *
// *	If TRUE, the system will attempt to populate the least loaded
// *	nodes FIRST in order to balance the load.  If FALSE, the system
// *	will attempt to fully populate one node prior to moving on
// *	to the next.
// *****************************************************************************
int SchedStatic::balanceFlag = 0;

// *****************************************************************************
// * SchedStatic::startCycle :
// * 	This method is called at the bgeinning of the scheduling cycle to 
// *	initialize internal variables.  This method will obtain the current 
// *	system time and will retain it in the 'now' variable through out the
// *	scheduling cycle.  Any other object that desires to know the current 
// *	time should request it from this object.
// *****************************************************************************
void SchedStatic::startCycle ( void )
	{
	struct timeval t;
	gettimeofday(&t, NULL);
	now = t.tv_sec;
	}
	
// *****************************************************************************
// * SchedStatic::endCycle :
// *	This method is called at the end of the cycle to clear any scheduling
// *	cycle specific variables that should be regenerated at the beginning
// *	of the next scheduling cycle.
// *****************************************************************************
void SchedStatic::endCycle ( void )
	{
	now = 0;
	}
	
// *****************************************************************************
// * SchedStatic::getTimeString :
// *	This method is called to convert a time_t object to its string 
// *	representation.
// *****************************************************************************
const char * SchedStatic::getTimeString ( time_t time )
	{
	static char timeBuf[64];
	
	if(time==timeMax) strcpy(timeBuf, "Infinity");
	else if(time)
		{
		struct tm * tptr = localtime(&time);
		if(tptr) strftime(timeBuf, 64, "%m/%d/%Y %H:%M:%S", tptr);
		else strcpy(timeBuf, "Indefinite");
		}
	else strcpy(timeBuf, "Unspecified");
	
	return (const char *)timeBuf;
	}
	
// *****************************************************************************
// * SchedStatic::stringToDuration :
// * 	This method converts a time string to a time_t. The defaultValue 
// *	provided by the caller will be used in the evnt that the str is NULL
// *	or blank.
// *****************************************************************************
time_t SchedStatic::stringToDuration ( const char * str, time_t defaultValue )
	{
	time_t result=defaultValue;
	
	if(str && *str)
		{
		int multx = 1;
		for(const char *ptr=str; multx<3600 && (ptr=strchr(ptr, ':'))!=NULL; ptr++)
			{
			multx*=60;
			}

		result    = 0;
		while(*str && multx)
			{
			result += atoi(str)*multx;
			if(*(str+=strcspn(str, ":"))) str++;
			multx/=60;
			}
		}
	
	return result;
	}
