// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	The SchedStatic.h file contains a list of static functions and methods
// 	that are used throughout the scheduler.  This object will be consulted
// 	to obtain the value of now (the time at the beginning of the cycle) and
// 	the maximum time that the scheduler will consider.
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#ifndef _SCHED_STATIC_H_
#define _SCHED_STATIC_H_ 1

#include <time.h>		// Needed for time_t

// *****************************************************************************
// * class SchedStatic :
// *	This class provides an interface to global data and functions that are
// *	common tomany of the scheduler objects.  Centralizing these objects
// *	here allows the scheduler to maintain a notion of 'now' that reflects
// *	the beginning of the scheduling cycle - rather than the current time
// *	during the cycle.
// *
// *	By using a relative notion of time, many of the system level calls
// *	may be elliminated and overall system performance will be improved.
// *****************************************************************************
class SchedStatic 
{
private:
	// *********************************************************************
	// * time_t now:
	// *	This value represents the system time at the beginning of the 
	// *	scheduling cycle.  The method startCycle must be called at the 
	// *	beginning of teh scheduling cycle to set the value of now.  At
	// *	the end of the scheduling cycle, the endCycle method is called
	// *	to clear the value of this variable.  
	// *
	// *	The 'now' value is used to avoid having to make multiple calls
	// *	out to the system to obtain the system time during a scheduling
	// *	cycle.
	// *********************************************************************
	static time_t now;

	// *********************************************************************
	// * time_t timeMax :
	// *	This is the largest time_t value that will be allowed within the
	// *	system.  This value equates to a time of infinity.
	// *********************************************************************
	static time_t timeMax;

	// *********************************************************************
	// * time_t durationMin :
	// *	This variable is used to gauge the availability of a processor.
	// *	The processor is only considered to be immediately available if
	// *	1) There are no jobs running on it now 
	// *	-AND-
	// *	2) No job is scheduled to start on it for 'durationMin' seconds.
	// *********************************************************************
	static time_t durationMin;

	// *********************************************************************
	// * time_t freeSlotMin :
	// *	This is the shortest duration of time for which a FreeTimeSlot
	// *	object will be created.  Note the FreeTimeSlot object represents
	// *	a period of availability on an individual processor of a compute
	// *	node.
	// *********************************************************************
	static time_t freeSlotMin;

	// *********************************************************************
	// * time_t freeSetMin :
	// *	This is the shortest duration of time for which a FreeTimeSet
	// *	object wil be created.  Note the FreeTimeSet object represents
	// *	a period of availability on multiple processors of a compute
	// *	node.
	// *********************************************************************
	static time_t freeSetMin;
	
	// *********************************************************************
	// * time_t reservationMax :
	// *	This is the longest contiguous period of time that can be
	// *	requested for a reservation. By default it is one year.
	// *********************************************************************
	static time_t reservationMax;
	
	// *********************************************************************
	// * time_t defaultWallTime :
	// *	This value is used in the event that there is no walltime
	// *	specified for a job.
	// *********************************************************************
	static time_t defaultWallTime;
	
	// *********************************************************************
	// * int balanceFlag :
	// *	This is a boolean value that tells the system to use 
	// *	utilization information to attempt to balance the load across 
	// *	all systems.
	// *
	// *	If TRUE, the system will attempt to populate the least loaded
	// *	nodes FIRST in order to balance the load.  If FALSE, the system
	// *	will attempt to fully populate one node prior to moving on
	// *	to the next.
	// *********************************************************************
	static int balanceFlag;
	
public:
	// *********************************************************************
	// * enum MAXSUBSETS :
	// *	This value is used to limit the maximum number of unique
	// *	JobCandidateSubSet requirements that can be supported by the
	// *	RangeTable.  
	// *********************************************************************
	enum { MAXSUBSETS = 32 };

	// *********************************************************************
	// * enum MAXPROC :
	// *	This is the maximum number of processors that is supported on 
	// *	any compute node. The value of this variable should be declared
	// *	during the build using the MAX_PROCESSOR_CNT definition.
	// *********************************************************************
	enum {MAXPROC = 16};
	
	// *********************************************************************
	// * enum MAXNODES :
	// *	This is the maximum number of compute nodes that can exist in
	// *	the system.  
	// *********************************************************************
	enum {MAXNODES = 256};

	// *********************************************************************
	// * The startCycle method is called at the beginning of each scheduling 
	// * cycle to set internal objects to the appropriae values.
	// *********************************************************************
	static void startCycle ( void );
	
	// *********************************************************************
	// * The endCycle method is called at the end of the scheduling cycle 
	// * to reset internal values to their defaults.
	// *********************************************************************
	static void endCycle ( void );
	
	// *********************************************************************
	// * These are the accessor methods that allow the internal
	// * variables to be retrieved.
	// *********************************************************************
	static time_t getNow             ( void )  { if(!now) startCycle(); return now; }
	static time_t getTimeMax         ( void )  { return timeMax; }
	static time_t getDurationMin     ( void )  { return durationMin; }
	static time_t getFreeSlotMin     ( void )  { return freeSlotMin; }
	static time_t getFreeSetMin      ( void )  { return freeSetMin; }
	static time_t getReservationMax  ( void )  { return reservationMax; }
	static time_t getDefaultWallTime ( void )  { return defaultWallTime; }
	static int    getBalanceFlag     ( void )  { return balanceFlag; }
	static void   setBalanceFlag     ( int x ) { balanceFlag = x; }
	
	// *********************************************************************
	// * These are utility methods used to perform common 
	// * computations used by other objects.
	// *********************************************************************
	static const char * getTimeString    ( time_t time );
	static time_t       stringToDuration ( const char *, time_t );
};

#endif /* _SCHED_STATIC_H_ */
