// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
//	The SimpleProcessor.h file contains the class definition for the 
// 	SimpleProcessor class.  It contains an array of JobAssigmnment objects
//	and an array of FreeTimeSlot objects.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#ifndef _SIMPLE_PROCESSOR_H_
#define _SIMPLE_PROCESSOR_H_ 1

#include <SchedStatic.h>	// Needed for static SchedStatic values
#include <FreeTimeSlot.h>	// Needed for inline FreeTimeSlot methods

// *****************************************************************************
// * class SimpleProcessor :
// *	This object is represents a single processor on a node.  It will contain
// *	a list of jobs that have been assigned to the node (JobAssignment) as 
// *	well as a list of unscheduled time periods that are still available for
// *	consumption ( FreeTimeSlot ).
// *****************************************************************************
class SimpleProcessor 
{
private :
	char                * name; 
	class JobAssignment * jobs;
	class FreeTimeSlot  * slots;
	
public:
	// *********************************************************************
	// * These are the contructor and destructor for the objects.
	// *********************************************************************
	SimpleProcessor  ( char * node, size_t index );
	~SimpleProcessor ( void );
	
	// *********************************************************************
	// * These are the inline accessor methods for the objects.
	// *********************************************************************
	const char    * getName  ( void ) { return (const char *) name; }
	JobAssignment * getJobs  ( void ) { return jobs; }
	FreeTimeSlot  * getSlots ( void ) { return slots; }
	
	// *********************************************************************
	// * SimpleProcessor::getSlot :
	// *	This method will locate a free time slot that encompasses the
	// *	start time and duration specified as parameters.  If the 
	// *	time is not represented, then NULL will be returned.
	// *********************************************************************
	FreeTimeSlot  * getSlot  ( time_t start, time_t duration );
	
	// *********************************************************************
	// * The available method is used to assess if the processor remains
	// * immediately available.
	// *********************************************************************
	inline int available ( void );

	// *********************************************************************
	// * SimpleProcessor::availableTime :
	// *	This method will return the length of time that this processor
	// *	is immediately available.
	// *********************************************************************
	inline time_t availableTime ( void );

	// *********************************************************************
	// * The claim method is used to install a new job on the processor. 
	// *********************************************************************
	char * claim ( char * job, time_t start, time_t duration );
	
	// *********************************************************************
	// * SimpleProcessor::scheduled :
	// *	This method returns a boolean flag indicating whether the
	// *	processor has been scheduled at a specific time.
	// *********************************************************************
	int scheduled ( time_t startTime );
	
	// *********************************************************************
	// * SimpleProcessor::reserveTime :
	// *	This method is called when the system needs to reserve a block
	// *	of time to prevent it from being utilized by a job.  This method
	// *	will walk through the list of FreeTimeSlot objects and will 
	// *	modify any FreeTimeSlot that intersects with the specified
	// *	period.
	// *********************************************************************
	void reserveTime ( time_t start, time_t duration );

	// *********************************************************************
	// * This method generates an ascii report of the contents of the object
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout, char * prefix = NULL );
};

// *****************************************************************************
// * SimpleProcessor::available :
// *	This method returns a boolean result indicating if the processor is
// *	immediately available for utilization.  A processor is considered
// *	available if it has a free time slot that begins immediately AND the
// *	duration of that free time slot is greater than SchedStatic::minDuration.
// *****************************************************************************
inline int SimpleProcessor::available ( void )
	{
	time_t slotStart    = (slots?slots->getStartTime():0);
	time_t slotLen      = (slots?slots->getDuration():0);

	return (slotStart<=SchedStatic::getNow() &&
	   	slotLen  > SchedStatic::getDurationMin())?1:0;
	}

// *****************************************************************************
// * SimpleProcessor::availableTime :
// *	This method will return the length of time that this processor
// *	is immediately available.
// *****************************************************************************
inline time_t SimpleProcessor::availableTime ( void )
	{
	time_t slotStart    = (slots?slots->getStartTime():0);
	time_t slotLen      = (slots?slots->getDuration():0);

	return (slotStart<=SchedStatic::getNow())?slotLen:0;
	}
#endif /* _SIMPLE_PROCESSOR_H_ */
