// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
//	This class computes the weight of a job based on the job's duration.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#include <DurationStage.h>
#include <SchedStatic.h>
#include <math.h>

// *****************************************************************************
// * DurationStage::compute :
// *	This stage of the scheduling algorithm will compute the weight of
// *	an individual job based on the amount of walltime that it has requested.
// *	
// *	jobs      : The array of PBS_Jobs that are being weighted.
// *	jobCnt    : The number of PBS_Jobs in the array.
// *	weightSum : The sum of the weights of all stages that will be used. 
// *****************************************************************************
void DurationStage::compute ( PBS_Job ** jobs, size_t jobCnt, double weightSum )
	{
	// *********************************************************************
	// * First requested the upper and lower bounds of significance from
	// * the configuration file.  When each value is tabulated and fitted
	// * to a curve, 90% of the curve is represented by those values that
	// * occur between the lower significance boundary and the upper
	// * significance boundary.
	// *
	// * Jobs whose durations are shorter than the lower significance 
	// * boundary will be fit in the bottom 5% of the curve, and those
	// * whose durations is greater than the upper significance boundary
	// * will be confined to the top 5% of the curve.
	// *
	// * The time expressed in the configuration file is given as whole
	// * seconds.  By default the lower significance boundary is 1 minute 
	// * and the upper significance boundary is 24 hours.
	// *********************************************************************
	PBS_Parameters & params = PBS_Parameters::attach();
	double LSB;
	double USB;
	
	LSB = SchedStatic::stringToDuration(
		params.getConfigString(
			"Job Duration Stage", 
			"lowerBound", 
			NULL), 60);
	USB = SchedStatic::stringToDuration(
		params.getConfigString(
			"Job Duration Stage", 
			"upperBound", 
			NULL), 24*3600);

	// *********************************************************************
	// * In order to insure a decent curve shape and proper behaviour, 
	// * The USB and LSB value MUST BE POSITIVE, and the LSB must be less 
	// * than the  USB.
	// *********************************************************************
	if(USB < 0) USB = 24*3600;
	if(LSB<0 || LSB>USB) LSB = USB/10.0;

	// *********************************************************************
	// * Next, request the countProcs flag from the configuration file. It
	// * is a yes or no value that indicate whether the number of processors
	// * should be considered when calculating the overall job duration.
	// *********************************************************************
	int countProcs = !strcasecmp
		("yes", params.getConfigString
			("Job Duration Stage", "countProcs", "yes"));

	// *********************************************************************
	// * Next we must obtain the weight of this stage (relative to all other
	// * stages that will be used. By default all stages are given a weight
	// * of 1.
	// *********************************************************************
	double weight = getWeight();
	
	// *********************************************************************
	// * If the stage weight is less than or equal to 0, then the stage will
	// * have no effect on the outcome and we will return without computing
	// * it.
	// *********************************************************************
	if(weight <= 0 || weightSum <= 0 ) return;
	
	// *********************************************************************
	// * If the weight is greater than 0, then we will compute the lower
	// * weight boundary using this value and the weightSum value.  The
	// * LWB indicates the bottom of the curve for this stage... Note: all
	// * stages will fit their values to a curve between LWB and 1.0.
	// *********************************************************************
	double LWB = 1.0-(weight/weightSum);
	
	// *********************************************************************
	// * If a stage recorder was provided with this class then 
	// * record debugging output for later review.
	// *********************************************************************
	if(recorder)
		{
		recorder->stagePrintf("[Duration Stage]", "Lower Significance Boundary = %i:%02i:%02i", (int)LSB/3600, ((int)LSB%3600)/60, ((int)LSB%60));
		recorder->stagePrintf("[Duration Stage]", "Upper Significance Boundary = %i:%02i:%02i", (int)USB/3600, ((int)USB%3600)/60, ((int)USB%60));
		recorder->stagePrintf("[Duration Stage]", "Count Processors Flag       = %s", countProcs?"yes":"no");
		recorder->stagePrintf("[Duration Stage]", "Stage Weight                = %g", weight);
		recorder->stagePrintf("[Duration Stage]", "Sum of All Stage Weights    = %g", weightSum);
		}
		
	// *********************************************************************
	// * Now we'll walk through each of the jobs and apply the weight that
	// * is appropriate for its job duration.
	// *********************************************************************
	for(size_t idx=0; idx<jobCnt; idx++)
		{
		double rawWeight = 0;
		
		// *************************************************************
		// * Read the job's duration from the job.
		// *************************************************************
		double duration = jobs[idx]->getDuration();
		
		// *************************************************************
		// * If countProcs is TRUE, then multiply the duration times
		// * the number of processors requested.
		// *************************************************************
		if(countProcs) duration*=jobs[idx]->getProcCnt();
		
		// *************************************************************
		// * If the duration is between the lower significance boundary 
		// * and the upper significance boundary, then we'll use the
		// * 'Central Formula' to compute its weight.
		// *************************************************************
		if(LSB<=duration && duration<=USB) 
			{
			rawWeight = -(18.0*duration - 9.0*(USB+LSB)) / (10.0*(USB-LSB));
			}
		// *************************************************************
		// * Otherwise use the boundary formula. Note the variable t is
		// * used to reduce the compute time for the operation and to 
		// * simplify the appearance of the code.
		// *************************************************************
		else	{
			double t = (2.0*duration - (USB+LSB));
			rawWeight = -(t / (fabs(t)+(USB-LSB)/9));
			}

		// *************************************************************
		// * Next use ths raw weight in conjunction with the lower
		// * weight boundary to fit the curve within the desired region
		// * and obtain the processed weight of this job in this stage.
		// *************************************************************
		double procWeight = ((rawWeight+1.0)/2.0)*(1-LWB)+LWB;
		
		// *************************************************************
		// * Multiply this value against the current weight of the 
		// * job to introduce this stage's influence on the overall 
		// * weight of the job. Note that if the initial weight is less 
		// * than (or equal to) 0, it is restored to 1 - making this 
		// * stage the only one of any significance.
		// *************************************************************
		double jobWeight;

		if((jobWeight = jobs[idx]->getWeight())<=0) jobWeight=1.0;
		jobs[idx]->setWeight(jobWeight * procWeight);

		// *************************************************************
		// * If a stage recorder was provided with this class then 
		// * record debugging output for later review.
		// *************************************************************
		if(recorder)
			{
			recorder->jobPrintf(jobs[idx]->getName(), 
				"[Duration Stage] %g hours on %3ld nodes yields %g", 
				((double)jobs[idx]->getDuration())/3600.0,
				jobs[idx]->getProcCnt(),
				procWeight );
			}
		}
	}
