// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	This file realizes the class methods that are defined in the 
//	QueuePriorityStage.h header file.
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#include <QueuePriorityStage.h>
#include <PBS_Server.h>
#include <PBS_Accounting.h>
#include <SchedStatic.h>

// *****************************************************************************
// * compareFunc :
// *	This is a static comparison method that is used to reorder the jobs
// *	in ascending order by the queueID.  This will reduce the effort expended
// *	in descending through the list multiple times. This function is static
// *	to prevent it from appearing outside of this object module and to keep
// *	the global name space clean.
// *****************************************************************************
static int compareFunc (const void * a, const void * b)
	{
	PBS_Job * jobA = *(PBS_Job **)a;
	PBS_Job * jobB = *(PBS_Job **)b;
	return jobA->getQueueID() - jobB->getQueueID();
	}


// *****************************************************************************
// * QueuePriorityStage::compute :
// *	This method is called to compute the weight associated with a job based 
// *	on the queues historical system utilization and its priority.  This 
// *	method will tend to balance system usage to correspond to the priority
// *	associated with the originating queue.
// *
// *	Note: The externally owned PBS_AccountingDatabase object should be updated 
// *	prior to calling this method.
// *****************************************************************************
void QueuePriorityStage::compute ( PBS_Job ** jobs, size_t jobCnt, double weightSum )
	{
	// *********************************************************************
	// * The QueuePriority class is an internal implementation object that
	// * is used to accumulate information about a queues historical 
	// * utilization.
	// *********************************************************************
	class QueuePriority
	{
	public:
		typedef enum { OLD=0, RECENT=1 } JobStat;
		
		QueuePriority * next;
		const char *    name;
		int             queueID;
		double          priority;
		double          actual;
		double          expected;
		double          projected;
		double          cpu[2];
		double          wall[2];
		double          jobs[2];

	QueuePriority ( void )
		: next(NULL),  name(NULL),    queueID(0),     priority(0.0), 
		  actual(0.0), expected(0.0), projected(0.0)
		{
		cpu[OLD]  = (cpu[RECENT] = 0.0);
		wall[OLD] = (wall[RECENT] = 0.0);
		jobs[OLD] = (jobs[RECENT] = 0.0);
		}
	};
	
	QueuePriority   total;
	QueuePriority * head         = NULL;
	QueuePriority * curr         = NULL;
	double          weight       = getWeight();
	double          maxProjected = 0.0;
	double          minPriority  = 1.0;
	size_t          queueCnt     = 0;
	size_t          idx          = 0;
				
	// *********************************************************************
	// * Start by getting the weight from the scheduler profile. If the 
	// * weight is 0 or the weightSum is less than or equal to 0, then this
	// * scheduling stage will be skipped. The stage will also be skipped
	// * if the jobCnt is 0 or the jobs array is NULL.
	// *********************************************************************
	if(weight<=0.0 || weightSum<=0 || jobCnt==0 || jobs==NULL) return;

	// *********************************************************************
	// * Before performing any operation, walk through the list of jobs
	// * and ensure that they represent at least two queues... If not,
	// * then there is no need to perform this operation.
	// *********************************************************************
	for(idx=1; idx<jobCnt && jobs[idx-1]->getQueueID()!=jobs[idx]->getQueueID(); idx++);
	if(idx>=jobCnt) return;

	// *********************************************************************
	// * If the weight is greater than 0, then we will compute the lower
	// * weight boundary using this value and the weightSum value.  The
	// * LWB indicates the bottom of the curve for this stage... Note: all
	// * stages will fit their values to a curve between LWB and 1.0.
	// *********************************************************************
	double LWB = 1.0-(weight/weightSum);

	// *********************************************************************
	// * Next, get the default share from the scheduler profile, as well as,
	// * the other factors that will be used to compute the share.
	// *********************************************************************
	PBS_Parameters & params = PBS_Parameters::attach();
	double wallFactor   = params.getConfigFloat("Queue Priority Stage", "wallTimeFactor", 1.0);
	double cpuFactor    = params.getConfigFloat("Queue Priority Stage", "cpuTimeFactor", 1.0);
	double runFactor    = params.getConfigFloat("Queue Priority Stage", "runrunFactor", 1.0);
	double oldFactor    = params.getConfigFloat("Queue Priority Stage", "oldFactor", 0.10);

	// *********************************************************************
	// * Ensure that all of the values that werew read are within legal
	// * range.
	// *********************************************************************
	if(wallFactor<0) wallFactor = 0.0;
	if(cpuFactor<0)  cpuFactor  = 0.0;
	if(runFactor<0)  runFactor  = 0.0;
	if(oldFactor<0)  oldFactor  = 0.0;

	// *********************************************************************
	// * Retrieve the thresholds for recent and old jobs.
	// *********************************************************************
	time_t rThreshold   = SchedStatic::stringToDuration (
		params.getConfigString (
			"Queue Priority Stage", 
			"recentThreshold", 
			NULL), 24*3600);
	time_t oThreshold   = SchedStatic::stringToDuration (
		params.getConfigString (
			"Queue Priority Stage", 
			"oldJobThreshold", 
			NULL), 168*3600);
		
	// *********************************************************************
	// * If a Stage Debug Recorder was provided to this object, then
	// * use it to record the events and data associated with this
	// * scheduling stage.
	// *********************************************************************
	if(recorder)
		{
		recorder->stagePrintf("[Queue Priority Stage]", "Wall Factor      : %g", wallFactor);
		recorder->stagePrintf("[Queue Priority Stage]", "CPU  Factor      : %g", cpuFactor);
		recorder->stagePrintf("[Queue Priority Stage]", "Run  Factor      : %g", runFactor);
		recorder->stagePrintf("[Queue Priority Stage]", "Old  Factor      : %g", oldFactor);
		recorder->stagePrintf("[Queue Priority Stage]", "Recent Threshold : %i:%02i:%02i hours", 
			rThreshold/3600, (rThreshold%3600)/60, rThreshold%60);
		recorder->stagePrintf("[Queue Priority Stage]", "Old Threshold    : %i:%02i:%02i hours", 
			oThreshold/3600, (oThreshold%3600)/60, oThreshold%60);
		recorder->stagePrintf("[Queue Priority Stage]", "Stage Weight     : %g", weight);
		recorder->stagePrintf("[Queue Priority Stage]", "Sum of Weights   : %g", weightSum);				
		}

	// *********************************************************************
	// * Presort the job list by queue ID to minimize the number of times
	// * that the data structure will need to be descended.
	// *********************************************************************
	qsort (jobs, jobCnt, sizeof(PBS_Job *), compareFunc);
	
	// *********************************************************************
	// * The next step is to walk through the jobs and identify each queue
	// * who is represented.  We'll create a QueuePriority object for each 
	// * of these queues. This structure defines its priority and other 
	// * pertinent information.
	// *********************************************************************
	idx = 0;
	do 	{
		// *************************************************************
		// * Extract a PBS_JobArray that defines the historical 
		// * utilization of this queue. Then walk through the list of 
		// * historical jobs and update the QueuePriority object to 
		// * reflect the queue's historical utilization information.
		// *************************************************************
		int                    queueID = jobs[idx]->getQueueID();
		const char           * name    = jobs[idx]->getQueueName();
		PBS_Queue            * queue   = server.getQueue((char *)name);
		PBS_Accounting       * qJobs   = serverDB.findQueue(queue->getName());
		PBS_AccountingRecord * qRec    = NULL;
		time_t                 now     = SchedStatic::getNow();
		
		// *************************************************************
		// * If a valid queue was found - continue processing.
		// *************************************************************
		if (queue)
			{
			// *****************************************************
			// * Increment the queueCnt variable.
			// *****************************************************
			queueCnt++;
			
			// *****************************************************
			// * Create a new QueuePriority object that represents 
			// * this user's historical utilization and share of 
			// * system resources.
			// *****************************************************
			if(curr)	
				{
				curr->next = new QueuePriority();
				curr = curr->next;
				}
			else	{
				head = new QueuePriority();
				curr = head;	
				}

			// *****************************************************
			// * Install the queueID and the queue's priority into 
			// * the QueuePriority object.
			// *****************************************************
			curr->queueID  = queueID;
			curr->name     = name;
			curr->priority = queue->getPriority(); 
			total.priority += curr->priority;
			
			// *****************************************************
			// * Track the minimum priority during processing. This
			// * will be needed if any of the queues were given a
			// * negative priority.
			// *****************************************************
			if(curr->priority < minPriority) 
				minPriority = curr->priority;
			 
			// *****************************************************
			// * Walk through each record returned from the 
			// * PBS_AccountingDatabase and add the utilization 
			// * values to the QueuePriority record.
			// *****************************************************
			for(size_t qIdx=0; qJobs && (qRec=(*qJobs)[qIdx])!=NULL; qIdx++)
				{
				// *********************************************
				// * If the job has been started... then add the 
				// * statistics.
				// *********************************************
				if(qRec->getStartTime()!=0)
					{
					time_t endTime = qRec->getEndTime();
					double cpu     = (double)qRec->getCpuTime()/3600.0;
					double wall    = (double)qRec->getWallTime()/3600.0;
					double job     = 1.0;

					if(endTime==0 || now-endTime <= rThreshold)
						{
						total.cpu[QueuePriority::RECENT]+=cpu;
						total.wall[QueuePriority::RECENT]+=wall;
						total.jobs[QueuePriority::RECENT]+=job;
						curr->cpu[QueuePriority::RECENT]+=cpu;
						curr->wall[QueuePriority::RECENT]+=wall;
						curr->jobs[QueuePriority::RECENT]+=job;
						}
					else if(now-endTime <= oThreshold)
						{
						total.cpu[QueuePriority::OLD]+=cpu;
						total.wall[QueuePriority::OLD]+=wall;
						total.jobs[QueuePriority::OLD]+=job;
						curr->cpu[QueuePriority::OLD]+=cpu;
						curr->wall[QueuePriority::OLD]+=wall;
						curr->jobs[QueuePriority::OLD]+=job;
						}
					}
				}
			// *****************************************************
			// * Using the computed statistics, calculate this
			// * queue's combined utilization statistics.
			// *****************************************************
			curr->actual =	(curr->cpu [QueuePriority::RECENT]+
					(curr->cpu [QueuePriority::OLD]*oldFactor))*
					cpuFactor+
					(curr->wall[QueuePriority::RECENT]+
					(curr->wall[QueuePriority::OLD]*oldFactor))*
					wallFactor+
					(curr->jobs[QueuePriority::RECENT]+
					(curr->jobs[QueuePriority::OLD]*oldFactor))*
					runFactor+1;
			total.actual += curr->actual;
			}
	
		// *************************************************************
		// * Walk through each of the jobs that is associated with this 
		// * particular queue.
		// *************************************************************
		while(idx<jobCnt && jobs[idx]->getQueueID()==queueID) idx++;
		} while(idx<jobCnt);

	// *********************************************************************
	// * If the minimum priority is less than 1, then advance all of the
	// * priorities AND the totals sufficiently to make all values greater
	// * than 0.
	// *********************************************************************
	if(minPriority<1) total.priority -= (minPriority-1)*queueCnt;

	// *********************************************************************
	// * Using the existing statistics and totals, walk through the list of
	// * queues and compute what the expected utilization should have been.
	// * Using this value in confjunction with the actual utilization, 
	// * compute what the projected utilization should be.
	// *
	// * Note: If fewer than two queues were detected - then these 
	// * computations will be skipped because it will have no effect on the
	// * outcome of processing.
	// *********************************************************************
	for(curr = head; queueCnt>1 && curr!=NULL; curr=curr->next)
		{
		if(minPriority<1) curr->priority -= (minPriority-1);
		// *************************************************************
		// * Compute the expected utilization based on actual use and
		// * percentage of overall priorities.
		// *************************************************************
		curr->expected = total.actual * (curr->priority/total.priority);

		// *************************************************************
		// * If the queueu had actual utilization, then project its
		// * future utilization as a proportion of its expected to its
		// * actual. If it had no utilization, then set its projected
		// * utilization equal to its priority.
		// *************************************************************
		if(curr->actual > 0.0) curr->projected = curr->expected/curr->actual;
		else curr->projected = curr->priority;

		// *************************************************************
		// * Keep track of the maximum projected utilization - this 
		// * value will be used to scale all values into the 0 - 1 range
		// * prior to computing weights.
		// *************************************************************
		if(curr->projected > total.projected) 
			total.projected = curr->projected;
		}
	
	// *********************************************************************
	// * Now we will walk through the queue list again, and compute the
	// * actual weight associated with the queue and apply this weight to
	// * each job that is associated with that queue.
	// *
	// * Note: If fewer than two queues were detected - then these 
	// * computations will be skipped because it will have no effect on the
	// * outcome of processing.
	// *********************************************************************
	for(curr=head, idx=0; queueCnt>1 && curr!=NULL && idx<jobCnt; curr=curr->next)
		{
		double queueWeight = (curr->projected/total.projected)*(1-LWB)+LWB;

		// *****************************************************
		// * If a Stage Debugging Recorder object was provided,
		// * then use it to report the results.
		// *****************************************************
		if(recorder)
			{
			recorder->stagePrintf("[Queue Priority Stage]",
				"Queue %s - Priority       : %g", curr->name, curr->priority);
			recorder->stagePrintf("[Queue Priority Stage]", 
				"Queue %s - Wall Time      : %g old hours, %g recent hours", 
				curr->name, 
				curr->wall[QueuePriority::OLD],
				curr->wall[QueuePriority::RECENT]);
			recorder->stagePrintf("[Queue Priority Stage]", 
				"Queue %s - CPU Time       : %g old hours, %g recent hours", 
				curr->name, 
				curr->cpu[QueuePriority::OLD],
				curr->cpu[QueuePriority::RECENT]);
			recorder->stagePrintf("[Queue Priority Stage]", 
				"Queue %s - Job Count      : %g old jobs, %g recent jobs", 
				curr->name, 
				curr->jobs[QueuePriority::OLD],
				curr->jobs[QueuePriority::RECENT]);
			recorder->stagePrintf("[Queue Priority Stage]",
				"Queue %s - Applied Weight : %g", curr->name, queueWeight);
			}								

		// *************************************************************
		// * Apply this weight to each of the queues's jobs in the array.
		// *************************************************************
		while(idx<jobCnt && jobs[idx]->getQueueID()==curr->queueID)
			{
			double oldWeight = jobs[idx]->getWeight();
			if(oldWeight<0.0 || oldWeight>1.0) oldWeight = 1.0;
			jobs[idx]->setWeight(oldWeight*queueWeight);

			// *****************************************************
			// * If a Stage Debugging Recorder object was provided,
			// * then use it to report the results.
			// *****************************************************
			if(recorder)
				{
				recorder->jobPrintf(jobs[idx]->getName(),
					"[Queue Priority Stage] Queue %s priority of %g - weight of %g applied", 
					curr->name, curr->priority, queueWeight);
				}

			idx++;
			}
		}

	// *********************************************************************
	// * Delete all of the temporary objects that were created for this
	// * method.
	// *********************************************************************
	while(head!=NULL)
		{
		curr = head;
		head = curr->next;
		delete curr;
		}
	}
