// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	This header file defines the QueuePriorityStage class which computes the
//	weight for a specific queue based on its historical system utilization.
//
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#ifndef _QUEUE_PRIORITY_STAGE_H_
#define _QUEUE_PRIORITY_STAGE_H_ 1

#include <UnderLordStage.h>
#include <PBS_Parameters.h>
#include <PBS_Job.h>
#include <PBS_Accounting.h>

// *****************************************************************************
// * class QueuePriorityStage:
// *	This class will compute the weight for a specific job based on the 
// *	historical system utilization by the owner.  Unlike the other stages,
// *	this stage must presort the jobs because the weight assigned for each
// *	user decreases as a job is scheduled.
// *****************************************************************************
class QueuePriorityStage : public UnderLordStage
{
protected:
	PBS_Server             & server;
	PBS_AccountingDatabase & serverDB;
	
public:
	inline QueuePriorityStage   ( PBS_Server & Server, PBS_AccountingDatabase & ServerDB );
	inline double getWeight     ( void );
	void          compute       ( PBS_Job ** jobs, size_t jobCnt, double weightSum );
};

// *****************************************************************************
// *  QueuePriorityStage::QueuePriorityStage :
// *	This is the constructor for the QueuePriorityStage object. It will 
// *	access a PBS_AccountingDatabase object to track system utilization 
// *	information.
// *****************************************************************************
inline QueuePriorityStage::QueuePriorityStage ( PBS_Server & Server, PBS_AccountingDatabase & ServerDB )
	: UnderLordStage (),
	  server         (Server),	  
	  serverDB       (ServerDB)
	{
	}

// *****************************************************************************
// * QueuePriorityStage::getWeight :
// *	This method will read from the scheduler profile and will return the
// *	weight associated with this scheduling stage. A default weight of 1.0
// *	will be returned if the parameter has not been specified in teh 
// *	scheduler profile.
// *****************************************************************************
double QueuePriorityStage::getWeight ( void )
	{
	PBS_Parameters & params = PBS_Parameters::attach();
	double weight = params.getConfigFloat("Queue Priority Stage", "weight", 1.0);
	return (weight<=0.0?0.0:weight);
	}

#endif /* _QUEUE_PRIORITY_STAGE_H_ */
