// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	This file realizes the class methods that are defined in the 
//	UserShareStage.h header file.
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#include <UserShareStage.h>
#include <PBS_Accounting.h>
#include <SchedStatic.h>

// *****************************************************************************
// * compareFunc :
// *	This is a static comparison method that is used to reorder the jobs
// *	in ascending order by the ownerID.  This will reduce the effort expended
// *	in descending through the list multiple times. This function is static
// *	to prevent it from appearing outside of this object module and to keep
// *	the global name space clean.
// *****************************************************************************
static int compareFunc (const void * a, const void * b)
	{
	PBS_Job * jobA = *(PBS_Job **)a;
	PBS_Job * jobB = *(PBS_Job **)b;
	return jobA->getOwnerID() - jobB->getOwnerID();
	}


// *****************************************************************************
// * UserShareStage::compute :
// *	This method is called to compute the weight associated with a job based 
// *	on the users historical system utilization and its share.  This 
// *	method will tend to balance system usage to correspond to the share
// *	associated with the originating user.
// *
// *	Note: The externally owned server database should be updated prior to
// *	calling this compute method.
// *****************************************************************************
void UserShareStage::compute ( PBS_Job ** jobs, size_t jobCnt, double weightSum )
	{
	// *********************************************************************
	// * The UserShare class is an internal implementation object that
	// * is used to accumulate information about a users historical 
	// * utilization.
	// *********************************************************************
	class UserShare
	{
	private:
		char       * name;
	
	public:
		typedef enum { OLD=0, RECENT=1 } JobStat;
		
		UserShare  * next;
		int          ownerID;
		double       share;
		double       actual;
		double       expected;
		double       projected;
		double       cpu[2];
		double       wall[2];
		double       jobs[2];

	UserShare ( void )
		: next(NULL),  name(NULL),    ownerID(0),     share(0.0), 
		  actual(0.0), expected(0.0), projected(0.0)
		{
		cpu[OLD]  = (cpu[RECENT]  = 0.0);
		wall[OLD] = (wall[RECENT] = 0.0);
		jobs[OLD] = (jobs[RECENT] = 0.0);
		}
	~UserShare ( void )
		{
		if(name) free(name);
		}

	void setName ( const char * Name ) 
		{ 
		if(name) free(name);
		name = (Name && *Name)?strdup(Name):NULL; 
		}
	const char * getName ( void ) 
		{
		return (const char *)name; 
		}
	};
	
	UserShare   total;
	UserShare * head         = NULL;
	UserShare * curr         = NULL;
	double      weight       = getWeight();
	double      maxProjected = 0.0;
	double      minShare  = 1.0;
	size_t      userCnt     = 0;
	size_t      idx          = 0;
				
	// *********************************************************************
	// * Start by getting the weight from the scheduler profile. If the 
	// * weight is 0 or the weightSum is less than or equal to 0, then this
	// * scheduling stage will be skipped. The stage will also be skipped
	// * if the jobCnt is 0 or the jobs array is NULL.
	// *********************************************************************
	if(weight<=0.0 || weightSum<=0 || jobCnt==0 || jobs==NULL) return;

	// *********************************************************************
	// * Before performing any operation, walk through the list of jobs
	// * and ensure that they represent at least two users... If not,
	// * then there is no need to perform this operation.
	// *********************************************************************
	for(idx=1; idx<jobCnt && jobs[idx-1]->getOwnerID()!=jobs[idx]->getOwnerID(); idx++);
	if(idx>=jobCnt) return;

	// *********************************************************************
	// * If the weight is greater than 0, then we will compute the lower
	// * weight boundary using this value and the weightSum value.  The
	// * LWB indicates the bottom of the curve for this stage... Note: all
	// * stages will fit their values to a curve between LWB and 1.0.
	// *********************************************************************
	double LWB = 1.0-(weight/weightSum);

	// *********************************************************************
	// * Next, get the default share from the scheduler profile, as well as,
	// * the other factors that will be used to compute the share.
	// *********************************************************************
	PBS_Parameters & params = PBS_Parameters::attach();
	double wallFactor = params.getConfigFloat("User Share Stage", "wallTimeFactor", 1.0);
	double cpuFactor  = params.getConfigFloat("User Share Stage", "cpuTimeFactor", 1.0);
	double runFactor  = params.getConfigFloat("User Share Stage", "runrunFactor", 1.0);
	double oldFactor  = params.getConfigFloat("User Share Stage", "oldFactor", 0.10);
	double defShare   = params.getConfigFloat("Fair Share", "default", 1.0);
	
	// *********************************************************************
	// * Ensure that all of the values that werew read are within legal
	// * range.
	// *********************************************************************
	if(wallFactor<0) wallFactor = 0.0;
	if(cpuFactor<0)  cpuFactor  = 0.0;
	if(runFactor<0)  runFactor  = 0.0;
	if(oldFactor<0)  oldFactor  = 0.0;
	if(defShare<=0)  defShare   = 1.0;
	
	// *********************************************************************
	// * Retrieve the thresholds for recent and old jobs.
	// *********************************************************************
	time_t rThreshold   = SchedStatic::stringToDuration (
		params.getConfigString (
			"User Share Stage", 
			"recentThreshold", 
			NULL), 24*3600);
	time_t oThreshold   = SchedStatic::stringToDuration (
		params.getConfigString (
			"User Share Stage", 
			"oldJobThreshold", 
			NULL), 168*3600);
		
	// *********************************************************************
	// * If a Stage Debug Recorder was provided to this object, then
	// * use it to record the events and data associated with this
	// * scheduling stage.
	// *********************************************************************
	if(recorder)
		{
		recorder->stagePrintf("[User Share Stage]", "Wall Factor      : %g", wallFactor);
		recorder->stagePrintf("[User Share Stage]", "CPU  Factor      : %g", cpuFactor);
		recorder->stagePrintf("[User Share Stage]", "Run  Factor      : %g", runFactor);
		recorder->stagePrintf("[User Share Stage]", "Old  Factor      : %g", oldFactor);
		recorder->stagePrintf("[User Share Stage]", "Recent Threshold : %i:%02i:%02i hours", 
			rThreshold/3600, (rThreshold%3600)/60, rThreshold%60);
		recorder->stagePrintf("[User Share Stage]", "Old Threshold    : %i:%02i:%02i hours", 
			oThreshold/3600, (oThreshold%3600)/60, oThreshold%60);
		recorder->stagePrintf("[User Share Stage]", "Stage Weight     : %g", weight);
		recorder->stagePrintf("[User Share Stage]", "Sum of Weights   : %g", weightSum);				
		recorder->stagePrintf("[User Share Stage]", "Default Share    : %g", defShare);
		}

	// *********************************************************************
	// * Presort the job list by user ID to minimize the number of times
	// * that the data structure will need to be descended.
	// *********************************************************************
	qsort (jobs, jobCnt, sizeof(PBS_Job *), compareFunc);
	
	// *********************************************************************
	// * The next step is to walk through the jobs and identify each user
	// * who is represented.  We'll create a UserShare object for each 
	// * of these users. This structure defines its share and other 
	// * pertinent information.
	// *********************************************************************
	idx = 0;
	do 	{
		// *************************************************************
		// * Extract a PBS_JobArray that defines the historical 
		// * utilization of this user. Then walk through the list of 
		// * historical jobs and update the UserShare object to 
		// * reflect the user's historical utilization information.
		// *************************************************************
		int                    ownerID = jobs[idx]->getOwnerID();
		PBS_Accounting       * uJobs   = serverDB.findUser(jobs[idx]->getOwner());
		const char           * name    = uJobs->getName();
		PBS_AccountingRecord * uRec    = NULL;
		time_t                 now     = SchedStatic::getNow();
		
		// *************************************************************
		// * Obtain the user's share of system resources from the 
		// * scheduler configuration file. Use the defShare value as
		// * the default (if the user doesn't have a defined share).
		// *************************************************************
		double userShare = params.getConfigFloat("Fair Share", (char *)name, defShare);
		
		// *************************************************************
		// * Increment the userCnt variable.
		// *************************************************************
		userCnt++;
			
		// *************************************************************
		// * Create a new UserShare object that represents 
		// * this user's historical utilization and share of 
		// * system resources.
		// *************************************************************
		if(curr)	
			{
			curr->next = new UserShare();
			curr = curr->next;
			}
		else	{
			head = new UserShare();
			curr = head;	
			}

		// *************************************************************
		// * Install the ownerID and the user's share into 
		// * the UserShare object.
		// *************************************************************
		curr->setName(name);
		curr->ownerID  = ownerID;
		curr->share    = userShare; 
		total.share   += userShare;
			
		// *************************************************************
		// * Track the minimum share during processing. This
		// * will be needed if any of the users were given a
		// * negative share.
		// *************************************************************
		if(curr->share < minShare) 
			minShare = curr->share;
			 
		// *************************************************************
		// * Walk through each record returned from the 
		// * PBS_AccountingDatabase and add the utilization values 
		// * to the UserShare record.
		// *************************************************************
		for(size_t uIdx=0; uJobs && (uRec=(*uJobs)[uIdx])!=NULL; uIdx++)
			{
			// *****************************************************
			// * If the job has been started... then add the 
			// * statistics.
			// *****************************************************
			if(uRec->getStartTime()!=0)
				{
				time_t endTime = uRec->getEndTime();
				double cpu     = (double)uRec->getCpuTime()/3600.0;
				double wall    = (double)uRec->getWallTime()/3600.0;
				double job     = 1.0;

				if(endTime==0 || now-endTime <= rThreshold)
					{
					total.cpu[UserShare::RECENT]+=cpu;
					total.wall[UserShare::RECENT]+=wall;
					total.jobs[UserShare::RECENT]+=job;
					curr->cpu[UserShare::RECENT]+=cpu;
					curr->wall[UserShare::RECENT]+=wall;
					curr->jobs[UserShare::RECENT]+=job;
					}
				else if(now-endTime <= oThreshold)
					{
					total.cpu[UserShare::OLD]+=cpu;
					total.wall[UserShare::OLD]+=wall;
					total.jobs[UserShare::OLD]+=job;
					curr->cpu[UserShare::OLD]+=cpu;
					curr->wall[UserShare::OLD]+=wall;
					curr->jobs[UserShare::OLD]+=job;
					}
				}
			}
		
		// *************************************************************
		// * Using the computed statistics, calculate this
		// * user's combined utilization statistics.
		// *************************************************************
		curr->actual =	(curr->cpu [UserShare::RECENT]+
				(curr->cpu [UserShare::OLD]*oldFactor))*
				cpuFactor+
				(curr->wall[UserShare::RECENT]+
				(curr->wall[UserShare::OLD]*oldFactor))*
				wallFactor+
				(curr->jobs[UserShare::RECENT]+
				(curr->jobs[UserShare::OLD]*oldFactor))*
				runFactor+1;
		total.actual += curr->actual;
	
		// *************************************************************
		// * Walk through each of the jobs that is associated with this 
		// * particular user.
		// *************************************************************
		while(idx<jobCnt && jobs[idx]->getOwnerID()==ownerID) idx++;
		} while(idx<jobCnt);

	// *********************************************************************
	// * If the minimum share is less than 1, then advance all of the
	// * priorities AND the totals sufficiently to make all values greater
	// * than 0.
	// *********************************************************************
	if(minShare<1) total.share -= (minShare-1)*userCnt;

	// *********************************************************************
	// * Using the existing statistics and totals, walk through the list of
	// * users and compute what the expected utilization should have been.
	// * Using this value in conjunction with the actual utilization, 
	// * compute what the projected utilization should be.
	// *
	// * Note: If fewer than two users were detected - then these 
	// * computations will be skipped because it will have no effect on the
	// * outcome of processing.
	// *********************************************************************
	for(curr = head; userCnt>1 && curr!=NULL; curr=curr->next)
		{
		if(minShare<1) curr->share -= (minShare-1);
		// *************************************************************
		// * Compute the expected utilization based on actual use and
		// * percentage of overall priorities.
		// *************************************************************
		curr->expected = total.actual * (curr->share/total.share);

		// *************************************************************
		// * If the user had actual utilization, then project his
		// * future utilization as a proportion of what he expected 
		// * compared with what he actually got. If he had no 
		// * utilization, then set his projected utilization equal to 
		// * his share.
		// *************************************************************
		if(curr->actual > 0.0) curr->projected = curr->expected/curr->actual;
		else curr->projected = curr->share;

		// *************************************************************
		// * Keep track of the maximum projected utilization - this 
		// * value will be used to scale all values into the 0 - 1 range
		// * prior to computing weights.
		// *************************************************************
		if(curr->projected > total.projected) 
			total.projected = curr->projected;
		}
	
	// *********************************************************************
	// * Now we will walk through the user list again, and compute the
	// * actual weight associated with the user and apply this weight to
	// * each job that is associated with that user.
	// *
	// * Note: If fewer than two users were detected - then these 
	// * computations will be skipped because it will have no effect on the
	// * outcome of processing.
	// *********************************************************************
	for(curr=head, idx=0; userCnt>1 && curr!=NULL && idx<jobCnt; curr=curr->next)
		{
		double userWeight = (curr->projected/total.projected)*(1-LWB)+LWB;

		// *****************************************************
		// * If a Stage Debugging Recorder object was provided,
		// * then use it to report the results.
		// *****************************************************
		if(recorder)
			{
			recorder->stagePrintf("[User Share Stage]",
				"User %s - Share          : %g", curr->getName(), curr->share);
			recorder->stagePrintf("[User Share Stage]", 
				"User %s - Wall Time      : %g old hours, %g recent hours", 
				curr->getName(), 
				curr->wall[UserShare::OLD],
				curr->wall[UserShare::RECENT]);
			recorder->stagePrintf("[User Share Stage]", 
				"User %s - CPU Time       : %g old hours, %g recent hours", 
				curr->getName(), 
				curr->cpu[UserShare::OLD],
				curr->cpu[UserShare::RECENT]);
			recorder->stagePrintf("[User Share Stage]", 
				"User %s - Job Count      : %g old jobs, %g recent jobs", 
				curr->getName(), 
				curr->jobs[UserShare::OLD],
				curr->jobs[UserShare::RECENT]);
			recorder->stagePrintf("[User Share Stage]",
				"User %s - Applied Weight : %g", curr->getName(), userWeight);
			}								

		// *************************************************************
		// * Apply this weight to each of the user's jobs in the array.
		// *************************************************************
		while(idx<jobCnt && jobs[idx]->getOwnerID()==curr->ownerID)
			{
			double oldWeight = jobs[idx]->getWeight();
			if(oldWeight<0.0 || oldWeight>1.0) oldWeight = 1.0;
			jobs[idx]->setWeight(oldWeight*userWeight);

			// *****************************************************
			// * If a Stage Debugging Recorder object was provided,
			// * then use it to report the results.
			// *****************************************************
			if(recorder)
				{
				recorder->jobPrintf(jobs[idx]->getName(),
					"[User Share Stage] User %s share of %g - weight of %g applied", 
					curr->getName(), curr->share, userWeight);
				}

			idx++;
			}
		}

	// *********************************************************************
	// * Delete all of the temporary objects that were created for this
	// * method.
	// *********************************************************************
	while(head!=NULL)
		{
		curr = head;
		head = curr->next;
		delete curr;
		}
	}
