#include <UnderLord.h>
#include <JobDeployment.h>
#include <PBS_Parameters.h>
#include <PBS_Log.h>
#include <stdio.h>

// *****************************************************************************
// * UnderLordStage::UnderLordStage :
// *	This is the constructor for the UnderLordStage object. It is defined
// *	out of line to ensure that the class has a basis in the object files.
// *****************************************************************************
UnderLordStage::UnderLordStage( void ) 
	: next(NULL), recorder(NULL) 
	{
	}

// *****************************************************************************
// * UnderLordStage::setRecorder :
// *	This method is called to install a new StageRecorder on the 
// *	UnderLordStage object.  A pointer to the previously installed 
// *	StageRecorder object is returned.
// *****************************************************************************
StageRecorder * UnderLordStage::setRecorder ( StageRecorder * Recorder )
	{
	StageRecorder * oldRecorder = recorder;
	recorder = Recorder;
	return oldRecorder;
	}

// *****************************************************************************
// * UnderLord::UnderLord :
// *	This is the constructor for the UnderLord class it will initialize
// *	the local variables and call the constructors for contained object.
// *****************************************************************************
UnderLord::UnderLord ( void ) 
	: server      (),
	  matrix      (),
	  lastSummary (0),
	  stages      (NULL)
	{
	LogRecord("UnderLord", "Creating UnderLord Object");
	}
	
// *****************************************************************************
// * UnderLord::addStage :
// *	This method is called by the user to add a new job weight stage to the
// *	UnderLord scheduler.  The scheduler will call each of these job weight
// *	stages in the order of their addition prior to sorting the jobs.
// *****************************************************************************
void UnderLord::addStage ( UnderLordStage * stage )
	{
	if(stage)
		{
		if(stages)
			{
			UnderLordStage * lastStage = stages;
			while(lastStage->next!=NULL) lastStage = lastStage->next;
			lastStage->next = stage;
			}
		else stages = stage;
		}
	}

// *****************************************************************************
// * UnderLord::processJobs :
// *	This method is called when it is time for the UnderLord scheduler to 
// *	process and dispatch jobs.  The method will extract a list of queued 
// *	jobs from the server, will order the jobs using internal alogorithms,
// *	will submit the jobs to the MatchMatrix for matching with available 
// *	nodes and will the submit those jobs to the server for processing. 
// *****************************************************************************
void UnderLord::processJobs ( int connID )
	{	
	// *********************************************************************
	// * Get the most recent data from the PBS_Server. If this doesn't work
	// * and the local connection exists - close it.
	// *********************************************************************
	if(connID>=0 && server.loadData(connID)<0)
		{
		LogError("UnderLord", "Error: Failed to communicate with server\n");
		connID = -1;
		}
	
	// *********************************************************************
	// * If the connID is not equal to -1, then data was successfully 
	// * extracted from the server and may be processed.
	// *********************************************************************
	if(connID>=0)
		{
		// *************************************************************
		// * The PBS_Parameters object provides an interface to the
		// * configuration files for the Scheduling system.
		// *************************************************************
		PBS_Parameters & parameters  = PBS_Parameters::attach();

		// *********************************************************************
		// * Call the SchedStatic::startCycle method to update the value of now.
		// *********************************************************************
		SchedStatic::startCycle();
	
		// *********************************************************************
		// * Call the MatchMatrix's startCycle method to begin the 
		// * processing cycle.
		// *********************************************************************
		matrix.startCycle(server);

		// *************************************************************
		// * The now variable represents the current time at the  
		// * beginning of the scheduling cycle.
		// *************************************************************
		time_t now = SchedStatic::getNow();
	
		// *************************************************************
		// * The frequency at which a summary of Job Deployments is 
		// * written is obtained from the PBS_Parameters objects.
		// *************************************************************
		time_t frequency = 
			SchedStatic::stringToDuration(
				parameters.getConfigString(
					"Scheduler", 
					"summaryRate", 
					"10:00"), 600);
	

		// *************************************************************
		// * Read the execSummary flag from the configuration file. If
		// * this flag is yes, then a summary of when jobs are projected
		// * to execute will be generated at the specified frequency.
		// *************************************************************
		int execSummary = 
			!strcasecmp("yes", 
				parameters.getConfigString(
					"Scheduler", 
					"execSummary", "yes"));
		// *************************************************************
		// * Read the stageSummary flag from the configuration file. If
		// * this flag is yes, then a report outlining the parameters
		// * used for scheduling decisions will be generated at the 
		// * specified frequency.
		// *************************************************************
		int stageSummary = 
			!strcasecmp("yes", 
				parameters.getConfigString(
					"Scheduler", 
					"stageSummary", "yes"));

		// *************************************************************
		// * The complete flag is used to indicate if the MatchMatrix  
		// * should perform a full scheduling cycle and generate an
		// * exhaustive summary, or conversly, only generate a list of 
		// * jobs that are eligible for immediate execution.
		// *************************************************************
		int quick = (now-lastSummary)<frequency;
		
		// *************************************************************
		// * Their are several types of jobs that are maintained in the 
		// * system. for our purposes, the most significant ones are 
		// * those that are queued and those that are waiting.
		// *
		// * A queued job is one that is immediately eligible for
		// * execution IF the queue is enabled and started. A waiting 
		// * job is one that must wait until another jobs is completed 
		// * in order to execute.  For the first draft of the UnderLord, 
		// * we will only consider those jobs that are queued to perform
		// * scheduling operations.
		// *************************************************************
		PBS_Queue    ** queues    = server.getQueues();
		PBS_Job      ** jobs      = NULL;
		size_t          jobCnt    = 0;
		size_t          jobIdx    = 0;
		size_t          queueIdx  = 0;
		JobDeployment * scheduled = NULL;
		
		// *************************************************************
		// * Walk through the list of queues retrieved from the server.
		// *************************************************************
		for(queueIdx=0; queues && queues[queueIdx]; queueIdx++)
			{
			PBS_Job ** qJobs = NULL;
			// *****************************************************
			// * If a queue is enabled, started, AND an execution
			// * queue, extract it's list of jobs.
			// *****************************************************
			if(queues[queueIdx]->getEnabledFlag() &&
			   queues[queueIdx]->getStartedFlag() &&
			   queues[queueIdx]->getType()==PBS_Queue::EXECUTION &&
			   (qJobs = queues[queueIdx]->getJobsByState(PBS_Job::QUEUED))!=NULL)
				{
				// *********************************************
				// * Count the jobs that were extracted from
				// * this queue.
				// *********************************************
				for(jobIdx=0; qJobs && qJobs[jobIdx]; jobIdx++);

				// *********************************************
				// * Realloc the jobs array to hold the current
				// * list of jobs + the new list + a NULL 
				// * terminator;
				// *********************************************
				if(jobIdx)
					{
					jobs = (PBS_Job **)realloc(jobs, sizeof(PBS_Job *)*(jobCnt+jobIdx+1));
					memcpy(jobs+jobCnt, qJobs, sizeof(PBS_Job *)*(jobIdx+1));
					jobCnt+=jobIdx;
					}
				
				// *********************************************
				// * Release the memory that was temporarily
				// * assigned to the qJobs array.
				// *********************************************
				delete[] qJobs;
				}
			}

		// *************************************************************
		// * Perform job sorting here using the locally developed 
		// * algorithms. Us the quick flag as an indicator of whether
		// * the sort steps should be recorded in the StageRecorder.
		// ************************************************************* 
		sortJobs(server, jobs, jobCnt, (!quick && stageSummary));
		
		// *************************************************************
		// * Pass the sorted array of jobs to the MatchMatrix to be
		// * matched against available nodes. Note the quick flag is set 
		// * to FALSE if it is time to generate another summary for the 
		// * log file.  If the user has specified that they do not want 
		// * execution summaries to be generated, then every sort will 
		// * be quick.
		// *************************************************************
		scheduled = matrix.processJobs(jobs, jobCnt, (quick || !execSummary));
		
		// *************************************************************
		// * Call the deployJobs method to begin execution of jobs that 
		// * are scheduled for immediate deployment.
		// *************************************************************
		deployJobs ( server, scheduled );

		// *************************************************************
		// * Generate a schedule report that identifies which jobs are
		// * scheduled on what machines and when they are projected to
		// * start. Sumamries will not be generated if the user has
		// * set the execSummary entry in the scheduler.cfg to FALSE.
		// *************************************************************
		if(scheduled && execSummary && !quick)
			{
			writeSummary ( server, scheduled );
			lastSummary = now;
			}
		
		// *************************************************************
		// * Free the jobs array that was allocated by malloc/realloc.
		// *************************************************************
		free(jobs);
		
		// *************************************************************
		// * Automicatically call endCycle to terminate this schedule
		// * cycle and prevent a recurrent call to the processJobs
		// * method using the same jobList.
		// *************************************************************
		matrix.endCycle();

		// *************************************************************
		// * Call the SchedStatic::endCycle method to clear global 
		// * variables.
		// *************************************************************
		SchedStatic::endCycle();
		}		
	}

// *****************************************************************************
// * UnderLord::~UnderLord :
// *	This method deletes all temporary variables that may have been used
// *	by the UnderLord object.
// *****************************************************************************
UnderLord::~UnderLord ( void )
	{
	UnderLordStage * current;
	while((current=stages)!=NULL)
		{
		stages = current->next;
		delete current;
		}
	}

// *****************************************************************************
// * UnderLord::sortJobs :
// *	This method is called to sort the list of jobs based on locally defined
// *	criteria.  This sorted list of jobs may then be submitted to the 
// *	MatchMatrix for selection of the best candidate node and execution time.
// *****************************************************************************
void UnderLord::sortJobs ( PBS_Server & server, PBS_Job ** jobs, size_t jobCnt, int record )
	{
	UnderLordStage * stage;
	double           weightSum  = 0.0;
	FILE           * fp         = NULL;
	time_t           now        = SchedStatic::getNow();

	if(record)
		{
		char             logName[1024];
		PBS_Parameters & parameters = PBS_Parameters::attach();		
		struct tm      * t          = localtime(&now);

		sprintf(logName, "%s/%04i%02i%02i.sorting",
			parameters.getConfigString("Scheduler", "summaryDir", "/usr/spool/PBS/sched_logs"),
			1900+t->tm_year, t->tm_mon+1, t->tm_mday);
		
		if((fp=fopen(logName, "a"))!=NULL) fseek(fp, 0L, SEEK_END);
		else LogError("UnderLord", "Failed to open output file %s", logName);
		}			

	if(fp!=NULL)
		{
		StageRecorder    recorder;
		StageRecorder  * prior;
		
		for(stage=stages; stage!=NULL; stage=stage->next)
			{
			prior = stage->setRecorder(&recorder);
			weightSum += stage->getWeight();
			stage->setRecorder(prior);
			}
		
		for(stage=stages; stage!=NULL; stage=stage->next)
			{
			prior = stage->setRecorder(&recorder);
			stage->compute(jobs, jobCnt, weightSum);
			stage->setRecorder(prior);
			}
		
		for(size_t idx=0; idx<jobCnt; idx++)
			{
			recorder.stagePrintf(
				jobs[idx]->getName(), 
				"Computed weight from all stages : %g", 
				jobs[idx]->getWeight());
			}
			
		qsort (jobs, jobCnt, sizeof(PBS_Job *), UnderLord::compareFunc);

		fprintf(fp, "--------------------------------------------------------\n");
		fprintf(fp, "UnderLord Sorting Summary for %s", ctime(&now));
		fprintf(fp, "--------------------------------------------------------");
		recorder.asciiDump(fp);
		fprintf(fp, "\n");
		fclose(fp);
		}
	else	{
		for(stage=stages; stage!=NULL; stage=stage->next)
			weightSum += stage->getWeight();

		for(stage=stages; stage!=NULL; stage=stage->next)
			stage->compute(jobs, jobCnt, weightSum);

		qsort (jobs, jobCnt, sizeof(PBS_Job *), UnderLord::compareFunc);		
		}
	}


// *****************************************************************************
// * UnderLord::sorter :
// *	This is the method called by qsort to sort the jobs based on their
// *	weights.
// *****************************************************************************
int UnderLord::compareFunc (const void * a, const void * b)
	{
	PBS_Job * jobA = *(PBS_Job **)a;
	PBS_Job * jobB = *(PBS_Job **)b;
	double result = jobB->getWeight() - jobA->getWeight();
	return (result==0)?0:(result<0)?-1:1;
	}
	
// *****************************************************************************
// * UnderLord::deployJobs :
// *	This method is called to deploy a list of jobs that are scheduled for
// *	immediate execution.
// *****************************************************************************
void UnderLord::deployJobs ( PBS_Server & server, JobDeployment * scheduled )
	{
	time_t          now      = SchedStatic::getNow();
	JobDeployment * job      = scheduled;
	const char    * resource = "neednodes";
	char          * nodeStr  = NULL;
	
	while(job!=NULL)
		{
		PBS_Job * jPtr = server.getJob(job->getJob());
		
		if(job->getStartTime()<=now &&
		   jPtr && jPtr->getState()==PBS_Job::QUEUED)
			{
			struct attrl attrib;
			
			// *****************************************************
			// * Set up the attrib structure to identify the nodes
			// * required in the neednodes resource.
			// *****************************************************
			attrib.name     = ATTR_l;
			attrib.resource = (char *)resource;
			attrib.value    = (nodeStr=(char *)job->getNodeStr());
			attrib.next     = NULL;
			
			// *****************************************************
			// * Alter the job's neednodes resource using the 
			// * attrib structure.  Report an error to the log
			// * file if this was unsuccessful.
			// *****************************************************
			if(pbs_alterjob(server.getConnectionID(), (char *)job->getJob(), &attrib, NULL)!=0)
				{
				LogError("UnderLord", "Failed to alter job %s", job->getJob());
				LogError("UnderLord", "Job %s error code is %i : %s\n", job->getJob(), pbs_errno, pbse_to_txt(pbs_errno));
				}

			// *****************************************************
			// * Execute the job.  In the event that the neednodes
			// * variable could not be successfully set, the server
			// * will still execute the job - on the first nodes 
			// * that match the specification.
			// *****************************************************
			if(pbs_runjob(server.getConnectionID(), (char *)job->getJob(), NULL, NULL)!=0)
				{
				LogError("UnderLord", "Failed to run job %s", job->getJob());
				LogError("UnderLord", "Job %s error code is %i : %s\n", job->getJob(), pbs_errno, pbse_to_txt(pbs_errno));
				}
			else	{
				LogRecord("UnderLord", "Job %s started on nodes %s", job->getJob(), nodeStr);
				}		
			}
		job=job->getNext();
		}
	}
	
// *****************************************************************************
// * UnderLord::writeSummary :
// *	This method is called to periodically write a summary of the jobs that
// *	are scheduled for execution and when and where they will execute.
// *****************************************************************************
void UnderLord::writeSummary ( PBS_Server & server, JobDeployment * scheduled )
	{
	FILE           * fp         = NULL;
	PBS_Parameters & parameters = PBS_Parameters::attach();
	time_t           now        = SchedStatic::getNow();
		
	char logName[1024];
	// *********************************************************************
	// * Convert the current time into a tm structure.
	// *********************************************************************
	struct tm * t = localtime(&now);

	// *********************************************************************
	// * Use the current date to generate a file name to which 
	// * summary information should be written.
	// *********************************************************************
	sprintf(logName, "%s/%04i%02i%02i.summary",
		parameters.getConfigString("Scheduler", "summaryDir", "/usr/spool/PBS/sched_logs"),
		1900+t->tm_year, t->tm_mon+1, t->tm_mday);

		
	// *********************************************************************
	// * Open the log file in append mode.
	// *********************************************************************
	if((fp=fopen(logName, "a"))!=NULL)
		{
		fseek(fp, 0L, SEEK_END);
		fprintf(fp, "Projected Job Scheduling for %02i/%02i/%04i %02i:%02i\n", 
			t->tm_mon+1, t->tm_mday, 1900+t->tm_year, t->tm_hour, t->tm_min);
	
		// *************************************************************
		// * Walk through each of the JobDeployment objects and
		// * write its contents to the summary file.
		// *************************************************************
		while(scheduled)
			{
			scheduled->asciiDump(fp, "   ");
			scheduled=scheduled->getNext();
			}
		fprintf(fp, "\n");
		
		// *************************************************************
		// * Close the output file.
		// *************************************************************
		fclose(fp);
		}
	else LogError("UnderLord", "Failed to open output file %s\n", logName);
	}

