#ifndef _UNDERLORD_STAGE_H_ 
#define _UNDERLORD_STAGE_H_ 1

#include <PBS_Job.h>
#include <stdarg.h>
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>
#include <PBS_Hash.h>

class UnderLordStage
{
friend class UnderLord;

private:
	UnderLordStage * next;

protected:
	class StageRecorder * recorder;
	
public:
	UnderLordStage             ( void );
	StageRecorder * setRecorder ( StageRecorder * Recorder );
	virtual double  getWeight   ( void ) = 0;
	virtual void    compute     ( PBS_Job ** jobs, size_t jobCnt, double weightSum ) = 0;
};



class StageRecorderEntry
{
private:
	StageRecorderEntry * next;
	char               * name;
	char              ** strings;
	size_t               max;
	size_t               cur;
	
public:	
	inline StageRecorderEntry  ( const char * Name );
	inline ~StageRecorderEntry ( void );
	inline void addString   ( const char * str ); 
	inline void asciiDump   ( FILE * fp = stdout, const char * prefix = NULL);

	inline char               * getName ( void ) { return name; }
	inline StageRecorderEntry * getNext ( void ) { return next; }
	inline void                 setNext ( StageRecorderEntry * Next ) { next = Next; }
};


class StageRecorder 
{
private:
	StageRecorderEntry * stageHead;
	StageRecorderEntry * stageTail;
	StageRecorderEntry * jobHead;
	StageRecorderEntry * jobTail;
	PBSStringHash        entries;
	
public:
	inline StageRecorder  ( void );
	inline ~StageRecorder ( void );	

	inline void addStageString ( const char * stage, const char * str );
	inline void stagePrintf    ( const char * job, const char * fmt, ... ); 

	inline void addJobString   ( const char * job, const char * str );
	inline void jobPrintf      ( const char * job, const char * fmt, ... );

	inline void clear          ( void );
	inline void asciiDump      ( FILE * fp = stdout, const char * prefix = NULL);
};


StageRecorderEntry::StageRecorderEntry ( const char * Name )
	: next   (NULL),
	  name   (strdup((Name && *Name)?Name:"Unknown")),
	  strings(NULL),
	  max    (16),
	  cur    (0)
	{
	strings = (char **)malloc(max*sizeof(char *));
	}
	
StageRecorderEntry::~StageRecorderEntry ( void )
	{
	while(cur>0) free ( strings[--cur] );
	free (strings);
	free (name);
	}
	
void StageRecorderEntry::addString ( const char * str )
	{
	if(cur+1 >= max)
		{
		max *= 2;
		strings = (char **)realloc(strings, max*sizeof(char*));
		}
	strings[cur++] = strdup(str);
	}
	
void StageRecorderEntry::asciiDump ( FILE * fp, const char * prefix )
	{
	char c = 0;
	const char * prefixString = (prefix?prefix:&c);
	
	fprintf(fp, "%s\n%sStage Debugging Entries for %s\n", 
		prefixString, prefixString, name);
	for(int idx=0; idx<cur; idx++)
		{
		fprintf(fp, "%s   %s\n", prefixString, strings[idx]); 
		}
	}
	
	
StageRecorder::StageRecorder ( void )
	: entries  (), 
	  stageHead(NULL), 
	  stageTail(NULL), 
	  jobHead  (NULL), 
	  jobTail  (NULL)
	{
	}
	
StageRecorder::~StageRecorder ( void )
	{
	clear();
	}

void StageRecorder::clear ( void )
	{
	while(stageHead!=NULL)
		{
		stageTail = stageHead;
		stageHead = stageTail->getNext();
		entries.remove(stageTail->getName());
		delete stageTail;
		}
	while(jobHead!=NULL)
		{
		jobTail = jobHead;
		jobHead = jobTail->getNext();
		entries.remove(jobTail->getName());
		delete jobTail;
		}	

	stageHead = NULL;
	stageTail = NULL;
	jobHead   = NULL;
	jobTail   = NULL;	
	}
		
void StageRecorder::addJobString ( const char * job, const char * string )
	{
	if(job && *job && string && *string)
		{
		StageRecorderEntry * entry;

		if((entry= (StageRecorderEntry *)entries.find(job))==NULL)
			{
			entry = new StageRecorderEntry(job);
			entries.insert(entry->getName(), entry);

			if(jobHead==NULL) jobHead = entry;
			else           jobTail->setNext(entry);
			jobTail = entry;
			}
		entry->addString(string);
		}	
	}
	
void StageRecorder::jobPrintf ( const char * job, const char * fmt, ... )
	{
	char msgBuf[1024];
	
	if(job && *job && fmt && *fmt)
		{
		va_list argp;
		va_start(argp, fmt);
		vsprintf(msgBuf, fmt, argp);
		va_end(argp);

		addJobString(job, msgBuf);
		}
	}
	
void StageRecorder::addStageString ( const char * stage, const char * string )
	{
	if(stage && *stage && string && *string)
		{
		StageRecorderEntry * entry;

		if((entry= (StageRecorderEntry *)entries.find(stage))==NULL)
			{
			entry = new StageRecorderEntry(stage);
			entries.insert(entry->getName(), entry);

			if(stageHead==NULL) stageHead = entry;
			else stageTail->setNext(entry);
			stageTail = entry;
			}
		entry->addString(string);
		}	
	}
	
void StageRecorder::stagePrintf ( const char * stage, const char * fmt, ... )
	{
	char msgBuf[1024];
	
	if(stage && *stage && fmt && *fmt)
		{
		va_list argp;
		va_start(argp, fmt);
		vsprintf(msgBuf, fmt, argp);
		va_end(argp);

		addStageString(stage, msgBuf);
		}
	}
	
void StageRecorder::asciiDump ( FILE * fp, const char * prefix )
	{
	StageRecorderEntry * entry = stageHead;
	while(entry!=NULL)
		{
		entry->asciiDump(fp, prefix);
		entry = entry->getNext();
		}
	entry = jobHead;
	while(entry!=NULL)
		{
		entry->asciiDump(fp, prefix);
		entry = entry->getNext();
		}
	}

#endif /* _UNDERLORD_STAGE_H_ */
